import fs from "node:fs";
import { createRequire } from "node:module";
import Discord from "discord.js";
import SQLite from "better-sqlite3";
import bootWeb from "./webfront/index.js";
import logging from "./utils/logging.js";
import common from "./utils/common.js";

const require = createRequire(import.meta.url);

const sql = new SQLite(`./src/db.sqlite`);

const client = new Discord.Client({
    partials: [Discord.Partials.Channel],
    intents: [
        Discord.GatewayIntentBits.Guilds,
        Discord.GatewayIntentBits.GuildMembers,
        Discord.GatewayIntentBits.GuildBans,
        Discord.GatewayIntentBits.GuildMessages,
        Discord.GatewayIntentBits.MessageContent,
        Discord.GatewayIntentBits.GuildMessageReactions,
        Discord.GatewayIntentBits.DirectMessages,
        Discord.GatewayIntentBits.GuildVoiceStates,
    ],
});

/**
 * @name loadLocalization
 * @description Loads the localization files.
 */
async function loadLocalization()
{
    global.locale = {};
    // Load all JSON files in ./locale and JSON parse them, then add them to the global.locale object
    const locales = fs.readdirSync(`./src/locale`).filter((file) => file.endsWith(`.json`));
    for (const locale of locales)
    {
        const data = JSON.parse(fs.readFileSync(`./src/locale/${ locale }`));
        const localeName = locale.split(`.`)[0];
        global.locale[localeName] = data;
    }

    // Get the defined locale from the config file
    const { locale } = require(`./config.json`);
    global.gLocale = locale;
}

await initSetup();

await loadLocalization(); // load all localizations first, so we can translate everything below.

const { token, isWorkflow, debugMode } = require(`./config.json`);
if (token.length < 50)
    logging.error(`${ global.locale[global.gLocale].invalidToken } `, true);

global.reactionMessageIds = []; // React to Delete Bot Message 1
global.elevatedServers = [`808962821697306654`, `872330479565680640`]; // Prod & testing

await setupDB();
await setupCommands();

client.login(token);
bootWeb(client, Discord);

if (!isWorkflow && !debugMode)
{ // catch everything if not workflow AND not dev env
    process.on(`unhandledRejection`, (error) =>
    {
        const d8 = new Date(Date.now());
        logging.error(`${ global.locale[global.gLocale].unhandledRejection } [${ d8.toDateString() } ${ d8.toTimeString() }]: ${ error }`, false, true);
    });

    process.on(`uncaughtException`, (error) =>
    {
        const d8 = new Date(Date.now());
        logging.error(`${ global.locale[global.gLocale].uncaughtException } [${ d8.toDateString() } ${ d8.toTimeString() }]: ${ error }`, false, true);
    });
}

/**
 * @name initSetup
 * @description Setup config.json if it doesn't exist.
 */
async function initSetup()
{
    if (fs.existsSync(`./src/config.json`))
    {
        const rawConfig = fs.readFileSync(`./src/config.json`);
        const config = JSON.parse(rawConfig);
        const rawDefaultConfig = fs.readFileSync(`./src/config.json.example`);
        const defaultConfig = JSON.parse(rawDefaultConfig);
        for (const key in defaultConfig)
        { // If we already have a config, check and see if everything from defaultConfig is set.
            if (!(key in config))
                config[key] = defaultConfig[key];
        }
        for (const key in config)
        { // Clean Legacy Config Options. (If something in config is not in defaultConfig, remove it from config).
            if (!(key in defaultConfig))
                delete config[key];
        }
        fs.writeFileSync(`./src/config.json`, JSON.stringify(config, undefined, 4));
        return;
    }

    fs.copyFileSync(`./src/config.json.example`, `./src/config.json`);

    var readlineSync = require(`readline-sync`);

    let localeList = fs.readdirSync(`./src/locale`).filter((file) => file.endsWith(`.json`));
    localeList = localeList.map((file) => file.split(`.`)[0]);
    localeList = localeList.join(`, `);

    let locale = readlineSync.question(`Please enter your locale (${ localeList }): `).trim().toLowerCase();
    const locales = fs.readdirSync(`./src/locale`).filter((file) => file.endsWith(`.json`));
    if (!locales.includes(`${ locale }.json`))
        locale = `en`; // Set to en if the locale is not found.

    // Load locales
    await loadLocalization();

    let token = ``;
    do
    {
        token = readlineSync.question(`${ global.locale[locale].setupBotToken } `, {
            hideEchoBack: true,
        });
    }
    while (token.length < 50);

    let clientSecret = ``;
    do
    {
        clientSecret = readlineSync.question(`${ global.locale[locale].setupClientSecret } `, {
            hideEchoBack: true,
        });
    }
    while (clientSecret.length < 20);

    let botOwners = readlineSync.question(`${ global.locale[locale].setupBotOwners } `);
    if (botOwners.length > 1)
    {
        botOwners = botOwners.split(`,`);
        for (let index = 0; index < botOwners.length; index++)
            botOwners[index] = botOwners[index].trim(); // this also handles "1, 2, 3"
    }
    else
        botOwners = [];

    const playingStatus = readlineSync.question(`${ global.locale[locale].setupPlayingStatus } `);

    let debugMode = readlineSync.question(`${ global.locale[locale].setupDebugMode } `);
    debugMode = !(debugMode.toLowerCase() !== `true` && debugMode.toLowerCase() !== `y` && debugMode.toLowerCase !== `yes`);

    let forwardDMs = readlineSync.question(`${ global.locale[locale].setupForwardDM } `);
    forwardDMs = !(forwardDMs.toLowerCase() !== `true` && forwardDMs.toLowerCase() !== `y` && forwardDMs.toLowerCase !== `yes`);

    // ask what port to use
    let webfrontPort = readlineSync.question(`${ global.locale[locale].setupWebfrontPort } `);
    webfrontPort = webfrontPort.trim();
    webfrontPort = webfrontPort.length === 0 ? 5000 : Number.parseInt(webfrontPort, 10);
    if (Number.isNaN(webfrontPort) || webfrontPort < 1000 || webfrontPort > 65_534)
    {
        webfrontPort = 5000;
        logging.log(`${ global.locale[locale].setupInvalidPort }`);
    }

    // ask what url to use (default: http://localhost:port)
    var url = readlineSync.question(`${ global.locale[locale].setupWebfrontURL.replaceAll(`%webfrontPort%`, `${ webfrontPort }`) } `);
    if (url.length === 0 || !url.startsWith(`http`))
        url = `http://localhost:${ webfrontPort }`;
    // remove trailing slash if found
    if (url.endsWith(`/`))
        url = url.slice(0, Math.max(0, url.length - 1));

    // ask for their webfront api key
    var webfrontAPIKey = ``;
    do
        webfrontAPIKey = readlineSync.question(`${ global.locale[locale].setupWebfrontAPIKey } `);
    while (webfrontAPIKey.length < 20);

    readlineSync.question(`\n${ global.locale[locale].setupOauthRedirect.replaceAll(`%url%`, `${ url }`) }\n`);

    const config = require(`./config.json`);
    config.token = token.trim();
    config.locale = locale;
    config.botOwnerID = botOwners;
    config.playingStatus = playingStatus.trim();
    config.clientSecret = clientSecret.trim();
    config.debugMode = debugMode;
    config.forwardDMs = forwardDMs;
    config.webfrontPort = webfrontPort;
    config.webfrontURL = url.trim();
    config.webfrontAPIKey = webfrontAPIKey.trim();
    fs.writeFileSync(`./src/config.json`, JSON.stringify(config, undefined, 4)); // save settings to config
}

/**
 * @name setupCommands
 * @description Loads and prepares Chat/Slash Commands & Events.
 */
async function setupCommands()
{
    // slash commands
    client.commands = new Discord.Collection();
    client.chatcommands = new Discord.Collection();

    const commandFolders = fs.readdirSync(`./src/commands/slash_commands`);
    const chatcommandFolders = fs.readdirSync(`./src/commands/chat_commands`);

    let numberFailed = 0;

    for (const folder of commandFolders)
    {
        const commandFiles = fs.readdirSync(`./src/commands/slash_commands/${ folder }`).filter((file) => file.endsWith(`.js`));
        for (const file of commandFiles)
        {
            try
            {
                let command = await import(`./commands/slash_commands/${ folder }/${ file }`);
                command = command.default || command;
                command = await injectLocale(command);
                client.commands.set(command.name, command);
                if (command.elevatedOnly)
                    common.elevatedCommands.push(command);
                else
                    common.commands.push(command);
            }
            catch (error)
            {
                logging.error(`${ global.locale[global.gLocale].failedToLoadCommand.replaceAll(`%file%`, file).replaceAll(`%error%`, error) }`);
                numberFailed += 1;
            }
        }
    }

    for (const folder of chatcommandFolders)
    {
        const commandFiles = fs.readdirSync(`./src/commands/chat_commands/${ folder }`).filter((file) => file.endsWith(`.js`));
        for (const file of commandFiles)
        {
            try
            {
                let command = await import(`./commands/chat_commands/${ folder }/${ file }`);
                command = command.default || command;
                client.chatcommands.set(command.name, command);
            }
            catch (error)
            {
                logging.error(`${ global.locale[global.gLocale].failedToLoadCommand.replaceAll(`%file%`, file).replaceAll(`%error%`, error) }`);
                numberFailed += 1;
            }
        }
    }

    logging.log(`${ global.locale[global.gLocale].loadedCommands.replaceAll(`%numberFailed%`, numberFailed) }`);

    // events
    const eventFiles = fs.readdirSync(`./src/events`).filter((file) => file.endsWith(`.js`));
    for (const file of eventFiles)
    {
        let event = await import(`./events/${ file }`);
        event = event.default || event;
        if (event.once)
            client.once(event.name, (...ourArguments) => event.execute(...ourArguments, Discord, client));
        else
            client.on(event.name, (...ourArguments) => event.execute(...ourArguments, Discord, client));
    }
}

/**
 * @name setupDB
 * @description Create the database if it doesn't exist.
 */
async function setupDB()
{
    sql.prepare(`CREATE TABLE IF NOT EXISTS punishments (guildAndUserId TEXT, guildId TEXT, id TEXT, warns TEXT, mutes TEXT, kicks TEXT, bans TEXT, punishmentreason TEXT, punishmentinfo TEXT);`).run();
    sql.prepare(`CREATE UNIQUE INDEX IF NOT EXISTS idx_punishments ON punishments (guildAndUserId);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS mutes (guildAndUserId TEXT, guildId TEXT, userId TEXT, time TEXT, moderatorId TEXT);`).run();
    sql.prepare(`CREATE UNIQUE INDEX IF NOT EXISTS idx_mutes ON mutes (guildAndUserId);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS allowedLinks (guildId TEXT, link TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS autoBannedPhrases (guildId TEXT, phrase TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS badWords (guildId TEXT, word TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS badNames (guildId TEXT, name TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS allowedDiscords (guildId TEXT, id TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS keywords (guildId TEXT, phrase TEXT, response TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS keywordsRegional (guildId TEXT, phrase TEXT, response TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS scheduledMessages (guildId TEXT, channelId TEXT, time TEXT, content TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS memberCounts (guildId TEXT, count TEXT, time TEXT);`).run();
    sql.prepare(`CREATE UNIQUE INDEX IF NOT EXISTS idx_mc_gid ON memberCounts (guildId);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS banAppealTokens (guildId TEXT, userId TEXT, token TEXT, threadId TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS configuration (guildId TEXT PRIMARY KEY, config TEXT);`).run();
    sql.prepare(`CREATE UNIQUE INDEX IF NOT EXISTS idx_configuration_guildId ON configuration (guildId);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS keywordSpam (guildId TEXT, userId TEXT, count TEXT, time TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS doNotForwardDMsFrom (userId TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS userNotes (guildAndUserId TEXT, userId TEXT, guildId TEXT, note TEXT, meta TEXT);`).run();
    sql.prepare(`CREATE UNIQUE INDEX IF NOT EXISTS idx_userNotes_guildAndUserId ON userNotes (guildAndUserId);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS bannedServers (guildId TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS tempbanUsers (guildAndUserId TEXT, guildId TEXT, userId TEXT, time TEXT, moderatorId TEXT);`).run();
    sql.prepare(`CREATE UNIQUE INDEX IF NOT EXISTS idx_tempbans ON tempbanUsers (guildAndUserId);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS rolePersists (guildId TEXT, userId TEXT, roleArray TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS giveaways (guildId TEXT, messageId TEXT, time TEXT, amountOfWinners TEXT, channelId TEXT);`).run();
    sql.prepare(`CREATE TABLE IF NOT EXISTS giveawayEntries (guildId TEXT, messageId TEXT, userId TEXT);`).run();
    sql.pragma(`synchronous = 1`);
    sql.pragma(`journal_mode = wal`);
    process.on(`exit`, () => sql.close());
    process.on(`SIGHUP`, () => process.exit(128 + 1));
    process.on(`SIGINT`, () => process.exit(128 + 2));
    process.on(`SIGTERM`, () => process.exit(128 + 15));

    // punishment variables (quick access)
    global.getInfoById = sql.prepare(`SELECT * FROM punishments WHERE guildId = ? AND id = ?;`);
    global.setInfo = sql.prepare(`INSERT OR REPLACE INTO punishments (guildAndUserId, guildId, id, warns, mutes, kicks, bans, punishmentreason, punishmentinfo) VALUES (@guildAndUserId, @guildId, @id, @warns, @mutes, @kicks, @bans, @punishmentreason, @punishmentinfo);`);

    // guild configuration variables
    global.getConfiguration = sql.prepare(`SELECT * FROM configuration WHERE guildId = ?`);
    global.setConfiguration = sql.prepare(`INSERT OR REPLACE INTO configuration (guildId, config) VALUES (@guildId, @config);`);
}

/**
 * @name injectLocale
 * @description Inject Localization into Slash Commands (https://discord.com/developers/docs/interactions/application-commands#localization)
 * @param {object} command The command object to inject the locale into.
 */
async function injectLocale(command)
{
    // https://discord.com/developers/docs/reference#locales
    const discordLocales = [`da`, `de`, `en-GB`, `en-US`, `es-ES`, `fr`, `hr`, `it`, `lt`, `hu`, `nl`, `no`, `pl`, `pt-BR`, `ro`, `fi`, `sv-SE`, `vi`, `tr`, `cs`, `el`, `bg`, `ru`, `uk`, `hi`, `th`, `zh-CN`, `ja`, `zh-TW`, `ko`];
    command.nameLocalizations = {};
    command.descriptionLocalizations = {};
    for (const locale of discordLocales)
    {
        let splitLocale = locale;
        if (splitLocale.includes(`-`))
            splitLocale = splitLocale.split(`-`)[0];

        // Inject nameLocalizations & descriptionLocalizations into the command
        let commandName = `${ command.name }CommandName`;
        commandName = commandName.replaceAll(` `, `_`);
        const commandDescription = `${ command.name }CommandDescription`;
        const workingLocale = global.findLocale(undefined, splitLocale);
        if (workingLocale !== `en`)
        {
            // Add commandName to the findLocale
            command.nameLocalizations = { ...command.nameLocalizations, [locale]: global.findLocaleString(workingLocale, commandName) };
            if (command.description)
                // eslint-disable-next-line max-len
                command.descriptionLocalizations = { ...command.descriptionLocalizations, [locale]: global.findLocaleString(workingLocale, commandDescription) };

            // For each option in the command, inject the locale into the option
            if (command.options)
            {
                for (let index = 0; index < command.options.length; index++)
                {
                    const option = command.options[index];
                    let optionName = `${ command.name }OptionName${ index }`;
                    optionName = optionName.replaceAll(` `, `_`);
                    const optionDescription = `${ command.name }OptionDescription${ index }`;
                    if (workingLocale !== `en`)
                    {
                        // Add optionName to the findLocale
                        option.nameLocalizations = { ...option.nameLocalizations, [locale]: global.findLocaleString(workingLocale, optionName) };
                        if (option.description)
                            // eslint-disable-next-line max-len
                            option.descriptionLocalizations = { ...option.descriptionLocalizations, [locale]: global.findLocaleString(workingLocale, optionDescription) };
                    }
                }
            }
        }
    }
    return command;
}
