import { execSync } from "node:child_process";
import { createRequire } from "node:module";
import common from "../../../utils/common.js";
import logging from "../../../utils/logging.js";

const require = createRequire(import.meta.url);
const { clientSecret, debugMode } = require(`../../../config.json`);

export default {
    /**
     * @name getCommitInfo
     * @description Gets the commit info for boot messages / version command.
     * @param {*} interactionOrMessage The interaction or message that called this function.
     * @param {*} auditChannel The audit channel to send the message to.
     * @param {boolean} isSlashCommand Whether or not this function was called from a slash command.
     * @param {boolean} botOnline Was this function called on the bot's startup?
     */
    async getCommitInfo(interactionOrMessage, auditChannel, isSlashCommand, botOnline = false)
    {
        const guildId = botOnline ? auditChannel.guild.id : interactionOrMessage.guild.id;

        const { banAppealsChannelID, banAppealsLink } = common.getGuildConfig(guildId);

        var gitMessage = execSync(`git show -s --format="*%s* (**%an**)"`).toString();

        // Don't @everyone/@here
        if (gitMessage.includes(`@everyone`) || gitMessage.includes(`@here`))
            gitMessage = gitMessage.replace(/@/g, `@${ String.fromCodePoint(8203) }`);
        if (gitMessage.length > 1024)
            gitMessage = common.truncateString(gitMessage, 1024);

        // bot online message
        if (botOnline)
        {
            // don't send anything if the bot is in debug mode
            if (debugMode)
                return;

            auditChannel.send(`${ global.locale[global.findLocale(guildId)].botIsOnline } ${ gitMessage }`);
            return;
        }

        const versionString = `**BOT:** ${ gitMessage }`;

        if (!banAppealsChannelID || !clientSecret)
            return common.sendInteractionOrMessage(interactionOrMessage, { content: versionString }, isSlashCommand);

        let banAppealsInfo = await common.fetchJsonResponse(
            `${ banAppealsLink }/commit-info?authKey=${ clientSecret }`,
            {
                methods: `GET`
            }
        ).catch(() => { logging.warn(`Ban Appeals API: Failed to get commit info.`); });
        if (!banAppealsInfo)
        {
            banAppealsInfo = {
                message: `Failed to reach ban appeals API.`
            };
        }

        common.sendInteractionOrMessage(interactionOrMessage, { content: `${ versionString }\n**APPEALS:** ${ banAppealsInfo.message }` }, isSlashCommand);
    }
};
