import SQLite from "better-sqlite3";
import st from "string-table";
import common from "../../../utils/common.js";

const sql = new SQLite(`./src/db.sqlite`);

export default {
    list,
    add,
    remove
};

/**
 * @param {*} interactionOrMessage The interaction or message object.
 * @param {boolean} isSlashCommand Whether or not the interaction is a slash command.
 * @param {string} guildId The guild ID.
 * @param {*} Discord The Discord object.
 */
export async function list(interactionOrMessage, isSlashCommand, guildId, Discord)
{
    const allowedDiscords = sql.prepare(`SELECT * FROM allowedDiscords WHERE guildId = ${ guildId }`).raw().all();
    if (allowedDiscords.length === 0)
    {
        return common.sendInteractionOrMessage(interactionOrMessage, {
            content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].noDiscordsInList
        }, isSlashCommand);
    }

    const servers = [];
    for (const allowedDiscord of allowedDiscords)
    {
        servers.push({
            id: allowedDiscord[1]
        });
    }

    let string = `\`\`\`${ st.create(servers) }\`\`\``;
    if (string.length < 2000)
        return common.sendInteractionOrMessage(interactionOrMessage, { content: string }, isSlashCommand);

    string = `${ st.create(servers) }`;

    // send message as attachment
    const attachment = new Discord.AttachmentBuilder(Buffer.from(string), { name: `allowed-discords.txt` });
    common.sendInteractionOrMessage(interactionOrMessage, { files: [attachment] }, isSlashCommand);
}

/**
 * @param {*} interactionOrMessage The interaction or message object.
 * @param {boolean} isSlashCommand Whether or not the interaction is a slash command.
 * @param {string} guildId The guild ID.
 * @param {string} inviteString The invite string.
 * @param {Discord} Discord The Discord object.
 * @param {Discord.client} client The client.
 */
export async function add(interactionOrMessage, isSlashCommand, guildId, inviteString, Discord, client)
{
    const discordInvite = await client.fetchInvite(inviteString);

    if (!discordInvite)
    {
        return common.sendInteractionOrMessage(interactionOrMessage, {
            content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].invalidDiscordInvite
        }, isSlashCommand);
    }

    // If discord invite is already in db, return
    const allowedDiscords = sql.prepare(`SELECT * FROM allowedDiscords WHERE guildId = ${ guildId }`).raw().all();
    for (const allowedDiscord of allowedDiscords)
    {
        if (guildId === allowedDiscord[1])
        {
            return common.sendInteractionOrMessage(interactionOrMessage, {
                content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].serverAlreadyInList
            }, isSlashCommand);
        }
    }

    const data = {
        guildId,
        id: discordInvite.guild.id
    };

    try
    {
        sql.prepare(`INSERT INTO allowedDiscords (guildId, id) VALUES (@guildId, @id)`).run(data);
    }
    catch (error)
    {
        return console.error(error);
    }

    common.sendInteractionOrMessage(interactionOrMessage, { content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].addedToList.replaceAll(`%item%`, `${ discordInvite.guild.name }`) }, isSlashCommand);
}

/**
 * @param {*} interactionOrMessage The interaction or message object.
 * @param {boolean} isSlashCommand Whether or not the interaction is a slash command.
 * @param {string} guildId The guild ID.
 * @param {string} inviteString The invite string.
 * @param {Discord} Discord The Discord object.
 * @param {Discord.client} client The client.
 */
export async function remove(interactionOrMessage, isSlashCommand, guildId, inviteString, Discord, client)
{
    const regex = /^\d*$/;

    let discordInvite = {
        guild: {
            guildId,
            id: undefined
        }
    };

    if (inviteString.includes(`.`))
    { // Is a discord invite link
        discordInvite = await client.fetchInvite(inviteString);
        if (!discordInvite)
        {
            return common.sendInteractionOrMessage(interactionOrMessage, {
                content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].invalidDiscordInvite
            }, isSlashCommand);
        }

        var guildIdToRm = discordInvite.guild.id;
    }
    else if (regex.test(inviteString))
    { // Is a guildId
        var guildIdToRm = inviteString;
    }

    try
    {
        sql.prepare(`DELETE FROM allowedDiscords WHERE id = ? AND guildId = ?`).run(guildIdToRm, guildId);
        common.sendInteractionOrMessage(interactionOrMessage, { content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].removedFromList.replaceAll(`%item%`, `${ guildIdToRm }`) }, isSlashCommand);
    }
    catch (error)
    {
        console.error(error);
    }
}
