import { AuditLogEvent } from "discord.js";
import sqlite from "../../../utils/sqlite.js";
import common from "../../../utils/common.js";

export default {
    guildBanAddEvent,
    sendBanEmbed,
    banAndSendEmbed,
};

/**
 * @name guildBanAddEvent
 * @description Event that fires when a user is banned from a guild.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client Object.
 * @param {*} data The data object.
 */
async function guildBanAddEvent(Discord, client, data)
{
    const guildId = data.guild.id;
    const guild = client.guilds.cache.get(guildId);

    const user = data.user;
    const fetchedLogs = await guild.fetchAuditLogs({
        limit: 1,
        type: AuditLogEvent.MemberBanAdd
    });
    const banLog = fetchedLogs.entries.first();

    if (!banLog)
        return; // no ban log, something went wrong?

    const { executor, target, action } = banLog;

    if (executor.id === client.user.id)
        return; // dont send embed if bot banned
    if (target.id !== user.id)
        return; // dont send embed if user wasn't banned
    if (action !== AuditLogEvent.MemberBanAdd)
        return; // dont send embed if user wasn't banned

    var reason = banLog.reason;
    if (!reason)
        reason = `${ global.locale[global.findLocale(guildId)].noReasonSpecified }`;

    var newReason = `🔨 ${ reason }`;
    await sqlite.modifyPunishmentsProfile(client, target.id, executor.id, guild.id, `bans`, 1, newReason, `permanently`);

    sendBanEmbed(Discord, client, target.id, target.username, executor.id, reason, guildId);
}

/**
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client Object.
 * @param {string} userId The user ID.
 * @param {string} displayName The user display name.
 * @param {string} moderatorId The moderator ID.
 * @param {string} reason The reason for the ban.
 * @param {Discord.GuildMember} member The guild member object.
 * @name banAndSendEmbed
 * @description Ban the user and send an embed.
 */
function banAndSendEmbed(Discord, client, userId, displayName, moderatorId, reason, member)
{
    const authorTag = client.users.cache.get(moderatorId).tag;
    member.ban({ deleteMessageSeconds: 7 * 24 * 60 * 60, reason: `${ authorTag } | ${ reason }` });
    sendBanEmbed(Discord, client, userId, displayName, moderatorId, reason, member.guild.id);
}

/**
 * @name sendBanEmbed
 * @description Send an embed to the audit channel when a user is banned.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client Object.
 * @param {string} userId The user's id.
 * @param {string} displayName The user's display name.
 * @param {string} moderatorId The moderator's id.
 * @param {string} reason The reason for the ban.
 * @param {string} guildId The guild's id.
 * @param {number} time Time in Epoch format.
 */
function sendBanEmbed(Discord, client, userId, displayName, moderatorId, reason, guildId, time)
{
    const { auditChannelID } = common.getGuildConfig(guildId);
    if (!auditChannelID || auditChannelID.length < 10) return; // no audit channel set, dont try and send embed
    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) // auditChannelID is invalid.
        return;
    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🔨 (${ global.locale[global.findLocale(guildId)].ban })` })
        .addFields([{ name: `${ global.locale[global.findLocale(guildId)].user }`, value: `<@!${ userId }> (${ displayName })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(guildId)].moderator }`, value: `<@!${ moderatorId }>`, inline: true }]);
    if (time && time > 0)
        embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].expires }`, value: `<t:${ Math.round(time / 1000) }:R>`, inline: true }]);
    embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].reason }`, value: (reason || `${ global.locale[global.findLocale(guildId)].noReasonSpecified }`), inline: true }]);

    auditChannel.send({ embeds: [embed] });
}
