import { ButtonStyle } from "discord.js";
import SQLite from "better-sqlite3";
import sqlite from "../../../utils/sqlite.js";
import common from "../../../utils/common.js";

const sql = new SQLite(`./src/db.sqlite`);

export default {
    handleLTAPMute,
    isMuted,
    sendMuteToDB,
    giveMuteRole,
    handleMute,
    checkLTAPStatus,
};

/**
 * @name giveMuteRole
 * @description Gives the mute role to the user.
 * @param {Discord.GuildMember} member The member to mute.
 * @param {Discord.Role} role The role to give the member.
 * @param {string} authorTag The author's tag.
 * @param {string} reason The reason for the mute.
 */
function giveMuteRole(member, role, authorTag, reason)
{
    member.roles.add(role, `${ authorTag } | ${ reason }`);
}

/**
 * @name handleMute
 * @description Sends embed in audit channel and give muted role if they are in guild.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client Object.
 * @param {Discord.GuildMember} member The member to mute.
 * @param {string} authorId The moderator's id.
 * @param {Discord.Role} role The role to give the member.
 * @param {string} authorTag The author's tag.
 * @param {string} reason The reason for the mute.
 * @param {number} time Time in Epoch format.
 * @param {boolean} extending Are we extending a current mute?
 * @param {string} guildId The guild's id.
 * @param {string} targetId The user's id.
 * @param {boolean} addQuickActions Add quick actions (Ban, BanScam, Add Msg Content to Ban List)
 */
function handleMute(Discord, client, member, authorId, role, authorTag, reason, time, extending, guildId, targetId, addQuickActions = false)
{
    const { auditChannelID } = common.getGuildConfig(guildId);

    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🤫 (${ global.locale[global.findLocale(guildId)].mute })` });

    if (member)
    {
        giveMuteRole(member, role, authorTag, reason);
        embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].user }`, value: `<@!${ member.id }> (${ member.user.username })`, inline: true }]);
    }
    else
        embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].user }`, value: `<@!${ targetId }>`, inline: true }]);

    embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].moderator }`, value: `<@!${ authorId }>`, inline: true }]);

    if (time > 0)
    {
        if (extending)
            embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].expires }`, value: `<t:${ Math.round(time / 1000) }:R>`, inline: true }]);

        else
            embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].expires }`, value: `<t:${ Math.round((Date.now() + time) / 1000) }:R>`, inline: true }]);
    }

    embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].reason }`, value: reason, inline: true }]);
    if (!auditChannelID || auditChannelID.length < 10) return; // If there is no audit channel, don't send the embed.
    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) // auditChannelID is invalid.
        return;

    if (addQuickActions)
    {
        const banButton = new Discord.ButtonBuilder()
            .setCustomId(`ban-${ targetId }`)
            .setLabel(`${ global.locale[global.findLocale(guildId)].ban }`)
            .setStyle(ButtonStyle.Danger)
            .setDisabled(false);
        const banScamButton = new Discord.ButtonBuilder()
            .setCustomId(`banScam-${ targetId }`)
            .setLabel(`${ global.locale[global.findLocale(guildId)].banScam }`)
            .setStyle(ButtonStyle.Danger)
            .setDisabled(false);
        const autoModerationAdd = new Discord.ButtonBuilder()
            .setCustomId(`autoModerationAdd-${ targetId }`)
            .setLabel(`${ global.locale[global.findLocale(guildId)].autoModerationAdd }`)
            .setStyle(ButtonStyle.Danger)
            .setDisabled(false);
        const buttonArray = [banButton, banScamButton, autoModerationAdd];
        const row = new Discord.ActionRowBuilder()
            .addComponents(buttonArray);

        auditChannel.send({
            embeds: [embed],
            components: [row]
        });
    }
    else
        auditChannel.send({ embeds: [embed] });
}

/**
 * @name handleLTAPMute
 * @description Re-mutes a user if they leave to avoid punishment.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client Object.
 * @param {Discord.GuildMember} member The member to mute.
 * @param {string} authorId The moderator's id.
 * @param {Discord.Role} role The role to give the member.
 * @param {string} authorTag The author's tag.
 * @param {string} reason The reason for the mute.
 * @param {number} time Time in Epoch format.
 */
function handleLTAPMute(Discord, client, member, authorId, role, authorTag, reason, time)
{
    const guildId = member.guild.id;
    const { auditChannelID } = common.getGuildConfig(guildId);

    if (member.voice) // Boot them from VC
        member.voice.disconnect();

    giveMuteRole(member, role, authorTag, reason);

    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🤫 (${ global.locale[global.findLocale(guildId)].mute })` })
        .addFields([{ name: `${ global.locale[global.findLocale(guildId)].user }`, value: `<@!${ member.id }> (${ member.user.username })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(guildId)].moderator }`, value: `<@!${ authorId }>`, inline: true }]);

    if (time > 0)
        embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].expires }`, value: `<t:${ Math.round((time / 1000)) }:R>`, inline: true }]);

    embed.addFields([{ name: `${ global.locale[global.findLocale(guildId)].reason }`, value: reason, inline: true }]);
    if (!auditChannelID || auditChannelID.length < 10) return; // If there is no audit channel, don't send the embed.
    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) // auditChannelID is invalid.
        return;
    auditChannel.send({ embeds: [embed] });
}

/**
 * @name isMuted
 * @description Checks if a user is muted.
 * @param {*} client The Discord Client Object.
 * @param {string} guildId The guild's id.
 * @param {Discord.GuildMember} member The member to check.
 * @returns {boolean} Is the user muted?
 */
function isMuted(client, guildId, member)
{
    const row = sql.prepare(`SELECT * FROM mutes WHERE guildId = ? AND userId = ?`).get(guildId, member.id);
    if (row)
        return true;
    return false;
}

/**
 * @name sendMuteToDB
 * @description Sends a mute to the database.
 * @param {*} client The Discord Client Object.
 * @param {string} guildId The guild's id.
 * @param {string} authorId The moderator's id.
 * @param {string} id The user's id.
 * @param {number} time Time in Epoch format.
 * @param {string} reason The reason for the mute.
 * @param {string} friendlyTime The time in a more friendly format.
 */
async function sendMuteToDB(client, guildId, authorId, id, time, reason, friendlyTime)
{
    const data = {
        guildAndUserId: `${ guildId }_${ id }`,
        id,
        guildId,
        time,
        moderatorId: authorId
    };
    const a = sql.prepare(`INSERT OR REPLACE INTO mutes (guildAndUserId, guildId, userId, time, moderatorId) VALUES (@guildAndUserId, @guildId, @id, @time, @moderatorId);`);
    a.run(data);

    var newReason = `🤫 ${ reason }`;
    if (time > 0)
        var timeForMuteMeta = `${ global.locale[global.findLocale(guildId)].for } ${ friendlyTime }`;
    else
        var timeForMuteMeta = `${ global.locale[global.findLocale(guildId)].permanently }`;
    await sqlite.modifyPunishmentsProfile(client, id, authorId, guildId, `mutes`, 1, newReason, timeForMuteMeta);
}

/**
 * @name checkLTAPStatus
 * @description Checks if a user left to avoid a punishment.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 * @param {string} guildId The guild ID we're in.
 * @param {Discord.GuildMember} member The member object.
 */
async function checkLTAPStatus(Discord, client, guildId, member)
{
    const sql = new SQLite(`./src/db.sqlite`);
    const { auditChannelID, mutedRoleID } = common.getGuildConfig(guildId);
    if (isMuted(client, member.guild.id, member))
    {
        const row = sql.prepare(`SELECT * FROM mutes WHERE guildId = ? AND userId = ?`).get(member.guild.id, member.id);
        if (!mutedRoleID)
            return;
        const mutedRole = member.guild.roles.cache.find((r) => r.id === mutedRoleID);

        if (member.roles.cache.some((r) => r.id === mutedRole.id))
            return; // User already has the muted role.

        if (row.time === `-1` || row.time === `-1.0`) // Evading Perma Mute, just give them the role and bail out.
            return giveMuteRole(member, mutedRole, client.user.tag, `${ global.locale[global.findLocale(guildId)].evadePermaMute }`);

        // log the user trying to mute evade.
        const guild = client.guilds.cache.get(row.guildId);

        if (auditChannelID && auditChannelID.length > 10)
        {
            const auditChannel = guild.channels.cache.get(auditChannelID);
            if (!auditChannel) // auditChannelID is not a valid channel.
                return;
            auditChannel.send(`${ global.locale[global.findLocale(guildId)].alertBold } <@${ member.user.id }> ${ global.locale[global.findLocale(guildId)].triedToEvade }`);
        }

        // call handle mute & give temp mute role.
        const newTime = Number.parseInt(row.time, 10) + (604_800 * 1000);
        handleLTAPMute(Discord, client, member, client.user.id, mutedRole, member.user.tag, `${ global.locale[global.findLocale(guildId)].mutedForExtraTime }`, newTime);
        await sendMuteToDB(client, guild.id, client.user.id, member.user.id, newTime, `${ global.locale[global.findLocale(guildId)].muteEvasion }`, `1w`);
    }
}
