import common from "../../../utils/common.js";

export default {
    user,
    amount,
    after,
};

/**
 * @name user
 * @description Prune a specific user's messages.
 * @param {*} interactionOrMessage The message object or interaction object.
 * @param {boolean} isSlashCommand Whether or not the command was run from a slash command.
 * @param {string} userId The user's ID.
 * @param {number} amount The amount of messages to delete.
 * @param {string} reason The reason for the prune.
 * @param {string} moderatorId The ID of the moderator who pruned the user.
 * @param {*} client The Discord Client Object.
 * @param {*} Discord The Discord Object.
 */
export async function user(interactionOrMessage, isSlashCommand, userId, amount, reason, moderatorId, client, Discord)
{
    amount = amountCheck(amount);

    const guildId = interactionOrMessage.guild.id;
    const { auditChannelID } = common.getGuildConfig(guildId);

    const member = await common.getGuildMember(client, guildId, userId);

    if (!member)
    {
        return common.sendInteractionOrMessage(interactionOrMessage, {
            content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].cantFetchUser
        }, isSlashCommand);
    }

    interactionOrMessage.channel.messages.fetch({
        limit: amount
    })
        .then((messages) =>
        {
            const targetMessages = [];

            const userMessages = messages.filter((m) =>
            {
                if (m.id === interactionOrMessage.id)
                    return false;
                if (m.author.id === member.id)
                    return m;
                return false;
            });
            for (const message of userMessages)
                targetMessages.push(message[1]);

            interactionOrMessage.channel.bulkDelete(targetMessages)
                .then(() =>
                {
                    if (isSlashCommand)
                        return interactionOrMessage.followUp(`${ global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].purgedMessages }`);
                });
        });

    if (!reason)
        reason = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].noReasonSpecified }`;
    reason = common.truncateString(reason, 256);

    if (!auditChannelID || auditChannelID.length < 10) return; // No audit channel set.
    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) // auditChannelID is invalid.
        return;
    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(interactionOrMessage.guild.id)].purgeSpecificUser })` })
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].target }`, value: `<@!${ member.id }> (${ member.user.username })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].amount }`, value: amount.toString(), inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].channel }`, value: `<#${ interactionOrMessage.channel.id }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].moderator }`, value: `<@!${ moderatorId }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].reason }`, value: reason.toString(), inline: true }]);

    auditChannel.send({ embeds: [embed] });
}

/**
 * @name amount
 * @description Prune a specific amount of messages.
 * @param {*} interactionOrMessage The message object or interaction object.
 * @param {boolean} isSlashCommand Whether or not the command was run from a slash command.
 * @param {number} amount The amount of messages to delete.
 * @param {string} reason The reason for the prune.
 * @param {string} moderatorId The ID of the moderator who pruned.
 * @param {*} client The Discord Client Object.
 * @param {*} Discord The Discord Object.
 */
export async function amount(interactionOrMessage, isSlashCommand, amount, reason, moderatorId, client, Discord)
{
    amount = amountCheck(amount);

    const guildId = interactionOrMessage.guild.id;
    const { auditChannelID } = common.getGuildConfig(guildId);

    interactionOrMessage.channel.bulkDelete(amount, true)
        .then(() =>
        {
            if (isSlashCommand)
                return interactionOrMessage.followUp(`${ global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].purgedMessages }`);
        });

    if (!reason)
        reason = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].noReasonSpecified }`;
    reason = common.truncateString(reason, 256);

    if (!auditChannelID || auditChannelID.length < 10) return; // No audit channel set.
    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) // auditChannelID is invalid.
        return;
    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(interactionOrMessage.guild.id)].prune })` })
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].amount }`, value: amount.toString(), inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].channel }`, value: `<#${ interactionOrMessage.channel.id }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].moderator }`, value: `<@!${ moderatorId }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].reason }`, value: reason.toString(), inline: true }]);

    auditChannel.send({ embeds: [embed] });
}

/**
 * @name after
 * @description Prune messages after a specific message ID.
 * @param {*} interactionOrMessage The message object or interaction object.
 * @param {boolean} isSlashCommand Whether or not the command was run from a slash command.
 * @param {number} messageId The ID of the message to start after from.
 * @param {string} reason The reason for the prune.
 * @param {string} moderatorId The ID of the moderator who pruned.
 * @param {*} client The Discord Client Object.
 * @param {*} Discord The Discord Object.
 */
export async function after(interactionOrMessage, isSlashCommand, messageId, reason, moderatorId, client, Discord)
{
    const guildId = interactionOrMessage.guild.id;
    const { auditChannelID } = common.getGuildConfig(guildId);

    if (!reason)
        reason = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].noReasonSpecified }`;
    reason = common.truncateString(reason, 256);

    interactionOrMessage.channel.messages.fetch({
        limit: 100,
        after: messageId
    })
        .then((messages) =>
        {
            interactionOrMessage.channel.bulkDelete(messages)
                .then(() =>
                {
                    if (isSlashCommand)
                        return interactionOrMessage.followUp(`${ global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].purgedMessages }`);
                });
        });

    if (!auditChannelID || auditChannelID.length < 10) return; // No audit channel set.
    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) // auditChannelID is invalid.
        return;
    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(interactionOrMessage.guild.id)].prune })` })
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].channel }`, value: `<#${ interactionOrMessage.channel.id }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].moderator }`, value: `<@!${ moderatorId }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].reason }`, value: `${ reason }`, inline: true }]);

    auditChannel.send({ embeds: [embed] });
}

/**
 * @name amountCheck
 * @description Check if the amount is valid.
 * @param {number} amount The amount of messages to delete.
 * @returns {number} A number between 2 and 100.
 */
function amountCheck(amount)
{
    if (amount < 2)
        amount = 2;
    if (amount > 100)
        amount = 100;
    return amount;
}
