import { AuditLogEvent } from "discord.js";
import sqlite from "../../../utils/sqlite.js";
import common from "../../../utils/common.js";

import logging from "../../../utils/logging.js";

export default {
    /**
     * @name unbanUser
     * @description Unbans the user and sends an embed.
     * @param {Discord.Message} message The message object.
     * @param {string} id The user's ID.
     * @param {string} authorId The ID of the moderator who unbanned the user.
     * @param {string} authorTag The tag of the moderator who unbanned the user.
     * @param {string} reason The reason for the unban.
     * @param {*} Discord The Discord Object.
     * @param {*} client The Discord Client Object.
     */
    async unbanUser(message, id, authorId, authorTag, reason, Discord, client)
    {
        await message.guild.members.unban(id, `${ authorTag } | ${ reason }`);

        const newReason = `🙏 ${ reason }`;
        const guildId = message.guild.id;
        await sqlite.modifyPunishmentsProfile(client, id, authorId, guildId, `kicks`, 0, newReason); // We hijack kicks here, but don't actually add anything to it. Just to get it into the punishment history.

        const { auditChannelID } = common.getGuildConfig(guildId);

        if (!auditChannelID || auditChannelID.length < 10) return; // No audit channel.
        const auditChannel = client.channels.cache.get(auditChannelID);
        if (!auditChannel) // auditChannelID is invalid.
        {
            logging.warn(`Could not get audit channel from cache with ID ${ auditChannelID }.`);
            return;
        }

        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🙏 (${ global.locale[global.findLocale(guildId)].unban })` })
            .addFields([{ name: `${ global.locale[global.findLocale(guildId)].user }`, value: `<@!${ id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(guildId)].moderator }`, value: `<@!${ authorId }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(guildId)].reason }`, value: reason, inline: true }]);

        auditChannel.send({ embeds: [embed] });
    },

    /**
     * @name guildBanRemoveEvent
     * @description Handles the guildBanRemove event.
     * @param {*} Discord The Discord Object.
     * @param {*} client The Discord Client Object.
     * @param {*} data The data object.
     */ async guildBanRemoveEvent(Discord, client, data)
    {
        const guildId = data.guild.id;
        const guild = client.guilds.cache.get(guildId);
        const { auditChannelID } = common.getGuildConfig(guildId);
        const user = data.user;
        const fetchedLogs = await guild.fetchAuditLogs({
            limit: 1,
            type: AuditLogEvent.MemberBanRemove
        });
        const banLog = fetchedLogs.entries.first();

        if (!auditChannelID || auditChannelID.length < 10) return; // No audit channel.
        const auditChannel = client.channels.cache.get(auditChannelID);
        if (!auditChannel) // auditChannelID is invalid.
            return;
        if (!banLog)
            return auditChannel.send(`${ user.tag } ${ global.locale[global.findLocale(guildId)].unknownUnbanExecutor }`);

        const { executor, target, action } = banLog;

        if (executor.id === client.user.id)
            return; // Don't send a message if the bot unbanned them.
        if (target.id !== user.id)
            return; // Don't send a message if the user was not banned.
        if (action !== AuditLogEvent.MemberBanRemove)
            return; // Don't send a message if the ban was not removed.

        var reason = banLog.reason;
        if (!reason)
            reason = `${ global.locale[global.findLocale(guildId)].manualUnban }`;

        var newReason = `🙏 ${ reason }`;
        await sqlite.modifyPunishmentsProfile(client, target.id, executor.id, guildId, `kicks`, 0, newReason); // We hijack kicks here, but don't actually add anything to it. Just to get it into the punishment history.

        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🙏 (${ global.locale[global.findLocale(guildId)].unban })` })
            .addFields([{ name: `${ global.locale[global.findLocale(guildId)].user }`, value: `<@!${ target.id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(guildId)].moderator }`, value: `<@!${ executor.id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(guildId)].reason }`, value: reason, inline: true }]);

        auditChannel.send({ embeds: [embed] });
    }
};
