import SQLite from "better-sqlite3";
import common from "../../../utils/common.js";
import sqlite from "../../../utils/sqlite.js";

const sql = new SQLite(`./src/db.sqlite`);

export default {
    /**
     * @param {*} interactionOrMessage The interaction or message object.
     * @param {boolean} isSlashCommand Whether or not the interaction is a slash command.
     * @param {string} targetId The ID of the user to unmute.
     * @param {string} moderatorId The ID of the moderator who unmuted the user.
     * @param {string} reason The reason for the unmute.
     * @param {*} client The Discord Client Object.
     * @param {*} Discord The Discord Object.
     */
    async unmute(interactionOrMessage, isSlashCommand, targetId, moderatorId, reason, client, Discord)
    {
        const guildId = interactionOrMessage.guild.id;
        const { auditChannelID, mutedRoleID } = common.getGuildConfig(guildId);
        const mutedRole = interactionOrMessage.guild.roles.cache.get(mutedRoleID);

        const guild = client.guilds.cache.get(interactionOrMessage.guild.id);
        const member = await common.getGuildMember(client, interactionOrMessage.guild.id, targetId);

        if (!reason)
            reason = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].noReasonSpecified }`;
        reason = common.truncateString(reason, 256);
        const moderator = interactionOrMessage.guild.members.cache.get(moderatorId);
        const authorTag = moderator.user.tag;

        if (member)
        {
            if (!member.roles.cache.some((r) => r.id === mutedRoleID))
            {
                return common.sendInteractionOrMessage(interactionOrMessage, {
                    content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].userNotMuted
                }, isSlashCommand);
            }

            if ((member.roles.highest.position >= moderator.roles.highest.position) && guild.ownerId !== moderatorId)
            {
                return common.sendInteractionOrMessage(interactionOrMessage, {
                    content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].cantUnmuteUser
                }, isSlashCommand);
            }

            member.roles.remove(mutedRole, `${ authorTag } | ${ reason }`);
            const row = sql.prepare(`SELECT * FROM mutes WHERE guildId = ? AND userId = ?`).get(guildId, member.id);
            if (row)
                sql.prepare(`DELETE FROM mutes WHERE guildId = ? AND userId = ?`).run(guildId, member.id);
        }
        else
        {
            // if they aren't in guild, try and remove their ID from the db anyways
            const databaseMuted = sql.prepare(`SELECT * FROM mutes WHERE guildId = ? AND userId = ?`).get(guildId, targetId);
            if (databaseMuted)
                sql.prepare(`DELETE FROM mutes WHERE guildId = ? AND userId = ?`).run(guildId, targetId);
        }

        var newReason = `🗣️ ${ reason }`;
        await sqlite.modifyPunishmentsProfile(client, targetId, moderatorId, guildId, `kicks`, 0, newReason); // We hijack kicks here, but don't actually add anything to it. Just to get it into the punishment history.
        await this.embed(auditChannelID, member, moderatorId, reason, targetId, client, Discord);
    },

    /**
     * @param {string} auditChannelID The ID of the audit channel.
     * @param {Discord.GuildMember} member The member object.
     * @param {string} moderatorId The ID of the moderator who unmuted the user.
     * @param {string} reason The reason for the unmute.
     * @param {string} targetId ID of victim.
     * @param {Discord.client} client The Discord Client Object.
     * @param {Discord} Discord The Discord Object.
     */
    async embed(auditChannelID, member, moderatorId, reason, targetId, client, Discord)
    {
        if (!auditChannelID || auditChannelID.length < 10) return; // no audit channel set
        const auditChannel = client.channels.cache.get(auditChannelID);
        if (!auditChannel) // auditChannelID is invalid.
            return;
        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🗣️ (${ global.locale[global.findLocale(member.guild.id)].unmute })` });
        if (member)
            embed.addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].user }`, value: `<@!${ member.id }> (${ member.user.username })`, inline: true }]);
        else
            embed.addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].user }`, value: `<@!${ targetId }>`, inline: true }]);

        embed.addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].moderator }`, value: `<@!${ moderatorId }>`, inline: true }]);
        embed.addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].reason }`, value: reason, inline: true }]);

        // look up latest mute for the user in the db
        const muteData = sql.prepare(`SELECT punishmentreason FROM punishments WHERE guildId = ? AND id = ?`).get(auditChannel.guild.id, targetId);
        if (muteData)
        {
            // search from the end of the array for "🤫"
            let reason = muteData.punishmentreason.split(`🤫`).pop();
            // strip """]" from the end of the string
            reason = reason.slice(0, Math.max(0, reason.length - 2));
            // If there's "," remove it and everything after it
            reason = reason.split(`",`).shift();
            embed.addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].reasonForMute }`, value: reason, inline: true }]);
        }

        auditChannel.send({ embeds: [embed] });
    }
};
