import sqlite from "../../../utils/sqlite.js";
import mute from "./mute.js";
import common from "../../../utils/common.js";

export default {
    /**
     * @param {*} interactionOrMessage The interaction or message object. * @param isAutomatic
     * @param {boolean} isAutomatic Whether or not the warn was automatic.
     * @param {boolean} isSlashCommand Whether or not the interaction is a slash command.
     * @param {string} targetId The ID of the user to warn.
     * @param {string} moderatorId The ID of the moderator who warned the user.
     * @param {string} reason The reason for the warn.
     * @param {string} messageContent The message content.
     * @param {*} Discord The Discord Object.
     * @param {*} client The Discord Client Object.
     */
    async warnUser(interactionOrMessage, isAutomatic, isSlashCommand, targetId, moderatorId, reason, messageContent, Discord, client)
    {
        const guildId = interactionOrMessage.guild.id;
        const { auditChannelID, mutedRoleID, generalChannelID } = common.getGuildConfig(guildId);

        const guild = client.guilds.cache.get(interactionOrMessage.guild.id);
        const member = await common.getGuildMember(client, interactionOrMessage.guild.id, targetId);

        const moderator = await common.getGuildMember(client, interactionOrMessage.guild.id, moderatorId);

        if (!isAutomatic)
        {
            if (!member)
            {
                return common.sendInteractionOrMessage(interactionOrMessage, {
                    content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].cantFetchUser
                }, isSlashCommand);
            }

            if (member.roles.highest.position >= moderator.roles.highest.position && guild.ownerId !== moderatorId)
            {
                return common.sendInteractionOrMessage(interactionOrMessage, {
                    content: global.locale[global.findLocale(interactionOrMessage.guild.id, interactionOrMessage.locale)].youCannotWarnThisUser
                }, isSlashCommand);
            }
        }

        if ((auditChannelID) && auditChannelID.length > 10)
        {
            const auditChannel = client.channels.cache.get(auditChannelID);
            if (!auditChannel) // auditChannelID is invalid.
                return;
            const embed = new Discord.EmbedBuilder()
                .setColor(`#ff1e00`)
                .setTimestamp()
                .setAuthor({ name: `⚠️ (${ global.locale[global.findLocale(interactionOrMessage.guild.id)].warn })` })
                .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].user }`, value: `<@!${ member.id }> (${ member.user.username })`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].moderator }`, value: `<@!${ moderator.id }>`, inline: true }]);

            if (messageContent)
            {
                messageContent = common.truncateString(messageContent, 1024);
                embed.addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].reason }`, value: `${ reason }\n${ global.locale[global.findLocale(interactionOrMessage.guild.id)].message }: \`\`\`${ messageContent }\`\`\``, inline: true }]);
            }
            else
                embed.addFields([{ name: `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].reason }`, value: `${ reason }`, inline: true }]);

            auditChannel.send({ embeds: [embed] });
        }

        var newReason = `⚠️ ${ reason }`;

        await sqlite.modifyPunishmentsProfile(client, targetId, moderator.id, interactionOrMessage.guild.id, `warns`, 1, newReason);

        // handle auto-muting.
        var warnings = sqlite.getPunishmentData(interactionOrMessage.guild.id, targetId, `warns`);
        var message = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].warnBody1 } \`${ guild.name }\` ${ global.locale[global.findLocale(interactionOrMessage.guild.id)].warnBody2 } \`${ reason }\`.`;

        if (warnings === 1)
            message += `\n${ global.locale[global.findLocale(interactionOrMessage.guild.id)].onlyFirstWarning }`;

        if (mutedRoleID && mutedRoleID.length > 10)
            var role = member.guild.roles.cache.find((r) => r.id === mutedRoleID);

        if (warnings === 2 && role)
        {
            var time = 86_400_000;
            var friendlyTime = `24h`;
            var muteReason = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].tempMute2ndWarn }`;
            message += `\n${ global.locale[global.findLocale(interactionOrMessage.guild.id)].tempMuteWhenExpire } <t:${ Math.round((Date.now() + time) / 1000) }:R>`;
        }
        else if (warnings > 2 && role)
        {
            var time = 604_800_000;
            var friendlyTime = `7d`;
            var muteReason = `${ global.locale[global.findLocale(interactionOrMessage.guild.id)].tempMute2Plus }`;
            message += `\n${ global.locale[global.findLocale(interactionOrMessage.guild.id)].tempMuteWhenExpire } <t:${ Math.round((Date.now() + time) / 1000) }:R>`;
        }

        if (warnings > 1)
        {
            if (!role)
                return;
            mute.handleMute(Discord, client, member, moderator.id, role, moderator.tag, muteReason, time, false, member.guild.id, member.user.id);
            await mute.sendMuteToDB(client, interactionOrMessage.guild.id, moderator.id, targetId, Date.now() + time, muteReason, friendlyTime);
        }

        await client.users.cache.get(member.id).send(message)
            .catch(() =>
            {
                // cannot dm, attempt to send in chat.
                if (generalChannelID && generalChannelID.length >= 15)
                {
                    const generalChannel = client.channels.cache.get(generalChannelID);
                    if (generalChannel)
                        generalChannel.send(`<@${ member.user.id }>, ${ message }`);
                }
            });
    }
};

