import { ApplicationCommandOptionType } from "discord.js";
import SQLite from "better-sqlite3";
import fs from "node:fs";
import fetch from "node-fetch";
import common from "../../../../utils/common.js";

const sql = new SQLite(`./src/db.sqlite`);

export default {
    name: `configure`,
    description: `Configure the bot.`,
    type: ApplicationCommandOptionType.Subcommand,
    options: [
        {
            name: `type`,
            description: `What do you want to do?`,
            type: ApplicationCommandOptionType.String,
            choices: [
                {
                    name: `Import Config`,
                    value: `import`
                },
                {
                    name: `Export Config as Template (No Server Data)`,
                    value: `export-template`
                },
                {
                    name: `Export All Server Data and Configs`,
                    value: `export`
                },
            ],
            required: true
        },
        {
            name: `file`,
            description: `The file to import.`,
            type: ApplicationCommandOptionType.Attachment,
            required: false
        },
    ],

    async callback(interaction, client, Discord)
    {
        const type = interaction.options.getString(`type`);
        if (type === `import`)
            return importConfig(interaction);
        if (type === `export`)
            return exportConfig(interaction, false, Discord);
        if (type === `export-template`)
            return exportConfig(interaction, true, Discord);
    }
};

/**
 * @name importConfig
 * @description Imports a configuration file.
 * @param {*} interaction The interaction object.
 */
async function importConfig(interaction)
{
    var attachment = interaction.options.getAttachment(`file`);
    if (!attachment)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].uploadAFile }`);

    var filePath = `./${ interaction.guild.id }_import.json`;
    // Write the file contents from the URL to a file
    const file = fs.createWriteStream(filePath);
    const response = await fetch(attachment.url);
    const arrayBuffer = await response.arrayBuffer();
    const buffer = Buffer.from(arrayBuffer);
    file.write(buffer);
    file.end();
    // wait for file to be written
    await common.sleep(1000);
    var newGuildId = interaction.guild.id;
    var fileContents = fs.readFileSync(filePath, `utf8`);
    // read the first line
    const firstLine = fileContents.split(`\n`)[0];
    if (firstLine.startsWith(`//`))
    {
        var oldGuildId = firstLine.replace(`// `, ``);
        fileContents = fileContents.replace(firstLine, ``);
    }
    else
        return interaction.channel.send(`<@!${ interaction.author.id }>, ${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].uploadAFile }`);

    fileContents = fileContents.split(oldGuildId).join(newGuildId); // This is rough...
    // Import the data
    const guildConfigs = JSON.parse(fileContents);
    // take this json data and insert it into the database
    for (const guildConfig of guildConfigs)
    {
        if (guildConfig !== undefined && guildConfig !== null)
        {
            const tableName = guildConfig.tableName;
            const rows = guildConfig.rows;
            for (const row of rows)
            {
                const values = Object.values(row);
                const keys = Object.keys(row);

                const newValues = values.map((value) =>
                {
                    if (typeof (value) === `string`)
                        value = value.replace(/'/g, `''`);
                    return `'${ value }'`;
                }).join(`, `);

                const query = `INSERT OR REPLACE INTO ${ tableName } (${ keys.join(`, `) }) VALUES (${ newValues });`;
                sql.prepare(query).run();
            }
        }
    }

    // delete file
    fs.unlinkSync(filePath);

    // Tell the user it's done
    return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].importedGuildData }`);
}

/**
 * @name exportConfig
 * @description Exports a configuration file.
 * @param {*} interaction The interaction object.
 * @param {boolean} isTemplate Whether or not the export is a template.
 * @param {*} Discord The Discord client.
 */
async function exportConfig(interaction, isTemplate, Discord)
{
    const guildId = interaction.guild.id;
    // Make array of every table in the database
    const tables = sql.prepare(`SELECT name FROM sqlite_master WHERE type='table'`).all();
    const tableNames = tables.map((table) => table.name);

    if (isTemplate)
    {
        tableNames.splice(tableNames.indexOf(`banAppealTokens`), 1);
        tableNames.splice(tableNames.indexOf(`bannedServers`), 1);
        tableNames.splice(tableNames.indexOf(`configuration`), 1);
        tableNames.splice(tableNames.indexOf(`giveawayEntries`), 1);
        tableNames.splice(tableNames.indexOf(`giveaways`), 1);
        tableNames.splice(tableNames.indexOf(`keywordSpam`), 1);
        tableNames.splice(tableNames.indexOf(`memberCounts`), 1);
        tableNames.splice(tableNames.indexOf(`mutes`), 1);
        tableNames.splice(tableNames.indexOf(`punishments`), 1);
        tableNames.splice(tableNames.indexOf(`rolePersists`), 1);
        tableNames.splice(tableNames.indexOf(`scheduledMessages`), 1);
        tableNames.splice(tableNames.indexOf(`tempbanUsers`), 1);
        tableNames.splice(tableNames.indexOf(`userNotes`), 1);
    }

    // For each table, get all of the rows that have guildId in the guildId column
    const guildConfigs = await Promise.all(tableNames.map(async (tableName) =>
    {
        // do we have a guildId column?
        const columns = sql.prepare(`PRAGMA table_info(${ tableName })`).all();
        const columnNames = columns.map((column) => column.name);
        if (!columnNames.includes(`guildId`))
            return;
        const rows = sql.prepare(`SELECT * FROM ${ tableName } WHERE guildId = ?`).all(guildId);
        if (!rows) return;
        // If we exported the guild's configuration, we need to remove some values.
        if (tableName === `configuration`)
        {
            for (const row of rows)
            {
                // json parse the config
                row.config = JSON.parse(row.config);
                // redact api key
                if (row.config.ITADAPIKey)
                    row.config.ITADAPIKey = ``;
                if (row.config.DeepLAPIKey)
                    row.config.DeepLAPIKey = ``;
                row.config = JSON.stringify(row.config);
            }
        }
        return {
            tableName,
            rows
        };
    }));

    // Attempt to remove any undefined values / empty rows (Sometimes this misses?)
    for (const guildConfig of guildConfigs)
    {
        if (guildConfig === undefined || guildConfig === null || guildConfig.rows.length === 0) // Remove empty rows
            guildConfigs.splice(guildConfigs.indexOf(guildConfig), 1);
    }

    // Write to file
    const fileName = `${ interaction.guild.id }_export.json`;
    const filePath = `./${ fileName }`;
    const file = fs.createWriteStream(filePath);
    // Write a "// guildId" comment to the file
    file.write(`// ${ interaction.guild.id }\n`);
    file.write(JSON.stringify(guildConfigs, undefined, 4));
    file.end();
    // wait for file to be written
    await common.sleep(1000);
    // Upload file to discord
    const attachment = await new Discord.AttachmentBuilder(filePath);
    await interaction.followUp({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].exportedGuildData }`, files: [attachment] });
    // Delete file
    fs.unlinkSync(filePath);
}
