import SQLite from "better-sqlite3";
import { ApplicationCommandOptionType } from "discord.js";
import sqlite from "../../../utils/sqlite.js";
import logging from "../../../utils/logging.js";
import common from "../../../utils/common.js";
import ban from "../../commands/moderation/ban.js";

const sql = new SQLite(`./src/db.sqlite`);

export default {
    name: `ban`,
    description: `Ban a user from this server`,
    defaultMemberPermissions: [`BanMembers`],
    options: [
        {
            name: `user`,
            description: `Which user to ban. NOTE: This can also be an ID`,
            type: ApplicationCommandOptionType.User,
            required: true
        },
        {
            name: `duration`,
            description: `Duration? (can be 1d or 1h30m. valid times: y, mo, w, d, h, m, s)`,
            type: ApplicationCommandOptionType.String,
            required: false
        },
        {
            name: `reason`,
            description: `Which reason`,
            type: ApplicationCommandOptionType.String,
            required: false
        },
        {
            name: `silent`,
            description: `Do NOT send a message to the user`,
            type: ApplicationCommandOptionType.Boolean,
            required: false
        }
    ],

    async execute(interaction, client, Discord)
    {
        await interaction.deferReply({ ephemeral: true });

        const guildId = interaction.guild.id;
        const { sendDmOnBan, banAppealsLink } = common.getGuildConfig(guildId);
        const user = interaction.options.getUser(`user`);
        const duration = interaction.options.getString(`duration`);

        const guild = client.guilds.cache.get(interaction.guild.id);
        const member = await common.getGuildMember(client, interaction.guild.id, user.id);

        let reason = interaction.options.getString(`reason`);
        if (!reason)
            reason = `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noReasonSpecified }`;
        reason = common.truncateString(reason, 256);
        const authorTag = interaction.user.tag;

        var isSilent = false;
        if (interaction.options.getBoolean(`silent`) === true)
            isSilent = true; // Do not send DM upon ban.

        if (duration)
        { // Did they specify a duration?
            var time = await common.getEpoch(duration, false);
            if (time === -1337)
                return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].badTimeUnit }`);

            const row = sql.prepare(`SELECT * FROM tempbanUsers WHERE guildId = ? AND userId = ?`).get(interaction.guild.id, user.id);
            if (row && row.time > 1)
            { // If they are already banned, and have a punishment, take their current ban time and add the new time to it.
                time = Number.parseInt(row.time, 10) + time;
                sql.prepare(`UPDATE tempbanUsers SET time = ?, SET moderatorId = ? WHERE guildId = ? AND userId = ?`).run(time, interaction.user.id, interaction.guild.id, user.id);
            }
            else if (row && row.time < 1)
                return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].permBanned }`);
            else
            { // fresh temp ban
                time += Date.now();
                sql.prepare(`INSERT INTO tempbanUsers VALUES (?, ?, ?, ?, ?)`).run(`${ interaction.guild.id }_${ user.id }`, interaction.guild.id, user.id, time, interaction.user.id);
            }
        }

        if (time > 0)
            var timeForMuteMeta = `${ global.locale[global.findLocale(interaction.guild.id)].for } ${ duration }`;
        else
            var timeForMuteMeta = `${ global.locale[global.findLocale(interaction.guild.id)].permanently }`;

        if (!member)
        {
            // prevent double banning
            let isBanned = true;
            guild.bans.fetch(user.id).catch(() =>
            {
                isBanned = false;
            });
            if (isBanned && time < 0)
                return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].alreadyBanned }`);
            if (isBanned && time > 0)
                return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].banTimeUpdate.replaceAll(`%ourArguments%`, `${ duration }`) }`);

            // They aren't in the server, try anyways
            await interaction.guild.members.ban(user.id, { deleteMessageSeconds: 7 * 24 * 60 * 60, reason: `${ authorTag } | ${ reason }` }).then(async () =>
            {
                ban.sendBanEmbed(Discord, client, user.id, user.username, interaction.user.id, reason, guildId, time);
                var newReason = `🔨 ${ reason }`;
                await sqlite.modifyPunishmentsProfile(client, user.id, interaction.user.id, interaction.guild.id, `bans`, 1, newReason, timeForMuteMeta);
                interaction.followUp(`${ user.tag } ${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].hasBeenBanned }`);
            }).catch((error) =>
            {
                logging.error(error);
                interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].invalidFormat }`).then((message) =>
                {
                    setTimeout(() =>
                    {
                        message.delete();
                    }, 10_000);
                });
            });
            return; // we banned someone not in the server, so we're done here
        }

        if ((member.roles.highest.position >= interaction.member.roles.highest.position && guild.ownerId !== interaction.user.id) || !member.bannable)
            return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].couldntBanUser }`);

        if (sendDmOnBan && banAppealsLink && banAppealsLink.length > 1 && isSilent === false)
        {
            const nitroSafety = `${ global.locale[global.findLocale(interaction.guild.id)].nitroSafety.replaceAll(`%guildName%`, `${ guild.name }`).replaceAll(`%banAppealsLink%`, `${ banAppealsLink }`) }`;
            const standardMessage = `${ global.locale[global.findLocale(interaction.guild.id)].standardBanDM.replaceAll(`%guildName%`, `${ guild.name }`).replaceAll(`%banAppealsLink%`, `${ banAppealsLink }`) }`;
            await (reason.toLowerCase().includes(`nitro`) || reason.toLowerCase().includes(`scam`) ? member.send(nitroSafety).catch(() => logging.warn(`Could not DM, ignoring`)) : member.send(standardMessage).catch(() => logging.warn(`Could not DM, ignoring`)));
        }

        member.ban({ deleteMessageSeconds: 7 * 24 * 60 * 60, reason: `${ authorTag } | ${ reason }` })
            .then(interaction.followUp(`${ user.tag } ${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].hasBeenBanned }`));
        ban.sendBanEmbed(Discord, client, member.id, member.user.username, interaction.user.id, reason, guildId, time);
        var newReason = `🔨 ${ reason }`;
        await sqlite.modifyPunishmentsProfile(client, user.id, interaction.user.id, interaction.guild.id, `bans`, 1, newReason, timeForMuteMeta);
    }
};
