import SQLite from "better-sqlite3";
import { createRequire } from "node:module";
import { ChannelType } from "discord.js";
import logging from "../utils/logging.js";
import common from "../utils/common.js";
import { invites } from "./ready.js";

const sql = new SQLite(`./src/db.sqlite`);

const require = createRequire(import.meta.url);
const { webfrontURL } = require(`../config.json`);

const defaultConfig = {
    // General Settings
    locale: `en`,
    prefix: [`cdb!`],

    // Role/Channel IDs
    vipRoleID: ``,
    auditChannelID: ``,
    deletedMessagesChanID: ``,
    mutedRoleID: ``,
    noPingRoleId: ``,
    generalChannelID: ``,
    suggestionChannelID: ``,
    regionalChannelID: ``,
    lockdownChannelIDs: [],
    banAppealsChannelID: ``,
    autoJoinRoleID: ``,

    // Toggleable Features
    allowAnyoneWithRoleToSendLinks: false,
    allowAnyoneWithRoleToBypassBadWordBanning: false,
    allowAnyoneWithRoleToBypassTriggerResponse: false,
    renameUntaggableNames: false,
    renameTo: `Unknown Soldier`,
    allowAnyoneToSendStickers: false,
    logSelfDeletes: false,
    logSelfEdits: false,
    trackInvites: false,
    reportApplet: false,
    sendDmOnBan: false,
    banAppealsLink: ``,
    altDetectionDays: -1,

    // AutoMod Settings
    autoModCaps: false,
    allowDenyMode: `allow`,
    autoModMode: `warn`,
    autoBanMode: `all`,
    autoWarnForAmountOfMentions: 4,
    amountOfSpoilers: -1,

    // Welcome Settings
    welcomeChannelID: ``,
    welcomeMessage: ``,
    leavingMessage: ``,

    // Keyword Spam
    amountOfKeywordSpam: -1,
    timeToMuteForKeywordSpamInMinutes: -1,
    timeBeforeRefreshingKeywordSpamInMinutes: -1,

    // Ticket Settings
    bannedfromTicketsRoleId: ``,
    ticketAdminRoleId: ``,
    openTixCatId: ``,
    closedTixCatId: ``,

    // API Keys
    ITADAPIKey: ``,
    DeepLAPIKey: ``,

    // Server Stat Categories
    serverStatsCatName: `📊 Server Stats 📊`,
    serverStatsMemberChanName: `Members: `,
    serverStatsBoosterChanName: `Boosters: `,
    serverStatsChanId: ``,
    memberStatsChanId: ``,
    boostersStatsChanId: ``
};

/**
 * @name guildCreate
 * @description Event that fires when the bot is invited to a guild while running, or on every fresh boot.
 */
export default {
    name: `guildCreate`,
    freshBootCheck,
    /**
     * @param {*} guild A valid guild object.
     * @param {*} Discord The Discord Object.
     */
    async execute(guild, Discord)
    { // This is called when we invite the bot to a guild while it is running **we already get a guildConfig from our raw packet call**.
        // (So we can send a welcome message.)
        const guildId = guild.id;

        // Check if guild is banned
        const inDatabase = sql.prepare(`SELECT * FROM bannedServers WHERE guildId = ?;`).get(guildId);
        if (inDatabase)
            return guild.leave();

        let guildConfig = global.getConfiguration.get(guildId);
        if (!guildConfig)
        { // For some reason, if we don't have a guildConfig (maybe we beat the raw packet?) make one.
            logging.log(`${ global.locale[global.gLocale].creatingNewGC } ${ guildId }`);
            guildConfig = {
                guildId,
                config: JSON.stringify(defaultConfig)
            };
            global.setConfiguration.run(guildConfig);
        }

        await common.deploySlashCommands(guild); // deploy slash commands to a newly joined guild (while the bot is running)
        await common.sleep(1500);
        // Attempt to create a logs channel.
        let logs = true;
        const logsChannel = await guild.channels.create({
            name: `logs`,
            type: ChannelType.GuildText,
            permissionOverwrites: [
                { // Deny @everyone from viewing the logs channel.
                    id: guild.id,
                    deny: [`ViewChannel`],
                }
            ],
        }).catch(() =>
        { // Unable to create logs channel.
            logs = false;
        });

        const welcomeTitle = `${ global.locale[global.findLocale(guild.id)].newGuildMessageTitle }`;
        const welcomeMessage = `${ global.locale[global.findLocale(guild.id)].newGuildMessageBody.replaceAll(`%webfrontURL%`, `${ webfrontURL }`).replaceAll(`%guildId%`, `${ guildId }`) }`;
        const welcomeMessage2 = `${ global.locale[global.findLocale(guild.id)].newGuildMessageBody2 }`;

        if (logs)
        {
            // Set this channel as the logs channel.
            const auditChannelID = logsChannel.id;
            const GUILD_CONFIG = JSON.parse(guildConfig.config);
            GUILD_CONFIG.auditChannelID = auditChannelID;
            const guildConfigString = JSON.stringify(GUILD_CONFIG);
            const data = {
                guildId,
                config: guildConfigString
            };
            global.setConfiguration.run(data);

            // Send message in that channel.
            const embed = new Discord.EmbedBuilder()
                .setColor(`BLUE`)
                .setTitle(welcomeTitle)
                .setDescription(
                    `${ welcomeMessage }.\n\n${ welcomeMessage2 }`
                )
                .setTimestamp();
            logsChannel.send({ embeds: [embed] });
        }
        else
        { // Unable to create logs channel, use random channel.
            // Find a channel we can view & type in.
            let channelToSend;
            for (const channel of guild.channels.cache.values())
            {
                if (
                    channel.type === ChannelType.GuildText
                    && !channelToSend
                    && channel.permissionsFor(guild.me).has(`SendMessages`)
                )
                    channelToSend = channel;
            }
            if (!channelToSend) return;
            const embed = new Discord.EmbedBuilder()
                .setColor(`BLUE`)
                .setTitle(welcomeTitle)
                .setDescription(
                    `${ welcomeMessage } ${ global.locale[global.findLocale(guild.id)].noAuditChanMessageBody }\n\n${ welcomeMessage2 }`
                )
                .setTimestamp();
            channelToSend.send({ embeds: [embed] });
        }

        guild.invites.fetch().then((guildInvites) =>
        {
            // Adding any new invites to the Map.
            invites.set(guild.id, new Map(guildInvites.map((invite) => [invite.code, invite.uses])));
        });
    },
};

/**
 * @name freshBootCheck
 * @description This is called on fresh boots & if you invite while running, so we setup/update the guild's config here.
 * @param {*} guild An incomplete guild object. Only useful for the guild id.
 */
export async function freshBootCheck(guild)
{
    const guildId = guild.id;
    let guildConfig = global.getConfiguration.get(guildId);
    if (!guildConfig)
    { // Create a guild config if it doesn't exist. (This only runs on fresh boots, so if someone invited the bot while it was offline.)
        logging.log(`${ global.locale[global.gLocale].creatingNewGC } ${ guildId }`);
        guildConfig = {
            guildId,
            config: JSON.stringify(defaultConfig)
        };
        global.setConfiguration.run(guildConfig);
        return;
    }
    guildConfig = JSON.parse(guildConfig.config);
    for (const key in defaultConfig)
    { // If we already have a config, check and see if everything from defaultConfig is set.
        if (!(key in guildConfig))
            guildConfig[key] = defaultConfig[key];
    }
    for (const key in guildConfig)
    { // Clean Legacy Config Options. (If something in guildConfig is not in defaultConfig, remove it from guildConfig).
        if (!(key in defaultConfig))
            delete guildConfig[key];
    }
    global.setConfiguration.run({
        guildId,
        config: JSON.stringify(guildConfig)
    });
    logging.log(`${ global.locale[global.gLocale].updatingGC } ${ guildId }`);
}
