import SQLite from "better-sqlite3";
import common from "../utils/common.js";
import mute from "../commands/commands/moderation/mute.js";
import { invites } from "./ready.js";

import logging from "../utils/logging.js";

const sql = new SQLite(`./src/db.sqlite`);

/**
 * @name guildMemberAdd
 * @description Event that fires when a user joins a guild.
 */
export default {
    name: `guildMemberAdd`,
    /**
     * @param {Discord.GuildMember} member A valid member object.
     * @param {*} Discord The Discord Object.
     * @param {*} client The client object.
     */
    async execute(member, Discord, client)
    {
        if (member.user.bot) return;
        const guildId = member.guild.id;
        const { auditChannelID, trackInvites } = common.getGuildConfig(guildId);
        if (!auditChannelID || auditChannelID.length < 10) return;
        const auditChannel = client.channels.cache.get(auditChannelID);
        if (!auditChannel) // auditChannelID is not a valid channel.
            return;

        await kickForBannedName(member, auditChannel); // Kick the user if they have a banned name.
        await common.checkName(Discord, client, member); // Check name for non-ASCII up here since we return later on.
        await common.checkAlt(member, Discord, client); // Check for alt accounts here.
        await mute.checkLTAPStatus(Discord, client, member.guild.id, member); // Check if the user is LTAP'ing.
        await checkInvite(member, auditChannel, trackInvites, client); // Check who invited this user.
        await checkPersistedRoles(member); // Check if the user has any persisted roles.
        await welcomeMember(member, client); // Send a welcome message.
    },
};

/**
 * @name kickForBannedName
 * @description Kick a user for using a banned name.
 * @param {Discord.GuildMember} member A valid member object.
 * @param {Discord.TextChannel} auditChannel A valid audit channel object.
 */
async function kickForBannedName(member, auditChannel)
{
    const badNamesH = sql.prepare(`SELECT name FROM badNames WHERE guildId = ${ member.guild.id }`).raw().all();
    const badNames = badNamesH.flat();
    if (badNames.length > 0)
    {
        // If their username contains any one of the items in badNames, kick them.
        for (const name of badNames)
        {
            if (member.user.username.toLowerCase().includes(name.toLowerCase()))
            {
                await member.send(`${ global.locale[global.findLocale(member.guild.id)].kickedForBadName.replaceAll(`%name%`, `${ name }`) }`).catch(() => logging.warn(`Could not DM, ignoring`));
                await member.kick(`${ global.locale[global.findLocale(member.guild.id)].kickedForBadNameAuditReason }`);
                const message = `${ global.locale[global.findLocale(member.guild.id)].alertBold } <@!${ member.user.id }> (${ member.user.username }) ${ global.locale[global.findLocale(member.guild.id)].hasBeenKickedForBadName }`;
                return auditChannel.send(message);
            }
        }
    }
}

/**
 * @name checkInvite
 * @description Check who invited this user.
 * @param {Discord.GuildMember} member A valid member object.
 * @param {Discord.TextChannel} auditChannel A valid audit channel object.
 * @param {boolean} trackInvites Whether or not to track invites.
 * @param {*} client The client object.
 */
async function checkInvite(member, auditChannel, trackInvites, client)
{
    if (!trackInvites) return; // If disabled, bail.
    member.guild.invites.fetch().then((newInvites) =>
    {
        const oldInvites = invites.get(member.guild.id);
        const invite = newInvites.find((index) => index.uses > oldInvites.get(index.code));
        if (!invite)
            return auditChannel.send(`${ member.user.tag } ${ global.locale[global.findLocale(member.guild.id)].joinedUnknownInvite }`);
        const inviter = client.users.cache.get(invite.inviter.id);
        if (inviter)
            auditChannel.send(`${ member.user.tag } ${ global.locale[global.findLocale(member.guild.id)].joinedWithInvite } \`${ invite.code }\` ${ global.locale[global.findLocale(member.guild.id)].from } ${ inviter.tag }.`);
    });
}

/**
 * @name welcomeMember
 * @description Send a welcome message.
 * @param {Discord.GuildMember} member A valid member object.
 * @param {*} client The client object.
 */
async function welcomeMember(member, client)
{
    const guildId = member.guild.id;
    const { welcomeChannelID, welcomeMessage } = common.getGuildConfig(guildId);
    if (!welcomeChannelID || welcomeChannelID.length < 10 || !welcomeMessage || welcomeMessage < 1) return;
    const channel = client.channels.cache.get(welcomeChannelID);
    if (!channel) return;
    const message = welcomeMessage.replaceAll(`{user}`, `<@!${ member.user.id }>`);
    return channel.send(message);
}

/**
 * @param {*} member A valid member object.
 */
async function checkPersistedRoles(member)
{
    const roleArray = sql.prepare(`SELECT roleArray FROM rolePersists WHERE guildId = ? AND userId = ?`).get(member.guild.id, member.id);
    if (!roleArray)
        return;
    if (roleArray.length === 0)
        return sql.prepare(`DELETE FROM rolePersists WHERE guildId = ? AND userId = ?`).run(member.guild.id, member.id);

    const roleArrayParsed = JSON.parse(roleArray.roleArray);
    for (const role of roleArrayParsed)
    {
        const roleObject = member.guild.roles.cache.get(role);
        if (roleObject)
            await member.roles.add(roleObject);
    }
}
