import { AuditLogEvent } from "discord.js";
import sqlite from "../utils/sqlite.js";
import common from "../utils/common.js";

/**
 * @name guildMemberRemove
 * @description Event that fires when a user leaves a guild.
 */
export default {
    name: `guildMemberRemove`,
    /**
     * @param {Discord.GuildMember} member A valid member object.
     * @param {*} Discord The Discord Object.
     * @param {*} client The client object.
     */
    async execute(member, Discord, client)
    {
        const guildId = member.guild.id;
        const { auditChannelID } = common.getGuildConfig(guildId);

        await goodbyeMember(member, client); // Send a goodbye message.
        await common.sleep(3000); // delay to get audit log entry

        const guild = client.guilds.cache.get(guildId);
        if (!guild)
            return;

        const fetchedLogs = await member.guild.fetchAuditLogs({
            limit: 1,
            type: AuditLogEvent.MemberKick,
        });

        const kickLog = fetchedLogs.entries.first();
        if (!kickLog)
            return;

        // Check what the time is on the audit log
        const auditLogTime = kickLog.createdTimestamp;
        const auditLogTimeDiff = Date.now() - auditLogTime;
        if (auditLogTimeDiff > 60_000)
            return; // if the audit log is more than a minute old, dont send the embed (because it could be from a previous punishment)

        const { executor, target, action } = kickLog;

        if (executor.id === client.user.id)
            return;
        if (target.id !== member.id)
            return;
        if (action !== AuditLogEvent.MemberKick)
            return;

        var reason = kickLog.reason;
        if (!reason)
            reason = `${ global.locale[global.findLocale(member.guild.id)].noReasonSpecified }`;

        var newReason = `🦶 ${ reason }`;
        await sqlite.modifyPunishmentsProfile(client, target.id, executor.id, guildId, `kicks`, 1, newReason);

        if (!auditChannelID || auditChannelID.length < 10) return;
        const auditChannel = client.channels.cache.get(auditChannelID);
        if (!auditChannel) // auditChannelID is invalid.
            return;
        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🦶 (${ global.locale[global.findLocale(member.guild.id)].kick })` })
            .addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].user }`, value: `<@!${ member.id }> (${ member.user.username })`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].moderator }`, value: `<@!${ executor.id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(member.guild.id)].reason }`, value: reason, inline: true }]);
        auditChannel.send({ embeds: [embed] });
    }
};

/**
 * @name goodbyeMember
 * @description Send a goodbye message.
 * @param {Discord.GuildMember} member A valid member object.
 * @param {*} client The client object.
 */
async function goodbyeMember(member, client)
{
    const guildId = member.guild.id;
    const { welcomeChannelID, leavingMessage } = common.getGuildConfig(guildId);
    if (!welcomeChannelID || welcomeChannelID.length < 10 || !leavingMessage || leavingMessage < 1) return;
    const channel = client.channels.cache.get(welcomeChannelID);
    if (!channel) return;
    const message = leavingMessage.replaceAll(`{user}`, `<@!${ member.user.id }>`);
    return channel.send(message);
}
