import { AuditLogEvent } from "discord.js";
import SQLite from "better-sqlite3";
import mute from "../commands/commands/moderation/mute.js";
import unmute from "../commands/commands/moderation/unmute.js";
import common from "../utils/common.js";
import sqlite from "../utils/sqlite.js";

const sql = new SQLite(`./src/db.sqlite`);

/**
 * @name guildMemberUpdate
 * @description Event that fires when a member is updated.
 */
export default {
    name: `guildMemberUpdate`,
    /**
     * @param {Discord.GuildMember} oldMember The old member object.
     * @param {Discord.GuildMember} newMember The new member object.
     * @param {*} Discord The Discord object.
     * @param {*} client The client object.
     */
    async execute(oldMember, newMember, Discord, client)
    {
        const guildId = oldMember.guild.id;
        const { auditChannelID, mutedRoleID } = common.getGuildConfig(guildId);

        if (!auditChannelID)
            return;

        const auditChannel = client.channels.cache.get(auditChannelID);
        if (!auditChannel)
            return;

        await sendRoleUpdates(oldMember, newMember, auditChannel, mutedRoleID, Discord, client);
        await sendNameChange(oldMember, newMember, auditChannel, Discord, client);
        await sendManualTimeout(oldMember, newMember, auditChannel, Discord, client);
        await autoRoleOnJoin(oldMember, newMember);
    }
};

/**
 * @name sendRoleUpdates
 * @description Sends an embed to the audit channel when a member's roles are updated.
 * @param {Discord.GuildMember} oldMember The old member object.
 * @param {Discord.GuildMember} newMember The new member object.
 * @param {Discord.TextChannel} auditChannel The audit channel object.
 * @param {Discord.Role.id} mutedRoleID The ID of the muted role.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function sendRoleUpdates(oldMember, newMember, auditChannel, mutedRoleID, Discord, client)
{
    // Role Handling
    const guildId = oldMember.guild.id;
    const fetchedLogs = await oldMember.guild.fetchAuditLogs({
        limit: 1,
        type: AuditLogEvent.MemberRoleUpdate,
    });
    const roleAddLog = fetchedLogs.entries.first();
    const { executor } = roleAddLog;
    if (oldMember.roles.cache.size > newMember.roles.cache.size)
    {
        for (let role of oldMember.roles.cache)
        {
            role = role[1];

            if (!newMember.roles.cache.has(role.id))
            { // Removed a Role
                return removedRole(oldMember, newMember, auditChannel, role, executor, mutedRoleID, guildId, Discord, client);
            }
        }
    }
    else if (oldMember.roles.cache.size < newMember.roles.cache.size)
    {
        for (let role of newMember.roles.cache)
        {
            role = role[1];
            if (!oldMember.roles.cache.has(role.id))
            { // Added a Role
                return addRole(oldMember, newMember, auditChannel, role, executor, mutedRoleID, guildId, Discord, client);
            }
        }
    }
}
/**
 * @name removedRole
 * @description Sends an embed to the audit channel when a member's roles are updated.
 * @param {Discord.GuildMember} oldMember The old member object.
 * @param {Discord.GuildMember} newMember The new member object.
 * @param {Discord.TextChannel} auditChannel The audit channel object.
 * @param {Discord.Role} role The role object being removed from the user.
 * @param {Discord.GuildMember} executor The member who removed the role.
 * @param {Discord.Role.id} mutedRoleID The ID of the muted role.
 * @param {Discord.Guild.id} guildId The ID of the guild.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function removedRole(oldMember, newMember, auditChannel, role, executor, mutedRoleID, guildId, Discord, client)
{
    if (role.id === mutedRoleID && executor && executor.id !== client.user.id)
    { // If someone was unmuted by removing the role manually:
        sql.prepare(`DELETE FROM mutes WHERE guildId = ? AND userId = ?`).run(guildId, oldMember.id);
        var reason = `${ global.locale[global.findLocale(oldMember.guild.id)].manuallyRemovedRole }`;
        unmute.embed(auditChannel.id, oldMember, executor.id, reason, false, oldMember.id, client, Discord);
        var newReason = `🗣️ ${ reason }`;
        sqlite.modifyPunishmentsProfile(client, oldMember.id, executor.id, guildId, `kicks`, 0, newReason); // We hijack kicks here, but don't actually add anything to it. Just to get it into the punishment history.
        return;
    }
    if (role.id === mutedRoleID) return; // Do not show role removal if unmute as we handle it elsewhere.
    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `❌ (${ global.locale[global.findLocale(oldMember.guild.id)].roleRemoved })` })
        .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].user }`, value: `<@!${ oldMember.id }> (${ oldMember.user.username })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].role }`, value: `<@&${ role.id }>`, inline: true }]);
    if (executor && !role.managed) // Don't show moderator for the "Server Booster" role, or other managed roles.
        embed.addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].moderator }`, value: `<@!${ executor.id }>`, inline: true }]);
    auditChannel.send({ embeds: [embed] });
}

/**
 * @name addRole
 * @description Sends an embed to the audit channel when a member's roles are updated.
 * @param {Discord.GuildMember} oldMember The old member object.
 * @param {Discord.GuildMember} newMember The new member object.
 * @param {Discord.TextChannel} auditChannel The audit channel object.
 * @param {Discord.Role} role The role object being added to the user.
 * @param {Discord.GuildMember} executor The member who added the role.
 * @param {Discord.Role.id} mutedRoleID The ID of the muted role.
 * @param {Discord.Guild.id} guildId The ID of the guild.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function addRole(oldMember, newMember, auditChannel, role, executor, mutedRoleID, guildId, Discord, client)
{
    if (role.id === mutedRoleID && executor && executor.id !== client.user.id)
    { // If we manually added the muted role, mute the user in the DB & send embed
        const reason = `${ global.locale[global.findLocale(oldMember.guild.id)].manuallyAddedRole }`;
        mute.sendMuteToDB(client, oldMember.guild.id, executor.id, oldMember.user.id, -1, reason);
        mute.handleMute(Discord, client, newMember, executor.id, role, executor.tag, reason, -1, false, newMember.guild.id, newMember.id);
        return; // Do not send role update for muting.
    }
    if (role.id === mutedRoleID) return; // If bot muted, do not send embed.
    const embed = new Discord.EmbedBuilder()
        .setColor(`#00ff00`)
        .setTimestamp()
        .setAuthor({ name: `✅ (${ global.locale[global.findLocale(oldMember.guild.id)].roleAdded })` })
        .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].user }`, value: `<@!${ oldMember.id }> (${ oldMember.user.username })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].role }`, value: `<@&${ role.id }>`, inline: true }]);
    if (executor && !role.managed) // Don't show moderator for the "Server Booster" role, or other managed roles.
        embed.addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].moderator }`, value: `<@!${ executor.id }>`, inline: true }]);
    auditChannel.send({ embeds: [embed] });
}
/**
 * @name sendNameChange
 * @description Sends an embed to the audit channel when a member's name is updated.
 * @param {Discord.GuildMember} oldMember The old member object.
 * @param {Discord.GuildMember} newMember The new member object.
 * @param {Discord.TextChannel} auditChannel The audit channel object.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function sendNameChange(oldMember, newMember, auditChannel, Discord, client)
{
    // Check if the user was renamed manually.
    if (oldMember.nickname !== newMember.nickname)
    {
        await common.sleep(1000);

        const fetchedLogs = await oldMember.guild.fetchAuditLogs({
            limit: 1,
            type: AuditLogEvent.MemberUpdate,
        });
        const memberUpdateLog = fetchedLogs.entries.first();
        // Check what the time is on the audit log
        const auditLogTime = memberUpdateLog.createdTimestamp;
        const auditLogTimeDiff = Date.now() - auditLogTime;
        if (auditLogTimeDiff > 60_000)
            return; // if the audit log is more than a minute old, dont send the embed (because it could be from a previous punishment)

        const { executor, target } = memberUpdateLog;
        if (memberUpdateLog && target.id === oldMember.id && executor && executor.id !== client.user.id)
        {
            const embed = new Discord.EmbedBuilder()
                .setColor(`#ff1e00`)
                .setTimestamp()
                .setAuthor({ name: `👤 (${ global.locale[global.findLocale(oldMember.guild.id)].rename })` })
                .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].oldName }`, value: oldMember.nickname || `${ oldMember.user.username }`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].newName }`, value: newMember.nickname || `${ newMember.user.username }`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].user }`, value: `<@!${ oldMember.id }>`, inline: true }]);
            if (executor && executor.id !== oldMember.id)
                embed.addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].moderator }`, value: `<@!${ executor.id }>`, inline: true }]);
            auditChannel.send({ embeds: [embed] });
        }
    }
}

/**
 * @name sendManualTimeout
 * @description Sends an embed to the audit channel when a member is manually timed out.
 * @param {Discord.GuildMember} oldMember The old member object.
 * @param {Discord.GuildMember} newMember The new member object.
 * @param {Discord.TextChannel} auditChannel The audit channel object.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function sendManualTimeout(oldMember, newMember, auditChannel, Discord, client)
{
    // Check if we timed the user out
    let memberTimeoutLog;
    let executor;

    // If we either timed someone out, or removed it. (Used to grab audit log & other info.)
    if (newMember.communicationDisabledUntilTimestamp > Date.now()
        || (oldMember.communicationDisabledUntilTimestamp !== null && newMember.communicationDisabledUntilTimestamp === null))
    {
        await common.sleep(1000);

        const fetchedLogs = await oldMember.guild.fetchAuditLogs({
            limit: 1,
            type: AuditLogEvent.MemberUpdate,
        });
        memberTimeoutLog = fetchedLogs.entries.first();
        if (!memberTimeoutLog)
            return;

        // Check what the time is on the audit log
        const auditLogTime = memberTimeoutLog.createdTimestamp;
        const auditLogTimeDiff = Date.now() - auditLogTime;
        if (auditLogTimeDiff > 60_000)
            return; // if the audit log is more than a minute old, dont send the embed (because it could be from a previous punishment)

        var reason = memberTimeoutLog.reason;
        if (!reason)
            reason = `${ global.locale[global.findLocale(oldMember.guild.id)].noReasonSpecified }`;

        executor = memberTimeoutLog.executor;
    }

    // If they were timed out and the bot didn't time them out
    if (newMember.communicationDisabledUntilTimestamp > Date.now()
        && memberTimeoutLog && memberTimeoutLog.target.id === oldMember.id && executor && executor.id !== client.user.id)
    {
        const friendlyTime = common.getFriendlyTime(newMember.communicationDisabledUntilTimestamp);
        mute.sendMuteToDB(client, oldMember.guild.id, executor.id, oldMember.user.id, newMember.communicationDisabledUntilTimestamp, reason, friendlyTime);
        // ^^ We send this to the DB regardless of if we have a mutedRoleID so it logs in `!lookup`, so we can see who times out people.

        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🤫 (${ global.locale[global.findLocale(oldMember.guild.id)].mute })` })
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].user }`, value: `<@!${ oldMember.id }> (${ oldMember.user.username })`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].moderator }`, value: `<@!${ executor.id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].expires }`, value: `<t:${ Math.round(newMember.communicationDisabledUntilTimestamp / 1000) }:R>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].reason }`, value: `${ reason }`, inline: true }]);
        auditChannel.send({ embeds: [embed] });
    }

    // If the ban was manually removed and the bot didn't remove the timeout
    if (newMember.communicationDisabledUntilTimestamp === null
        && oldMember.communicationDisabledUntilTimestamp !== null
        && executor && executor.id !== client.user.id)
    {
        // Remove the timeout from the DB
        const row = sql.prepare(`SELECT * FROM mutes WHERE guildId = ? AND userId = ?`).get(oldMember.guild.id, oldMember.id);
        if (row)
            sql.prepare(`DELETE FROM mutes WHERE guildId = ? AND userId = ?`).run(oldMember.guild.id, oldMember.id);

        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🗣️ (${ global.locale[global.findLocale(oldMember.guild.id)].unmute })` })
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].user }`, value: `<@!${ newMember.id }> (${ newMember.user.username })`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].moderator }`, value: `<@!${ executor.id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(oldMember.guild.id)].reason }`, value: `Manually removed timeout`, inline: true }]);

        auditChannel.send({ embeds: [embed] });
    }
}

/**
 * @name autoRoleOnJoin
 * @description Automatically adds a role to a member when they join.
 * @param {Discord.GuildMember} oldMember The old member object.
 * @param {Discord.GuildMember} newMember The new member object.
 */
async function autoRoleOnJoin(oldMember, newMember)
{
    if (oldMember.pending && !newMember.pending)
    {
        // Auto Role on Join
        const guildId = oldMember.guild.id;
        const guildConfig = global.getConfiguration.get(guildId);
        const GUILD_CONFIG = JSON.parse(guildConfig.config);
        const { autoJoinRoleID } = GUILD_CONFIG;
        if (!autoJoinRoleID || autoJoinRoleID === ``) // If there is no autoJoinRoleID, then we don't need to do anything.
            return;
        const role = oldMember.guild.roles.cache.get(autoJoinRoleID);
        if (!role) // Role doesn't exist? Let's empty out the config so we don't keep trying to add it.
        {
            GUILD_CONFIG.autoJoinRoleID = ``;
            global.setConfiguration.run({
                guildId,
                config: JSON.stringify(GUILD_CONFIG)
            });
            return;
        }
        await oldMember.roles.add(role);
    }
}
