import
{
    InteractionType, TextInputBuilder, ModalBuilder, ActionRowBuilder
} from "discord.js";
import SQLite from "better-sqlite3";
import logging from "../utils/logging.js";
import sale from "../commands/commands/utility/sale.js";
import { accept, reject } from "../utils/appeal.js";
import common from "../utils/common.js";
import
{
    makeTicket, closeTicket, reopenTicket, saveTranscriptAdmin, saveTranscriptUser
} from "../utils/tickets.js";
import giveaway from "../utils/giveaway.js";
import { handleButtonRole } from "../utils/selectroles.js";
import embed from "../commands/commands/utility/embed.js";
import ban from "../commands/commands/moderation/ban.js";
import sqlite from "../utils/sqlite.js";

reactionMessageIds = global.reactionMessageIds;
const sql = new SQLite(`./src/db.sqlite`);

/**
 * @name interactionCreate
 * @description Event that fires when a user interacts with a message, or sends a Slash Command.
 */
export default {
    name: `interactionCreate`,
    /**
     * @param {Discord.Interaction} interaction A valid interaction object.
     * @param {*} Discord The Discord Object.
     * @param {*} client The client object.
     */
    async execute(interaction, Discord, client)
    {
        // command/app handling
        if (interaction.type === InteractionType.ApplicationCommand
            || interaction.type === InteractionType.isContextMenu
            || interaction.type === InteractionType.isUserContextMenu
            || interaction.type === InteractionType.isMessageContextMenu)
        {
            const command = client.commands.get(interaction.commandName);
            try
            {
                command.execute(interaction, client, Discord);
            }
            catch (error)
            {
                logging.error(error);
            }
            return;
        }

        // modal handling
        if (interaction.type === InteractionType.ModalSubmit)
        {
            await interaction.deferReply({ ephemeral: true });

            switch (interaction.customId)
            {
                case `appealInfo_accept`:
                case `appealInfo_reject`: {
                    const [appealInfoResponse] = [`appealInfo_${ interaction.user.id }_response`]
                        .map((id) => interaction.fields.getTextInputValue(id));

                    let reason = `${ global.locale[global.findLocale(interaction.guild.id)].noReasonSpecified }`;
                    if (appealInfoResponse)
                        reason = appealInfoResponse;

                    if (interaction.customId === `appealInfo_accept`)
                        await accept(interaction, interaction.message.id, reason, Discord, client);
                    else if (interaction.customId === `appealInfo_reject`)
                        await reject(interaction, interaction.message.id, reason, Discord, client);
                    return;
                }
            }
        }

        // sending modal interactions
        switch (interaction.customId)
        {
            // ban appeals
            case `accept`:
            case `reject`: {
                const guildId = interaction.guild.id;
                const { banAppealsChannelID } = common.getGuildConfig(guildId);

                if (!common.canRunCommand(interaction, `BanMembers`, true))
                    return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`);
                if (!banAppealsChannelID || banAppealsChannelID.length < 10)
                    return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].thisCommandIsNotSetup } ${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].banAppealsChannelIdNotSetup }`);
                const banAppealsChannel = client.channels.cache.get(banAppealsChannelID);
                if (!banAppealsChannel)
                    return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].banAppealsChannelIdInvalid }`);

                const modal = new ModalBuilder()
                    .setTitle(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].appealAction }`)
                    .setCustomId(`appealInfo_${ interaction.customId }`);

                const rows = new ActionRowBuilder().addComponents(
                    new TextInputBuilder()
                        .setCustomId(`appealInfo_${ interaction.user.id }_response`)
                        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].reason }:`)
                        .setStyle(1)
                        .setMinLength(0)
                        .setMaxLength(250)
                        .setRequired(false)
                );

                modal.addComponents(rows);
                interaction.showModal(modal);
                return;
            }
        }

        if (interaction.customId === `openticket`)
            return makeTicket(interaction, Discord, client);
        if (interaction.customId.startsWith(`closeticket`))
            return closeTicket(interaction, Discord);
        if (interaction.customId.startsWith(`reopenticket`))
            return reopenTicket(interaction, Discord);
        if (interaction.customId.startsWith(`savetranscriptadmin`))
            return saveTranscriptAdmin(interaction, Discord, client);
        if (interaction.customId.startsWith(`savetranscriptuser`))
            return saveTranscriptUser(interaction, Discord, client);
        if (interaction.customId === `entergiveaway`)
            return giveaway.giveawayEntry(interaction);
        if (interaction.customId === `selectroles`)
            return handleButtonRole(interaction);
        if (interaction.customId.startsWith(`embedgenerator`))
        {
            const channelId = interaction.customId.split(`-`)[1];
            return embed.handleEmbedGenerator(interaction, channelId, Discord);
        }
        if (interaction.customId.startsWith(`modal-embedgenerator`))
        {
            const channelId = interaction.customId.split(`-`)[2];
            return embed.handleEmbedGeneratorModal(interaction, channelId, Discord, client);
        }
        if (interaction.customId.startsWith(`ban`))
        {
            if (!common.canRunCommand(interaction, `BanMembers`, true))
                return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`, ephemeral: true }); // dont fail interaction, but tell them what happened.

            const userId = interaction.customId.split(`-`)[1];
            let reason = `https://discord.com/channels/${ interaction.guild.id }/${ interaction.channel.id }/${ interaction.message.id }`;
            if (interaction.customId.includes(`Scam`))
                reason = `Scam Message (${ reason })`;
            const member = await common.getGuildMember(client, interaction.guild.id, userId);
            if (member)
                ban.banAndSendEmbed(Discord, client, userId, member.user.username, interaction.user.id, reason, member);
            else
            {
                const authorTag = interaction.user.tag;
                await interaction.guild.members.ban(userId, { deleteMessageSeconds: 7 * 24 * 60 * 60, reason: `${ authorTag } | ${ reason }` }).then(async () =>
                {
                    ban.sendBanEmbed(Discord, client, userId, userId, interaction.user.id, reason, interaction.guild.id, -1);
                    var newReason = `🔨 ${ reason }`;
                    await sqlite.modifyPunishmentsProfile(client, userId, interaction.user.id, interaction.guild.id, `bans`, 1, newReason, `permanently`);
                });
            }
            return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].interactionBasicSuccess }`, ephemeral: true }); // Done.
        }
        if (interaction.customId.startsWith(`autoModerationAdd`))
        {
            if (!common.canRunCommand(interaction, `BanMembers`, true))
                return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`, ephemeral: true }); // dont fail interaction, but tell them what happened.

            const messageId = interaction.message.id;

            // Fetch that message
            const message = await interaction.channel.messages.fetch(messageId);
            if (!message)
                return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].cantFindMessageToAddToAutoBan }`, ephemeral: true });

            // Look at the embed
            const fields = message.embeds[0].fields;
            let content = ``;
            for (const [index, field] of fields.entries())
            {
                if (index.name === global.locale[global.findLocale(interaction.guild.id)].content
                    || global.locale[global.findLocale(undefined, interaction.locale)].content)
                    content = field.value.toLowerCase();
            }

            content = content.replaceAll(`@everyone`, ``).replaceAll(`@here`, ``);
            if (content.includes(`\`\`\``))
                content = content.split(`\`\`\``)[1];
            const domainsRegex = /(http:\/\/|https:\/\/|www\.|hxxp:\/\/|hxxps:\/\/)?([\da-z][\da-z-]*\.)+[\da-z][\da-z-]*/gi; // http:// , https:// , www. , hxxp:// , hxxps://
            const domains = content.match(domainsRegex);
            if (domains)
            {
                for (const domain of domains)
                    content = content.replace(domain, ``);
            }
            content = content.trim();
            content = content.slice(0, 100);
            try
            {
                sql.prepare(`INSERT INTO autoBannedPhrases (guildId, phrase) VALUES (?, ?)`).run(interaction.guild.id, content);
            }
            catch
            {
                return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].cantFindMessageToAddToAutoBan }`, ephemeral: true });
            }
            return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].addedToList.replace(`%item%`, content) } <@${ interaction.user.id }>`, ephemeral: false });
        }

        // React to Delete Bot Message 3
        if (reactionMessageIds.includes(interaction.message.id))
        {
            if (interaction.user.id !== interaction.customId
                && !common.canRunCommand(interaction, `ManageMessages`, true)) // customId == message author Id. bypass for manage message perms
                return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`, ephemeral: true }); // dont fail interaction, but tell them what happened.
        }
        else if (sale.mw2SaleIds.includes(interaction.message.id) && !common.canRunCommand(interaction, `ManageMessages`, true)) // customId == message author Id. bypass for manage message perms)
            return interaction.reply({ content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`, ephemeral: true });

        let invalidChannel = false;
        const channel = await client.channels.fetch(interaction.message.channelId)
            .catch(() =>
            {
                invalidChannel = true;
            });
        if (invalidChannel)
            return;

        let invalidMessage = false;
        const message = await channel.messages.fetch(interaction.message.id)
            .catch(() =>
            {
                invalidMessage = true;
            });
        if (invalidMessage)
            return;

        message.delete()
            .then(() =>
            {
                if (reactionMessageIds.includes(interaction.message.id))
                    reactionMessageIds.splice(reactionMessageIds.indexOf(interaction.message.id), 1); // remove from array since we no longer need
                else if (sale.mw2SaleIds.includes(interaction.message.id))
                    sale.mw2SaleIds.splice(sale.mw2SaleIds.indexOf(interaction.message.id), 1);
            });
    },
};
