import SQLite from "better-sqlite3";
import { createRequire } from "node:module";
import { ChannelType } from "discord.js";
import logging from "../utils/logging.js";
import messageDetection from "../utils/messageDetection.js";
import common from "../utils/common.js";

const require = createRequire(import.meta.url);
const { forwardDMs, botOwnerID } = require(`../config.json`);

const sql = new SQLite(`./src/db.sqlite`);

/**
 * @name messageCreate
 * @description Event that fires when a message is created.
 */
export default {
    name: `messageCreate`,
    /**
     * @param {Discord.Message} message The message object.
     * @param {*} Discord The Discord Object.
     * @param {*} client The Discord Client Object.
     */
    async execute(message, Discord, client)
    {
        if (message.author.bot && message.author.id !== client.user.id) return; // ignore other bots
        if (message.channel.type === ChannelType.DM && message.author.id !== client.user.id && forwardDMs)
        {
            await forwardMessage(message, Discord, client);
            return;
        }
        if (message.channel.type === ChannelType.DM)
            return; // ignore DMs after forwarding / if forwarding is disabled

        const { prefix } = common.getGuildConfig(message.guild.id);

        // Prepare multiple prefix(es)
        if (prefix)
            var isPrefix = prefix.find((p) => message.content.startsWith(p));

        await messageDetection.messageDetection(message, Discord, client);

        // Command Handling
        if (isPrefix)
        {
            if (message.content === `${ isPrefix }`)
                return; // Don't try and do anything if you just enter a single character that happens to be a prefix.
            const commandsBotIsAllowedToRun = [`cdb!lookup`, `cdb!bat`, `cdb!forcejoin`];
            if (message.author.id === `${ client.user.id }` && !commandsBotIsAllowedToRun.some((c) => message.content.toLowerCase().startsWith(c)))
                return; // Let bot run above commands.

            return runCommand(isPrefix, message, client, Discord);
        }

        await common.checkName(Discord, client, message.member);
    }
};

/**
 * @name forwardMessage
 * @description Forwards a message to the bot owner.
 * @param {Discord.Message} message The message object.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client Object.
 */
async function forwardMessage(message, Discord, client)
{
    // Do not Forward DMs from ignored users.
    const checkIfTheyAreDMIgnored = sql.prepare(`SELECT userId FROM doNotForwardDMsFrom WHERE userId = ?`).get(message.author.id);
    if (checkIfTheyAreDMIgnored)
        return;

    // handle stickers in a message
    // as far as I can tell, you should only be able to send 1 sticker. - mikey
    if (message.stickers.size > 0)
    {
        const stickers = [...message.stickers]; // convert to a array, i don't like collections
        message.content = `${ message.content } **[${ stickers[0][1].name } Sticker]**`;
    }

    message.content = common.truncateString(message.content, 1024);
    logging.log(`${ message.author.tag } ${ global.locale[global.gLocale].userSentDM }`);
    const embed = new Discord.EmbedBuilder()
        .setColor(`#0099ff`)
        .setTitle(`${ global.locale[global.gLocale].newDM } ${ global.locale[global.gLocale].from } ${ message.author.tag } (${ message.author.id })`)
        .setDescription(`${ message.content }`);

    if (message.stickers.size > 0)
        embed.addFields([{ name: `Stickers Data`, value: `**Size: ${ message.stickers.size }**`, inline: true }]);

    for (const ownerID of botOwnerID)
    {
        const owner = await client.users.fetch(ownerID);
        if (owner)
        {
            owner.send({ embeds: [embed] });

            // sending message attachments
            if (message.attachments.size > 0)
            {
                const files = [];
                for (const file of message.attachments.values())
                    files.push(file.url);

                owner.send({ content: `**${ message.attachments.size } message attachments:**\n${ files.join(` \n`) }` }, { files });
            }
        }
    }
}

/**
 * @name runCommand
 * @description Runs a command.
 * @param {string} isPrefix The prefix used to run the command.
 * @param {Discord.Message} message The message object.
 * @param {*} client The Client Object.
 * @param {*} Discord The Discord Object.
 * @returns {undefined} Command Execution if checks pass
 */
export function runCommand(isPrefix, message, client, Discord)
{
    const suppliedArguments = isPrefix.startsWith(`<@`) ? message.content.slice(isPrefix.length).trim().split(/ +/) : message.content.slice(isPrefix.length).split(/ +/);

    const commandName = suppliedArguments.shift().toLowerCase();
    const command = client.chatcommands.get(commandName)
        || client.chatcommands.find((cmd) => cmd.aliases && cmd.aliases.includes(commandName));

    if (!command)
        return;

    if (command.ourArguments && suppliedArguments.length === 0)
        return message.delete();

    if ((command.elevatedOnly && global.elevatedServers.includes(message.guild.id)) || !command.elevatedOnly)
    { // If the command is elevated only, check we are in a valid server. If not, don't run the command.
        // If the command is not elevated only, run it also.
        try
        {
            command.execute(message, suppliedArguments, Discord, client);
            return message.delete();
        }
        catch (error)
        {
            logging.error(error);
            message.delete();
        }
    }

    message.delete();
}
