import { ChannelType } from "discord.js";
import messageDetection from "../utils/messageDetection.js";
import common from "../utils/common.js";

/**
 * @name messageUpdate
 * @description Event that fires when a message is updated.
 */
export default {
    name: `messageUpdate`,
    /**
     * @param {Discord.Message} oldMessage The old message object.
     * @param {Discord.Message} newMessage The new message object.
     * @param {*} Discord The Discord Object.
     * @param {*} client The Client Object.
     */
    async execute(oldMessage, newMessage, Discord, client)
    {
        await messageDetection.messageDetection(newMessage, Discord, client);
        await logSelfEdits(oldMessage, newMessage, Discord, client);
    }
};

/**
 * @name logSelfEdits
 * @description Sends a message to the log channel if a message is edited by the user.
 * @param {Discord.Message} oldMessage The old message object.
 * @param {Discord.Message} newMessage The new message object.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Client Object.
 */
async function logSelfEdits(oldMessage, newMessage, Discord, client)
{
    if (oldMessage.author.bot && oldMessage.author.id !== client.user.id) return; // ignore other bots
    if (oldMessage.channel.type === ChannelType.DM)
        return; // ignore DMs

    const { logSelfEdits, deletedMessagesChanID } = common.getGuildConfig(oldMessage.guild.id);

    // If disabled, do nothing.
    if (!logSelfEdits)
        return;

    if (oldMessage.content === newMessage.content)
        return; // ignore if content is the same

    const embed = new Discord.EmbedBuilder()
        .setColor(`WHITE`)
        .setTimestamp()
        .setAuthor({ name: `✏️ (${ global.locale[global.findLocale(oldMessage.guild.id)].selfEdited })` })
        .addFields([{ name: `${ global.locale[global.findLocale(oldMessage.guild.id)].user }`, value: ` <@!${ oldMessage.member.id }> (${ oldMessage.member.user.username })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(oldMessage.guild.id)].channel }`, value: ` <#${ oldMessage.channel.id }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(oldMessage.guild.id)].oldContent }`, value: common.truncateString(oldMessage.content, 1024), inline: false }])
        .addFields([{ name: `${ global.locale[global.findLocale(oldMessage.guild.id)].newContent }`, value: common.truncateString(newMessage.content, 1024), inline: false }]);

    const delMessageChan = client.channels.cache.get(deletedMessagesChanID);
    if (delMessageChan)
        delMessageChan.send({ embeds: [embed] });
}
