import SQLite from "better-sqlite3";
import { createRequire } from "node:module";
import { ActivityType } from "discord.js";

// other js files in the project
import sqlite from "../utils/sqlite.js";
import logging from "../utils/logging.js";
import sale from "../commands/commands/utility/sale.js";
import version from "../commands/commands/informational/version.js";
import common from "../utils/common.js";
import giveaway from "../utils/giveaway.js";
import embed from "../commands/commands/moderation/unmute.js";

const require = createRequire(import.meta.url);
const { playingStatus, isWorkflow } = require(`../config.json`);

const sql = new SQLite(`./src/db.sqlite`);
export const invites = new Map();

/**
 * @name ready
 * @description Event that fires when the bot is ready.
 */
export default {
    name: `ready`,
    once: true,
    invites,
    /**
     * @param {*} client The Client Object.
     * @param {*} Discord The Discord Object.
     */
    async execute(client, Discord)
    {
        // cache stuff (this doesn't take long at all)
        await client.guilds.fetch()
            .then(async () =>
            {
                logging.log(`${ global.locale[global.gLocale].guildsCached }`);
                // once guild is cached, let's cache the guild's roles
                for (const guild of client.guilds.cache.values())
                {
                    // Check if guild is banned
                    const inDatabase = sql.prepare(`SELECT * FROM bannedServers WHERE guildId = ?;`).get(guild.id);
                    if (inDatabase)
                        return guild.leave();

                    await common.deploySlashCommands(guild);

                    // Cache Invites
                    const firstInvites = await guild.invites.fetch();
                    invites.set(guild.id, new Map(firstInvites.map((invite) => [invite.code, invite.uses])));
                }
            });

        if (client.guilds.cache.size === 0)
            console.warn(`${ global.locale[global.gLocale].inviteLink }:\nhttps://discord.com/api/oauth2/authorize?client_id=${ client.user.id }&permissions=8&scope=bot%20applications.commands`);

        // elevated server specific things
        try
        {
            for (const guildId of global.elevatedServers)
            {
                // Check if the bot is in the guild
                const guild = client.guilds.cache.get(guildId);
                if (guild)
                {
                    const { auditChannelID, generalChannelID } = common.getGuildConfig(guildId);

                    if (auditChannelID && auditChannelID.length > 10)
                    {
                        const auditChannel = client.channels.cache.get(auditChannelID);
                        if (auditChannel)
                            version.getCommitInfo(false, auditChannel, false, true);
                    }

                    if (generalChannelID && generalChannelID.length > 10)
                    {
                        sale.checkSteamMW2Sale(generalChannelID, client, true); // On boot, we check. Then Interval handles it every 4~ hours.

                        setInterval(() =>
                        {
                            sale.checkSteamMW2Sale(generalChannelID, client, true);
                        }, 14_400_000); // every 4~ hours
                    }
                    await common.deploySlashCommands(guild);
                }
            }
        }
        catch (error)
        {
            logging.error(error);
        }

        if (playingStatus !== ``)
            client.user.setActivity(`${ playingStatus }`, { type: ActivityType.Playing });

        await intervalChecks(client, Discord);

        // Migrate any bans that are not in the database.
        for (const guild of client.guilds.cache.values())
        {
            const banArray = await common.getAllGuildBans(guild);
            for (const userId of banArray)
            {
                const punishment = sql.prepare(`SELECT * FROM punishments WHERE guildId = ${ guild.id } AND id = ${ userId }`).get();
                if (!punishment)
                {
                    // fetch the ban reason
                    const ban = await guild.bans.fetch(userId);
                    const reason = ban.reason || `${ global.locale[global.findLocale(guild.id)].noReasonSpecified }`;
                    var newReason = `🔨 ${ reason }`;
                    await sqlite.modifyPunishmentsProfile(client, userId, undefined, guild.id, `bans`, 1, newReason);
                }
            }
        }

        logging.log(`${ global.locale[global.gLocale].botIsReady }`);

        if (isWorkflow)
            process.exit(); // We've successfully started the bot from a workflow so we can bail.
    },
};

/**
 * @name intervalChecks
 * @description Set of functions that are repeating.
 * @param {*} client The Client Object.
 * @param {*} Discord The Discord Object.
 */
async function intervalChecks(client, Discord)
{
    /*
        Every 10 minutes:
        - Update Server Counter (if enabled)
    */
    setInterval(async () =>
    {
        const guilds = client.guilds.cache.map((guild) => guild.id);
        for (const guildId of guilds)
        {
            const guild = client.guilds.cache.get(guildId);

            // Loop through all roles until we find the premium one
            let boosterRole;
            for (const role of guild.roles.cache.values())
            {
                if (role.tags && role.tags.premiumSubscriberRole)
                    boosterRole = role;
            }
            // If we get through all roles & didn't find the premium one, fake one and set size to 0.
            if (!boosterRole)
            {
                boosterRole = {
                    members: {
                        size: 0,
                    },
                };
            }

            const {
                memberStatsChanId, boostersStatsChanId, serverStatsMemberChanName, serverStatsBoosterChanName
            } = common.getGuildConfig(guildId);
            if (memberStatsChanId)
            {
                const memberStatsChan = guild.channels.cache.find((channel) => channel.id === memberStatsChanId);
                if (memberStatsChan)
                    memberStatsChan.setName(`${ serverStatsMemberChanName }${ guild.memberCount }`);
            }
            if (boostersStatsChanId)
            {
                const boostersStatsChan = guild.channels.cache.find((channel) => channel.id === boostersStatsChanId);
                if (boostersStatsChan)
                    boostersStatsChan.setName(`${ serverStatsBoosterChanName }${ boosterRole.members.size }`);
            }
        }
    }, 600_000);

    /*
        Every second, check:
        - Mutes
        - Tempbans
    */
    setInterval(async () =>
    {
        const guilds = client.guilds.cache.map((guild) => guild.id);
        for (const guildId of guilds)
        {
            const { mutedRoleID, auditChannelID } = common.getGuildConfig(guildId);

            const allMutes = sql.prepare(`SELECT * FROM mutes WHERE guildId = ${ guildId }`);
            if (allMutes)
            {
                const allMutesArray = allMutes.all();
                for (const index of allMutesArray)
                {
                    let { time } = index; // Grab the time from the row.
                    time = Number.parseInt(time, 10);
                    const { guildId, userId } = index; // Grab said data from i (the current row we are checking).
                    const guild = client.guilds.cache.get(guildId);

                    if ((Date.now() > time) && (time > 0))
                    {
                        sql.prepare(`DELETE FROM mutes WHERE guildId = ? AND userId = ?`).run(guildId, userId);

                        if (mutedRoleID)
                        {
                            let member = await common.getGuildMember(client, guildId, userId);
                            const reason = `${ global.locale[global.findLocale(guildId)].muteRemovedTimeLapsed }`;

                            // if they are in server
                            if (member)
                            {
                                const mutedRole = guild.roles.cache.find((mR) => mR.id === mutedRoleID);
                                if (mutedRole)
                                    member.roles.remove(mutedRole, reason);
                            }
                            else // if we don't have a member object, we fake it since unmute embed only uses guildId from member.
                            {
                                member = {
                                    guild: {
                                        id: guildId,
                                    }
                                };
                            }

                            embed.embed(auditChannelID, member, client.user.id, reason, userId, client, Discord);
                            var newReason = `🗣️ ${ reason }`;
                            sqlite.modifyPunishmentsProfile(client, userId, client.user.id, guildId, `kicks`, 0, newReason); // We hijack kicks here, but don't actually add anything to it. Just to get it into the punishment history.
                        }
                    }
                }
            }

            const allTemporaryBans = sql.prepare(`SELECT * FROM tempbanUsers WHERE guildId = ${ guildId }`);
            if (allTemporaryBans)
            {
                const allTemporaryBansArray = allTemporaryBans.all();
                for (const index of allTemporaryBansArray)
                {
                    const { userId, time } = index;
                    const timeLeft = Number.parseInt(time, 10);
                    const guild = client.guilds.cache.get(guildId);
                    if (guild && Date.now() > timeLeft)
                    {
                        await guild.bans.fetch().then(async (bans) =>
                        {
                            if (bans.size === 0)
                                return; // No bans to unban.
                            let isBanned = true;
                            guild.bans.fetch(userId).catch(() =>
                            {
                                isBanned = false;
                            });
                            if (!isBanned)
                                return; // User is not banned.
                            await guild.members.unban(userId, `${ global.locale[global.findLocale(guild.id)].tempbanLifted }`);
                        });
                        sql.prepare(`DELETE FROM tempbanUsers WHERE guildId = ? AND userId = ?`).run(guildId, userId);
                    }
                }
            }
        }
    }, 1000);

    // Every 30 seconds check if any giveaways are ending.
    setInterval(async () =>
    {
        const databaseEntry = sql.prepare(`SELECT messageId FROM giveaways`).raw().all().flat();
        const databaseEntryNoDupes = [...new Set(databaseEntry)];
        for (const messageId of databaseEntryNoDupes)
            await giveaway.checkActiveGiveaway(messageId, client, Discord);
    }, 30_000);

    // Every 15 seconds check & send out any scheduled messages.
    setInterval(async () =>
    {
        // Check the DB for any scheduled messages.
        for (const row of sql.prepare(`SELECT * FROM scheduledMessages`).all())
        {
            const {
                guildId, channelId, time, content
            } = row;
            const databaseContent = content;
            if (Date.now() > time)
            {
                const guild = client.guilds.cache.get(guildId);
                if (guild)
                {
                    const channel = guild.channels.cache.get(channelId);
                    if (channel)
                    {
                        // regex for any link ending in a valid image extension
                        var regex = /(?:https?:\/\/)?[\w%./=?-]+\.(?:jpe?g|png|gif|tiff?)/gi;
                        var matches = content.match(regex);
                        if (matches)
                        { // If we have images
                            if (matches.length <= 10)
                            { // If we have <= 10 images
                                const messageWithoutMatches = content.replace(regex, ``); // Remove the matches from the message
                                if (messageWithoutMatches === ``)
                                { // Just Images
                                    channel.send({
                                        files: matches
                                    });
                                }
                                else
                                { // Message Content w/ Images
                                    channel.send({
                                        files: matches,
                                        content: messageWithoutMatches
                                    });
                                }
                            }
                        }
                        else
                            channel.send({ content: `${ content }` }); // Send the message if there are no images
                        sql.prepare(`DELETE FROM scheduledMessages WHERE guildId = ? AND channelId = ? AND content = ? AND time = ?`).run(guildId, channelId, databaseContent, time);
                    }
                }
            }
        }
    }, 15_000);
}
