import fetch from "node-fetch";
import { createRequire } from "node:module";
import fs from "node:fs";
import logging from "./logging.js";
import sqlite from "./sqlite.js";

const require = createRequire(import.meta.url);

const commands = [];
const elevatedCommands = [];
const whitelistedAltUsers = [];

const DISCORD_API_URL = `https://discord.com/api/v10`;

export default {
    commands,
    elevatedCommands,
    whitelistedAltUsers,
    DISCORD_API_URL,
    sleep,
    hasNoPingRole,
    checkName,
    checkAlt,
    deploySlashCommands,
    truncateString,
    canRunCommand,
    loadSubSC,
    getFriendlyTime,
    getEpoch,
    getGuildConfig,
    getAllGuildBans,
    getGuildMember,
    getUser,
    fetchJsonResponse,
    sendInteractionOrMessage
};

/**
 * @name sleep
 * @description Sleeps for a specified amount of time.
 * {@link https://stackoverflow.com/questions/951021/what-is-the-javascript-version-of-sleep StackOverflow}
 * @param {number} ms The amount of time to sleep for in ms.
 */
async function sleep(ms)
{
    // eslint-disable-next-line no-promise-executor-return
    return new Promise((resolve) => setTimeout(resolve, ms));
}

/**
 * @name hasNoPingRole
 * @description Checks if a user has the no ping role.
 * @param {string} userId The user ID to check.
 * @param {string} guildId The guild ID we're in.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 * @returns {boolean} Whether or not the user has the no ping role.
 */
async function hasNoPingRole(userId, guildId, Discord, client)
{
    const guild = client.guilds.cache.get(guildId);
    const { noPingRoleId } = getGuildConfig(guildId);
    if (!noPingRoleId)
        return false;

    const noPingRole = guild.roles.cache.find((r) => r.id === noPingRoleId);
    if (!noPingRole)
        return false;
    const repliedUser = await getGuildMember(client, guildId, userId); // GuildMemberManager#fetch caches the user.

    let isNoPingRole = false;
    repliedUser.roles.cache.each((role) =>
    {
        if (role.id === noPingRole.id)
            isNoPingRole = true;
    });

    if (isNoPingRole)
        return true;
    return false;
}

/**
 * @name checkName
 * @description Checks if a name is valid.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 * @param {Discord.GuildMember} member The member object.
 */
async function checkName(Discord, client, member)
{
    const guildId = member.guild.id;
    const { renameUntaggableNames } = getGuildConfig(guildId);
    // If the username does not start with 2 letters of ASCII, force name change.
    // eslint-disable-next-line no-control-regex
    const regex = /^[\d\u0000-\u007F\u00C0-\u00FF]{2,}/i;
    if (renameUntaggableNames && regex.test(member.user.username) === false && !member.nickname)
    {
        const { renameTo } = getGuildConfig(guildId);
        const reason = `${ global.locale[global.findLocale(member.guild.id)].untaggableNameError }`;
        // Only rename if they don't have a nickname.
        member.setNickname(`${ renameTo }`, `${ reason }`);
    }
}

/**
 * @name checkAlt
 * @description Checks if a user is an alt / a new account.
 * @param {Discord.GuildMember} member The member object.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function checkAlt(member, Discord, client)
{
    const guildId = member.guild.id;

    const { auditChannelID, altDetectionDays, banAppealsLink } = getGuildConfig(guildId);
    if (!auditChannelID || auditChannelID.length < 10) return;

    const auditChannel = client.channels.cache.get(auditChannelID);
    if (!auditChannel) return; // auditChannelID is not a valid channel.

    if (altDetectionDays < 0) return; // -1 is disabled.

    // this should ONLY EVER BE TRUE if they solved a captcha and we used their token to get them back
    if (whitelistedAltUsers.includes(member.user.id))
        return;

    // Check how old a users account is.
    const accountAge = Math.floor((Date.now() - member.user.createdTimestamp) / (1000 * 60 * 60 * 24));

    // Ban with reasoning
    if (accountAge < altDetectionDays && banAppealsLink && banAppealsLink.length > 1)
    {
        const userId = member.user.id;

        const message = `${ global.locale[global.findLocale(member.guild.id)].alertBold } <@!${ userId }> ${ global.locale[global.findLocale(member.guild.id)].hasBeenCaptchaBanned.replaceAll(`%count%`, `${ altDetectionDays }`) }`;
        const banMessage = `${ global.locale[global.findLocale(member.guild.id)].captchaBanDm.replaceAll(`%count%`, `${ altDetectionDays }`).replaceAll(`%appealLink%`, `${ banAppealsLink }`) }`;

        const timeForMuteMeta = `${ global.locale[global.findLocale(guildId)].permanently }`;

        await member.send(banMessage).catch((error) => logging.log(`Could not DM, ignoring\n${ error }`));

        const reason = `Auto Mod: New account age detected.`;

        await member.ban({ deleteMessageSeconds: 7 * 24 * 60 * 60, reason })
            .then(() => auditChannel.send(message));

        var newReason = `🔨 ${ reason }`;
        await sqlite.modifyPunishmentsProfile(client, userId, client.user.id, guildId, `bans`, 1, newReason, timeForMuteMeta);
    }
}

/**
 * @name deploySlashCommands
 * @description Deploys slash commands to the guild.
 * @param {Discord.Guild} guild The guild object.
 */
async function deploySlashCommands(guild)
{
    // For this guild, see if reporting is enabled.
    const guildId = guild.id;
    const { reportApplet } = getGuildConfig(guildId);
    if (!reportApplet)
    { // If reporting is disabled, don't deploy the applet.
        const index = commands.findIndex((c) => c.name === `Report`);
        if (index > -1)
            commands.splice(index, 1);
    }
    if (global.elevatedServers.includes(guild.id))
    {
        const allCommands = [...commands, ...(Array.isArray(elevatedCommands) ? elevatedCommands : [elevatedCommands])]; // Both "global" (not actually) and elevated commands.
        // Set slash commands for all elevated servers
        guild.commands.set(allCommands)
            .catch((error) =>
            {
                console.error(`${ global.locale[global.gLocale].errorSettingSlashCommands } ${ guild.name } (${ guild.id }).`, error);
            });
    }
    else
    {
        guild.commands.set(commands)
            .catch((error) =>
            {
                console.error(`${ global.locale[global.gLocale].errorSettingSlashCommands } ${ guild.name } (${ guild.id }).`, error);
            });
    }
}

/**
 * @name truncateString
 * @description Truncates a string to fit in 1024 characters.
 * @param {string} item The string to truncate.
 * @param {number} length The length to truncate to. (Including the ellipsis)
 * @returns {string} The truncated string.
 */
function truncateString(item, length)
{
    const ellipsis = ` [...]`;
    if (item.length > length - ellipsis.length)
    {
        item = item.slice(0, length - ellipsis.length);
        item += ellipsis;
        return item;
    }
    return item;
}

/**
 * @name canRunCommand
 * @description Checks if a user has permission to run a command.
 * @param {*} interactionOrMessage The interaction or message object.
 * @param {Discord.PermissionResolvable} permission The permission to check.
 * @param {boolean} isSlashCommand Whether or not the command being ran is a slash command.
 * @returns {boolean} Whether or not the user has permission to run the command.
 */
function canRunCommand(interactionOrMessage, permission, isSlashCommand = false)
{
    if (isSlashCommand)
    {
        if (interactionOrMessage.guild.ownerID !== interactionOrMessage.user.id && !interactionOrMessage.member.permissions.has(permission))
            return false;
        return true;
    }

    if (interactionOrMessage.guild.ownerID !== interactionOrMessage.author.id && !interactionOrMessage.member.permissions.has(permission))
        return false;
    return true;
}

/**
 * @name loadSubSC
 * @description Loads a sub-folder for Slash Commands.
 * @param {string} __dirname The directory of the file being called.
 * @param {string} name The name of the sub-folder/command.
 * @param {Array} options The options for the command.
 */
async function loadSubSC(__dirname, name, options)
{
    const subCommandFolder = fs.readdirSync(`${ __dirname }/${ name }/`);
    for (const subCommand of subCommandFolder)
    {
        let cmd;
        cmd = await (process.platform === `win32` ? import(`file://${ __dirname }/${ name }/${ subCommand }`) : import(`${ __dirname }/${ name }/${ subCommand }`));
        cmd = cmd.default || cmd;
        const data = {
            name: cmd.name,
            description: cmd.description,
            type: cmd.type,
            options: cmd.options || undefined,
            callback: cmd.callback
        };

        options.push(data);
    }
    return options;
}

/**
 * @name getFriendlyTime
 * @description Take Epoch Time and return a friendly time. (e.g. 1d, 1h, 1m, 1s)
 * @param {number} time The time in Epoch format.
 * @returns {string} The time in a friendly format.
 */
function getFriendlyTime(time)
{
    const timeDiff = Math.abs(Date.now() - time);
    if (timeDiff <= 60_000)
        return `${ Math.round(timeDiff / 1000) }s`;
    if (timeDiff <= 3_600_000)
        return `${ Math.round(timeDiff / 60_000) }m`;
    if (timeDiff <= 86_400_000)
        return `${ Math.round(timeDiff / 3_600_000) }h`;
    if (timeDiff <= 604_800_000)
        return `${ Math.round(timeDiff / 86_400_000) }d`;
    if (timeDiff <= 2_592_000_000)
        return `${ Math.round(timeDiff / 604_800_000) }w`;
    if (timeDiff <= 31_104_000_000)
        return `${ Math.round(timeDiff / 2_592_000_000) }mo`;
    return `${ Math.round(timeDiff / 31_104_000_000) }y`;
}

/**
 * @name getTime
 * @description Take a friendly time and return the Epoch time.
 * @param {string} time The time in a friendly format.
 * @param {boolean} addDateNow Whether or not to add the current date to the time.
 * @returns {number} The time in Epoch format.
 */
async function getEpoch(time, addDateNow)
{
    const regex = /^(\d+y)?(\d+mo)?(\d+w)?(\d+d)?(\d+h)?(\d+m)?(\d+s)?$/;
    if (regex.test(time)) // If it passes our regex:
    {
        let totalEpoch = 0;
        const groups = time.match(regex); // Split each matching group.
        groups.shift(); // Remove first object (the whole string)
        for (const group of groups)
        {
            if (group !== undefined)
            {
                // Carve out the number from the group.
                const number = group.match(/\d+/);
                // Get the unit from the group.
                const unit = group.replace(number, ``);
                const epoch = await findTime(number, unit);
                totalEpoch += Number.parseInt(epoch, 10);
            }
        }
        if (addDateNow)
            totalEpoch += Date.now();
        if (totalEpoch !== -1337)
            return totalEpoch;
        return -1337;
    }
    return -1337;
}

/**
 * @name findTime
 * @description Take a number and a unit and return the Epoch time.
 * @param {number} time The number to convert.
 * @param {string} timeUnit The unit to convert.
 * @returns {number} The time in Epoch format.
 */
async function findTime(time, timeUnit)
{
    switch (timeUnit)
    {
        case `y`: {
            time = time * 31_536_000 * 1000;
            break;
        }
        case `mo`: {
            time = time * 30 * 24 * 60 * 60 * 1000;
            break;
        }
        case `w`: {
            time = time * 7 * 24 * 60 * 60 * 1000;
            break;
        }
        case `d`: {
            time = time * 24 * 60 * 60 * 1000;
            break;
        }
        case `h`: {
            time = time * 60 * 60 * 1000;
            break;
        }
        case `m`: {
            time = time * 60 * 1000;
            break;
        }
        case `s`: {
            time *= 1000;
            break;
        }
        default: {
            time = -1337;
        }
    }
    return time;
}

/**
 * @name getGuildConfig
 * @description Returns the configuration for the guild.
 * @param {string} id The guild ID to check
 * @returns {object} The configuration
 */
function getGuildConfig(id)
{
    const guildConfig = global.getConfiguration.get(id);
    const GUILD_CONFIG = JSON.parse(guildConfig.config);
    return GUILD_CONFIG;
}

/**
 * @name getAllGuildBans
 * @description Returns an array of IDs of all banned users.
 * @param {object} guild The guild to check.
 * @returns {Array} The array of banned users IDs.
 */
async function getAllGuildBans(guild)
{
    const { token } = require(`../config.json`);
    const bansArray = [];
    await guild.bans.fetch().then(async (bans) =>
    {
        const temporaryArray = [...bans.keys()];
        if (temporaryArray.length < 1000)
        {
            for (const userId of temporaryArray)
                bansArray.push(userId);
        }
        else if (temporaryArray.length >= 1000)
        { // Is 1000 or more (unknown amount)
            // push the first 1000
            for (const userId of temporaryArray)
                bansArray.push(userId);
            var lastIdInList = `${ temporaryArray[temporaryArray.length - 1] }`;
            for (; ;) // inf. loop
            {
                const apiResult = await fetchJsonResponse(
                    `${ DISCORD_API_URL }/guilds/${ encodeURIComponent(guild.id) }/bans?after=${ lastIdInList }`,
                    {
                        method: `GET`,
                        headers: {
                            Authorization: `Bot ${ token }`,
                            'Content-Type': `application/json`
                        }
                    }
                );

                for (const ban of apiResult)
                    bansArray.push(ban.user.id);
                if (apiResult.length < 1000) // Exit inf loop
                    break;
                lastIdInList = `${ apiResult[apiResult.length - 1].user.id }`;
            }
        }
    });
    // make a new set to avoid duplicates
    const bansSet = new Set(bansArray);
    return [...bansSet];
}

/**
 * @param {Discord.Client} client The client
 * @param {string} guildId ID of target guild
 * @param {string} memberId ID of target member
 */
async function getGuildMember(client, guildId, memberId)
{
    const guild = client.guilds.cache.get(guildId);

    let invalidMember = false;
    const member = await guild.members.fetch(memberId)
        .catch(() =>
        {
            invalidMember = true;
        });

    if (invalidMember)
        return false;
    return member;
}

/**
 * @param {Discord.Client} client The client
 * @param {string} memberId ID of target member
 */
async function getUser(client, memberId)
{
    let invalidUser = false;
    const member = await client.users.fetch(memberId)
        .catch(() =>
        {
            invalidUser = true;
        });

    if (invalidUser)
        return false;
    return member;
}

/**
 *
 * @param {string} url URL to fetch for response
 * @param {object} data Data to pass to fetch
 */
async function fetchJsonResponse(url, data)
{
    const response = await fetch(url, data)
        .catch((error) => logging.error(error, false));

    const json = await response.json()
        .catch((error) => logging.error(`Could not reach target URL ${ url } (${ error })`));

    if (json)
        return json;
    return false;
}

/**
 * @name sendInteractionOrMessage
 * @description Determines whether or not to send a interaction or message.
 * @param {*} interactionOrMessage The object containing the interaction/message
 * @param {object} options Options for the message
 * @param {boolean} isSlashCommand Whether or not the object is a interaction or message
 * @param {boolean} ephemeral Whether or not the interaction is a ephemeral
 * @returns {*} A message, either a reply or in channel.
 */
function sendInteractionOrMessage(interactionOrMessage, options, isSlashCommand, ephemeral = true)
{
    if (!isSlashCommand)
        return interactionOrMessage.channel.send(options);
    if (isSlashCommand && !ephemeral)
        return interactionOrMessage.reply(options);
    return interactionOrMessage.followUp(options);
}

global.findLocale = function (guildId, specifiedLocale)
{
    let finalLocale = ``;

    if (specifiedLocale)
    {
        // https://discord.com/developers/docs/reference#locales
        if (specifiedLocale.includes(`-`))
            specifiedLocale = specifiedLocale.split(`-`)[0];
        finalLocale = specifiedLocale;
    }
    else
    {
        const guildConfig = getGuildConfig(guildId);
        finalLocale = guildConfig.locale;
    }

    if (!global.locale[finalLocale])
        finalLocale = `en`; // Default to English if locale is not found.
    return finalLocale;
};

global.findLocaleString = function (specifiedLocale, word)
{
    return global.locale[specifiedLocale][word];
};
