import { ButtonStyle, ChannelType } from "discord.js";
import unidecode from "unidecode-plus";
import SQLite from "better-sqlite3";

import ban from "../commands/commands/moderation/ban.js";
import mute from "../commands/commands/moderation/mute.js";
import warn from "../commands/commands/moderation/warn.js";
import logging from "./logging.js";
import sqlite from "./sqlite.js";
import common from "./common.js";
import debug from "../commands/chat_commands/utility/debug.js";

reactionMessageIds = global.reactionMessageIds;
const sql = new SQLite(`./src/db.sqlite`);

export default {
    messageDetection,
    areTheySpammingKeywords
};

/**
 * @name messageDetection
 * @description Various Auto Moderation checks. (Bad Words, Links, Un-necessary Pings, etc.)
 * @param {Discord.Message} message The message object.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Client Object.
 */
async function messageDetection(message, Discord, client)
{
    if (message.channel.type === ChannelType.DM)
        return; // Don't check DMs.
    if (message.author.bot)
        return;
    if (message.author.id === message.guild.ownerId && (!debug.isDebugUser(message.author.id)))
        return; // Don't chump-check the owner.

    const {
        regionalChannelID, mutedRoleID, deletedMessagesChanID, vipRoleID, allowAnyoneWithRoleToSendLinks,
        allowAnyoneWithRoleToBypassBadWordBanning, allowAnyoneWithRoleToBypassTriggerResponse, noPingRoleId,
        sendDmOnBan, banAppealsLink, autoWarnForAmountOfMentions, allowAnyoneToSendStickers, autoModCaps, amountOfSpoilers,
        allowDenyMode, autoModMode, autoBanMode
    } = common.getGuildConfig(message.guild.id);

    // get vip role
    const vipRole = message.guild.roles.cache.find((r) => r.id === vipRoleID);
    if (!vipRole)
        return;

    // Sticker Detection: If the message contains a sticker, and the server has "Allow Anyone To Send Stickers" disabled, delete it, unless they're VIP.
    if (!allowAnyoneToSendStickers && message.stickers.size > 0 && (vipRole && message.member.roles.highest.position < vipRole.position))
        return message.delete();

    if (banAppealsLink)
        var nitroSafetyMessage = `${ global.locale[global.findLocale(message.guild.id)].nitroSafetyMessage.replaceAll(`%banAppealsLink%`, `${ banAppealsLink }`) }`;

    if (message.member && message.member.guild && message.member.guild.id)
        await mute.checkLTAPStatus(Discord, client, message.member.guild.id, message.member);

    // Anti-Zalgo Handling
    // [̀-ͯ᪰-᫿᷀-᷿⃐-⃿︠-︯]
    // https://codegolf.stackexchange.com/a/119313
    // eslint-disable-next-line no-misleading-character-class
    const regex = /[̀-ͯ᪰-᫿᷀-᷿⃐-⃿︠-︯]/umi;
    const reason = `${ global.locale[global.findLocale(message.guild.id)].containsZalgo }`;
    if ((regex.test(message.content))
        && (debug.isDebugUser(message.author.id)
            || (allowAnyoneWithRoleToBypassBadWordBanning && message.member.roles.highest.position < 1)
            || (!allowAnyoneWithRoleToBypassBadWordBanning && message.member.roles.highest.position < vipRole.position)))
    {
        message.delete({ reason });

        const embed = new Discord.EmbedBuilder()
            .setColor(`#ff1e00`)
            .setTimestamp()
            .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(message.guild.id)].zalgoDeleted })` })
            .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].user }`, value: ` <@!${ message.member.id }> (${ message.member.user.username })`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].channel }`, value: ` <#${ message.channel.id }>`, inline: true }])
            .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].content }`, value: common.truncateString(message.content, 1024), inline: false }]);

        const delMessageChan = client.channels.cache.get(deletedMessagesChanID);
        if (delMessageChan)
            delMessageChan.send({ embeds: [embed] });
        return;
    }

    if (!common.canRunCommand(message, `KickMembers`))
        message.content = unidecode(message.content); // Convert funky characters to ASCII.

    if (message.member.roles.highest.position < vipRole.position // Let VIPs ping people with "No Mentions" role.
        // Did the user ping without replying? (more intentional than reply)
        && message.mentions.users && message.mentions.users.size > 0 && message.mentions.repliedUser === null)
    {
        for (let index = 0; index < message.mentions.users.size; index++)
        {
            // doing some weird stuff to use the Collection's values and make it a Array so i can iterate through :P
            const member = [...message.mentions.users.values()][index];
            if (!noPingRoleId || noPingRoleId.length === 0)
                return;
            const result = await common.hasNoPingRole(member.id, message.guild.id, Discord, client);
            if (result)
            {
                const reason = `${ global.locale[global.findLocale(message.guild.id)].autoModPingNoMentions.replaceAll(`%user-and-discrim%`, `${ member.username }#${ member.discriminator }`) }`;
                message.delete({ reason });
                if (autoModMode === `warn`)
                    return warn.warnUser(message, true, false, message.author.id, client.user.id, reason, message.content, Discord, client);
                await sendAutoModeratorDeleteMessage(message, deletedMessagesChanID, reason, Discord, client);
                return;
            }
        }
    }

    // Auto Banned Phrases Handling
    if (!common.canRunCommand(message, `ManageMessages`))
    {
        const bannedPhrasesH = sql.prepare(`SELECT phrase FROM autoBannedPhrases WHERE guildId = ${ message.guild.id } `).raw().all();
        const bannedPhrases = bannedPhrasesH.flat();
        if (bannedPhrases.length > 0) // Don't break the bot if there are no banned phrases.
        {
            let messageContentWithoutDomains = message.content;
            const domainsRegex = /(http:\/\/|https:\/\/|www\.)?([\da-z][\da-z-]*\.)+[\da-z][\da-z-]*/gi;
            const domains = message.content.match(domainsRegex);
            if (domains)
            {
                for (const domain of domains)
                    messageContentWithoutDomains = messageContentWithoutDomains.replace(domain, ``);
            }

            let doBan = false;
            for (const bannedPhrase of bannedPhrases)
            {
                if (messageContentWithoutDomains.toLowerCase().includes(bannedPhrase.toLowerCase()))
                {
                    if (!message.member.bannable)
                        return;

                    if (autoBanMode === `phrase`)
                        doBan = true;

                    const messageContainsLink = message.content.toLowerCase().includes(`https`)
                        || message.content.toLowerCase().includes(`http`)
                        || message.content.toLowerCase().includes(`www.`)
                        || message.content.toLowerCase().includes(`discord.gg`);

                    if (autoBanMode === `link`
                        && messageContainsLink
                        && message.author.id !== client.user.id)
                        doBan = true;

                    if (autoBanMode === `all`
                        && messageContainsLink
                        && (message.content.toLowerCase().includes(`@everyone`)
                            || message.content.toLowerCase().includes(`@here`))
                        && message.author.id !== client.user.id)
                        doBan = true;

                    if (doBan)
                    {
                        if ((sendDmOnBan) && banAppealsLink && banAppealsLink.length > 1)
                            await message.member.send(nitroSafetyMessage).catch(() => logging.warn(`Could not DM, ignoring`));

                        const unmodified = message.content;
                        let safeString = unmodified.replace(/http:\/\//gi, `hxxp://`).replace(/https:\/\//gi, `hxxps://`);
                        safeString = common.truncateString(safeString, 1024);

                        let reason = `${ global.locale[global.findLocale(message.guild.id)].autoModBannedPhrase }`;
                        const unmodifiedReason = reason;
                        reason = `${ reason }\n\`\`\`${ safeString }\n\`\`\``;
                        ban.banAndSendEmbed(Discord, client, message.member.id, message.member.user.username, client.user.id, reason, message.member);

                        const newReason = `🔨 ${ unmodifiedReason }`;
                        await sqlite.modifyPunishmentsProfile(client, message.member.id, client.user.id, message.guild.id, `bans`, 1, newReason);
                        return;
                    }
                }
            }
        }
    }

    // Anti-@everyone / @here Handling
    if ((message.content.toLowerCase().includes(`@everyone`)
        || message.content.toLowerCase().includes(`@here`))
        && !common.canRunCommand(message, `ManageMessages`))
    {
        if (message.content.toLowerCase().includes(`http`)
            || message.content.toLowerCase().includes(`https`)
            || message.content.toLowerCase().includes(`www.`)
            || message.content.toLowerCase().includes(`discord.gg`))
        { // If msg contains: @everyone + a link -> 7d mute.
            const time = 7 * 24 * 60 * 60 * 1000; // 7d
            const reason = `${ global.locale[global.findLocale(message.guild.id)].autoModEveryoneAndLink }`;
            const authorTag = client.user.tag;
            const isExtending = false;
            const mutedRole = message.guild.roles.cache.find((r) => r.id === mutedRoleID);
            const unmodified = message.content;
            let safeString = unmodified.replace(/http:\/\//gi, `hxxp://`).replace(/https:\/\//gi, `hxxps://`);
            safeString = common.truncateString(safeString, 1024);
            message.delete({ reason });
            if (!mutedRole)
                return;
            await mute.sendMuteToDB(client, message.guild.id, client.user.id, message.author.id, Date.now() + time, reason, `7d`);
            mute.handleMute(Discord, client, message.member, client.user.id, mutedRole, authorTag, `${ reason }\`\`\`\n${ safeString }\n\`\`\``, time, isExtending, message.member.guild.id, message.member.user.id, true); // handles rest of mute process
            return;
        }

        // If it's just an @everyone, mute for 1d.
        const time = 1 * 24 * 60 * 60 * 1000; // 1d
        const reason = `${ global.locale[global.findLocale(message.guild.id)].autoModEveryone }`;
        const authorTag = client.user.tag;
        const isExtending = false;
        const mutedRole = message.guild.roles.cache.find((r) => r.id === mutedRoleID);
        const unmodified = message.content;
        let safeString = unmodified.replace(/http:\/\//gi, `hxxp://`).replace(/https:\/\//gi, `hxxps://`);
        safeString = common.truncateString(safeString, 1024);
        message.delete({ reason });
        if (!mutedRole)
            return;
        await mute.sendMuteToDB(client, message.guild.id, client.user.id, message.author.id, Date.now() + time, reason, `1d`);
        mute.handleMute(Discord, client, message.member, client.user.id, mutedRole, authorTag, `${ reason }\`\`\`\n${ safeString }\n\`\`\``, time, isExtending, message.member.guild.id, message.member.user.id, true); // handles rest of mute process
        return;
    }

    // Discord Link Handling
    // check if discord is valid and allowed
    if ((message.content.toLowerCase().includes(`discord.gg/`)
        || message.content.toLowerCase().includes(`discord.com/`)
        || message.content.toLowerCase().includes(`discordapp.com/`))
        && message.author.id !== client.user.id && (debug.isDebugUser(message.author.id)
            || (allowAnyoneWithRoleToSendLinks && message.member.roles.highest.position < 1)
            || (!allowAnyoneWithRoleToSendLinks && message.member.roles.highest.position < vipRole.position)))
    {
        let found = 0;
        const discordInvite = await client.fetchInvite(message.content).catch((error) => logging.error(`${ global.locale[global.gLocale].isThisAValidInvite }\n${ error }`));
        if (discordInvite)
        { // If a valid invite
            // get entries for whitelisted discord invites
            const allowedDiscords = sql.prepare(`SELECT id FROM allowedDiscords WHERE guildId = ${ message.guild.id }`).raw().all();

            for (const allowedDiscord of allowedDiscords)
            {
                // If allowedDiscord array contains discordInvite.guild.id
                if (allowedDiscord.toString() === discordInvite.guild.id)
                {
                    found = 1;
                    break;
                }
            }

            if ((found === 0 && allowDenyMode === `allow`) || (found === 1 && allowDenyMode === `deny`))
            {
                message.delete({ reason: message.content });
                const unmodified = message.content;
                let safeString = unmodified.replace(/http:\/\//gi, `hxxp://`).replace(/https:\/\//gi, `hxxps://`);
                safeString = common.truncateString(safeString, 1024);
                const embed = new Discord.EmbedBuilder()
                    .setColor(`#ff1e00`)
                    .setTimestamp()
                    .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(message.guild.id)].discordInviteDeleted })` })
                    .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].user }`, value: `<@!${ message.member.id }> (${ message.member.user.username })`, inline: true }])
                    .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].channel }`, value: `<#${ message.channel.id }>`, inline: true }])
                    .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].content }`, value: safeString, inline: false }]);

                const banButton = new Discord.ButtonBuilder()
                    .setCustomId(`ban-${ message.author.id }`)
                    .setLabel(`${ global.locale[global.findLocale(message.guild.id)].ban }`)
                    .setStyle(ButtonStyle.Danger)
                    .setDisabled(false);
                const banScamButton = new Discord.ButtonBuilder()
                    .setCustomId(`banScam-${ message.author.id }`)
                    .setLabel(`${ global.locale[global.findLocale(message.guild.id)].banScam }`)
                    .setStyle(ButtonStyle.Danger)
                    .setDisabled(false);
                const autoModerationAdd = new Discord.ButtonBuilder()
                    .setCustomId(`autoModerationAdd-${ message.author.id }`)
                    .setLabel(`${ global.locale[global.findLocale(message.guild.id)].autoModerationAdd }`)
                    .setStyle(ButtonStyle.Danger)
                    .setDisabled(false);
                const buttonArray = [banButton, banScamButton, autoModerationAdd];
                const row = new Discord.ActionRowBuilder()
                    .addComponents(buttonArray);

                const delMessageChan = client.channels.cache.get(deletedMessagesChanID);
                if (delMessageChan)
                {
                    delMessageChan.send({
                        embeds: [embed],
                        components: [row]
                    });
                }
                return;
            }
        }
    }

    // Link Handling
    const approvedLinksH = sql.prepare(`SELECT link FROM allowedLinks WHERE guildId = ${ message.guild.id }`).raw().all();
    const approvedLinks = approvedLinksH.flat();
    let found = 0;
    if ((message.content.toLowerCase().includes(`https`)
        || message.content.toLowerCase().includes(`http`)
        || message.content.toLowerCase().includes(`www.`))
        && message.author.id !== client.user.id && (debug.isDebugUser(message.author.id)
            || (allowAnyoneWithRoleToSendLinks && message.member.roles.highest.position < 1)
            || (!allowAnyoneWithRoleToSendLinks && message.member.roles.highest.position < vipRole.position)))
    {
        // Check if it was a discord invite, and return if so (we handled this earlier)
        const domainsRegex = /(http:\/\/|https:\/\/|www\.)?([\da-z][\da-z-]*\.)+[\da-z][\da-z-]*/gi;
        const domainsInMessage = message.content.match(domainsRegex);
        // remove any discord links from domains
        for (const domain of domainsInMessage)
        {
            if (domain.includes(`discord.gg`) || domain.includes(`discord.com`) || domain.includes(`discordapp.com`))
                domainsInMessage.splice(domainsInMessage.indexOf(domain), 1);
        }
        if (approvedLinks.length > 0 && domainsInMessage.length > 0)
        { // Don't break the bot if there are no approved links.
            for (let approvedLink of approvedLinks)
            {
                approvedLink = approvedLink.replace(/http:\/\//gi, ``).replace(/https:\/\//gi, ``).replace(/www\./gi, ``); // Get raw domain only
                approvedLink = approvedLink.split(`?`)[0]; // Strip anything after ?
                approvedLink = approvedLink.split(`#`)[0]; // Strip anything after #
                approvedLink = approvedLink.replace(/\/$/, ``); // Strip last / if it exists
                for (let domain of domainsInMessage)
                {
                    domain = domain.replace(/http:\/\//gi, ``).replace(/https:\/\//gi, ``).replace(/www\./gi, ``); // Get raw domain only
                    domain = domain.split(`?`)[0]; // Strip anything after ?
                    domain = domain.split(`#`)[0]; // Strip anything after #
                    domain = domain.replace(/\/$/, ``); // Strip last / if it exists
                    if (domain.includes(approvedLink) || approvedLink === `*`)
                    {
                        found = 1;
                        break;
                    }
                }
            }
        }
        else if (domainsInMessage.length === 0)
            return; // We handle discord links earlier.

        // we did not find a match?
        if ((found === 1 && allowDenyMode === `deny`) // If we found something that's blacklisted, delete.
            || (found === 0 && allowDenyMode === `allow`)) // If we didn't find anything that's whitelisted, delete.
        {
            message.delete({ reason: message.content });
            const unmodified = message.content;
            let safeString = unmodified.replace(/http:\/\//gi, `hxxp://`).replace(/https:\/\//gi, `hxxps://`);
            safeString = common.truncateString(safeString, 1024);
            const embed = new Discord.EmbedBuilder()
                .setColor(`#ff1e00`)
                .setTimestamp()
                .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(message.guild.id)].linkDeleted })` })
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].user }`, value: `<@!${ message.member.id }> (${ message.member.user.username })`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].channel }`, value: `<#${ message.channel.id }>`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].content }`, value: safeString, inline: false }]);

            const banButton = new Discord.ButtonBuilder()
                .setCustomId(`ban-${ message.author.id }`)
                .setLabel(`${ global.locale[global.findLocale(message.guild.id)].ban }`)
                .setStyle(ButtonStyle.Danger)
                .setDisabled(false);
            const banScamButton = new Discord.ButtonBuilder()
                .setCustomId(`banScam-${ message.author.id }`)
                .setLabel(`${ global.locale[global.findLocale(message.guild.id)].banScam }`)
                .setStyle(ButtonStyle.Danger)
                .setDisabled(false);
            const autoModerationAdd = new Discord.ButtonBuilder()
                .setCustomId(`autoModerationAdd-${ message.author.id }`)
                .setLabel(`${ global.locale[global.findLocale(message.guild.id)].autoModerationAdd }`)
                .setStyle(ButtonStyle.Danger)
                .setDisabled(false);
            const buttonArray = [banButton, banScamButton, autoModerationAdd];
            const row = new Discord.ActionRowBuilder()
                .addComponents(buttonArray);

            const delMessageChan = client.channels.cache.get(deletedMessagesChanID);
            if (delMessageChan)
            {
                delMessageChan.send({
                    embeds: [embed],
                    components: [row]
                });
            }
            return;
        }
    }

    // Bad Word Handling
    const badWordsH = sql.prepare(`SELECT word FROM badWords WHERE guildId = ${ message.guild.id }`).raw().all();
    const badWords = badWordsH.flat();
    if (badWords.length > 0)
    {
        let foundBadWord;

        // search through each word in message
        const messageContentArray = message.content.split(` `);
        for (const word of messageContentArray)
        {
            // search through each word in bad words
            for (const badWord of badWords)
            {
                if (word.toLowerCase().includes(badWord.toLowerCase()) && badWord.length > (word.length * 0.51))
                { // If 51% of word is smaller than the banned word, delete.
                    foundBadWord = badWord;
                    break;
                }
            }
        }

        if (foundBadWord && (debug.isDebugUser(message.author.id)
            || (allowAnyoneWithRoleToBypassBadWordBanning && message.member.roles.highest.position < 1)
            || (!allowAnyoneWithRoleToBypassBadWordBanning && message.member.roles.highest.position < vipRole.position)))
        {
            message.delete({ reason: message.content });
            const unmodified = message.content;
            let safeString = unmodified.replace(/http:\/\//gi, `hxxp://`).replace(/https:\/\//gi, `hxxps://`);
            safeString = common.truncateString(safeString, 1024);

            const embed = new Discord.EmbedBuilder()
                .setColor(`#ff1e00`)
                .setTimestamp()
                .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(message.guild.id)].badWordDeleted })` })
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].user }`, value: `<@!${ message.member.id }> (${ message.member.user.username })`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].channel }`, value: `<#${ message.channel.id }>`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].word }`, value: `${ foundBadWord }`, inline: true }])
                .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].content }`, value: `${ safeString }`, inline: false }]);

            const banButton = new Discord.ButtonBuilder()
                .setCustomId(`ban-${ message.author.id }`)
                .setLabel(`${ global.locale[global.findLocale(message.guild.id)].ban }`)
                .setStyle(ButtonStyle.Danger)
                .setDisabled(false);
            const banScamButton = new Discord.ButtonBuilder()
                .setCustomId(`banScam-${ message.author.id }`)
                .setLabel(`${ global.locale[global.findLocale(message.guild.id)].banScam }`)
                .setStyle(ButtonStyle.Danger)
                .setDisabled(false);
            const autoModerationAdd = new Discord.ButtonBuilder()
                .setCustomId(`autoModerationAdd-${ message.author.id }`)
                .setLabel(`${ global.locale[global.findLocale(message.guild.id)].autoModerationAdd }`)
                .setStyle(ButtonStyle.Danger)
                .setDisabled(false);
            const buttonArray = [banButton, banScamButton, autoModerationAdd];
            const row = new Discord.ActionRowBuilder()
                .addComponents(buttonArray);

            const delMessageChan = client.channels.cache.get(deletedMessagesChanID);
            if (delMessageChan)
            {
                delMessageChan.send({
                    embeds: [embed],
                    components: [row]
                });
            }
            return;
        }
    }

    // Multiple Mentions Handling
    if (autoWarnForAmountOfMentions > 0 // Lets us set to -1 to disable.
        && (message.mentions.users.size >= autoWarnForAmountOfMentions)
        && (debug.isDebugUser(message.author.id)
            || message.member.roles.highest.position < vipRole.position))
    {
        const reason = `${ global.locale[global.findLocale(message.guild.id)].mentionedTooManyUsers }`;
        message.delete({ reason });
        if (autoModMode === `warn`)
            return warn.warnUser(message, true, false, message.author.id, client.user.id, reason, message.content, Discord, client);
        await sendAutoModeratorDeleteMessage(message, deletedMessagesChanID, reason, Discord, client);
        return;
    }

    // Caps Handling
    if (autoModCaps
        && message.content.length >= 20
        && (debug.isDebugUser(message.author.id)
            || message.member.roles.highest.position < vipRole.position))
    {
        // Check if 70% of the message is caps
        const caps = message.content.match(/[A-Z]/g);
        if (caps && caps.length >= (message.content.length * 0.7))
        {
            const reason = `${ global.locale[global.findLocale(message.guild.id)].autoModExcessiveCaps }`;
            message.delete({ reason });
            if (autoModMode === `warn`)
                return warn.warnUser(message, true, false, message.author.id, client.user.id, reason, message.content, Discord, client);
            await sendAutoModeratorDeleteMessage(message, deletedMessagesChanID, reason, Discord, client);
            return;
        }
    }

    // Excessive Spoilers Handling
    if (amountOfSpoilers > 0
        && (debug.isDebugUser(message.author.id)
            || message.member.roles.highest.position < vipRole.position))
    { // Lets us set to -1 to disable.
        const spoilers = message.content.match(/\|\|/g);
        if (spoilers && spoilers.length / 2 >= amountOfSpoilers)
        {
            const reason = `${ global.locale[global.findLocale(message.guild.id)].autoModExcessiveSpoilers }`;
            message.delete({ reason });
            if (autoModMode === `warn`)
                return warn.warnUser(message, true, false, message.author.id, client.user.id, reason, message.content, Discord, client);
            await sendAutoModeratorDeleteMessage(message, deletedMessagesChanID, reason, Discord, client);
            return;
        }
    }

    // Keyword Handling
    if ((debug.isDebugUser(message.author.id)
        || (allowAnyoneWithRoleToBypassTriggerResponse && message.member.roles.highest.position < 1)
        || (!allowAnyoneWithRoleToBypassTriggerResponse && message.member.roles.highest.position < vipRole.position)))
    { // Check for keywords only if you are not a bot, and you do not have a role.
        // If message is sent in a regional channel, check for keywords.

        let responseObject = sql.prepare(`SELECT * FROM keywords WHERE guildId = ${ message.guild.id }`).raw().all();

        if (regionalChannelID && message.channel.id === regionalChannelID)
            responseObject = sql.prepare(`SELECT * FROM keywordsRegional WHERE guildId = ${ message.guild.id }`).raw().all();

        const messageContentArray = message.content.split(` `);
        // search through each word in bad words
        for (const element of responseObject)
        {
            const triggerWord = element[1];
            if (!triggerWord.includes(` `))
            { // If keyword does not have a space, do the math to avoid things like vac triggering on vacation
                for (const word of messageContentArray)
                {
                    if (word.toLowerCase().includes(element[1].toLowerCase()) && element[1].length >= (word.length * 0.5))
                    { // If 50% of the word matches the trigger, reply.
                        return sendKeywordReply(message, element, Discord, client);
                    }
                }
            }
            else if (message.content.toLowerCase().includes(triggerWord.toLowerCase()))
            {
                // keyword has a space so i mean its probably intentional to trigger against
                return sendKeywordReply(message, element, Discord, client);
            }
        }
    }

    if (client.user.id === `992838421192118292` // Only allow easter eggs if we're using Chase's bot directly, just so people self-hosting don't have to deal with unintended code.
        && message.content.toLowerCase().includes(`<a:goodbot:890540504217972737>`)) // :goodbot: -> react w/ :cat_kiss:
        message.react(`<a:cat_kiss:861991427151888404>`);
}

/**
 *
 * @param {Discord.Message} message The message object that triggered the keyword.
 * @param {object} element The keyword object from the database.
 * @param {*} Discord The Discord Object.
 * @param {*} client The Discord Client.
 */
async function sendKeywordReply(message, element, Discord, client)
{
    await areTheySpammingKeywords(message, Discord, client);
    // React to Delete Bot Message 2
    const button = new Discord.ActionRowBuilder().addComponents(
        new Discord.ButtonBuilder()
            .setStyle(ButtonStyle.Danger)
            .setLabel(`${ global.locale[global.findLocale(message.guild.id)].delete }`)
            .setDisabled(false)
            .setCustomId(message.author.id)
    );
    message.reply({
        content: element[2],
        components: [button]
    })
        .then((message) =>
        {
            reactionMessageIds.push(message.id); // remember message ID because we react to it
        });
}

/**
 * @name areTheySpammingKeywords
 * @description Checks if a user is spamming keywords.
 * @param {Discord.Message} message The message object.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function areTheySpammingKeywords(message, Discord, client)
{
    const {
        amountOfKeywordSpam, timeToMuteForKeywordSpamInMinutes, timeBeforeRefreshingKeywordSpamInMinutes, mutedRoleID
    } = common.getGuildConfig(message.guild.id);

    if (amountOfKeywordSpam <= 0 || timeToMuteForKeywordSpamInMinutes <= 0 || timeBeforeRefreshingKeywordSpamInMinutes <= 0)
        return; // They have disabled this feature.

    // Are they already in the list we're checking against?
    const spammer = sql.prepare(`SELECT * FROM keywordSpam WHERE guildId = ? AND userId = ?`).get(message.guild.id, message.author.id);
    if (!spammer) // They are not; add them.
        return sql.prepare(`INSERT INTO keywordSpam (guildId, userId, count, time) VALUES (?, ?, 1, ?)`).run(message.guild.id, message.author.id, Date.now());

    if (spammer.time + (timeBeforeRefreshingKeywordSpamInMinutes * 60 * 1000) >= Date.now())
    { // If their current time + the time before refreshing is greater than the current time, reset their count.
        sql.prepare(`DELETE FROM keywordSpam WHERE guildId = ? AND userId = ?`).run(message.guild.id, message.author.id);
        return sql.prepare(`INSERT INTO keywordSpam (guildId, userId, count, time) VALUES (?, ?, 1, ?)`).run(message.guild.id, message.author.id, Date.now());
    }
    // Time is not up, so update their count.
    sql.prepare(`UPDATE keywordSpam SET count = ? WHERE guildId = ? AND userId = ?`).run(spammer.count + 1, message.guild.id, message.author.id);
    if (spammer.count + 1 >= amountOfKeywordSpam)
    { // If they've hit the amountOfKeywordSpam, mute them for amount specified in config.
        const time = timeToMuteForKeywordSpamInMinutes * 60 * 1000; // minutes to milliseconds
        const reason = `${ global.locale[global.findLocale(message.guild.id)].autoModSpammingKeywords }`;
        const mutedRole = message.guild.roles.cache.find((r) => r.id === mutedRoleID);
        if (!mutedRole) return;
        const friendlyTime = common.getFriendlyTime(time);
        await mute.sendMuteToDB(client, message.guild.id, client.user.id, message.author.id, Date.now() + time, reason, friendlyTime);
        mute.handleMute(Discord, client, message.member, client.user.id, mutedRole, client.user.tag, `${ reason }`, time, false, message.member.guild.id, message.member.user.id); // handles rest of mute process
        // Remove their entry since they are now muted.
        sql.prepare(`DELETE FROM keywordSpam WHERE guildId = ? AND userId = ?`).run(message.guild.id, message.author.id);
    }
}

/**
 * @name sendAutoModeratorDeleteMessage
 * @description Sends a log to the deleted messages channel when we moderate a message.
 * @param {Discord.Message} message The message object.
 * @param {string} deletedMessagesChanID The ID of the deleted messages channel.
 * @param {string} reason The reason for the deletion.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
async function sendAutoModeratorDeleteMessage(message, deletedMessagesChanID, reason, Discord, client)
{
    const embed = new Discord.EmbedBuilder()
        .setColor(`#ff1e00`)
        .setTimestamp()
        .setAuthor({ name: `🗑️ (${ global.locale[global.findLocale(message.guild.id)].autoModDeletion })` })
        .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].user }`, value: ` <@!${ message.author.id }> (${ message.member.user.username })`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].channel }`, value: ` <#${ message.channel.id }>`, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].reason }`, value: reason, inline: true }])
        .addFields([{ name: `${ global.locale[global.findLocale(message.guild.id)].content }`, value: common.truncateString(message.content, 1024), inline: false }]);

    const banButton = new Discord.ButtonBuilder()
        .setCustomId(`ban-${ message.author.id }`)
        .setLabel(`${ global.locale[global.findLocale(message.guild.id)].ban }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(false);
    const banScamButton = new Discord.ButtonBuilder()
        .setCustomId(`banScam-${ message.author.id }`)
        .setLabel(`${ global.locale[global.findLocale(message.guild.id)].banScam }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(false);
    const autoModerationAdd = new Discord.ButtonBuilder()
        .setCustomId(`autoModerationAdd-${ message.author.id }`)
        .setLabel(`${ global.locale[global.findLocale(message.guild.id)].autoModerationAdd }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(false);
    const buttonArray = [banButton, banScamButton, autoModerationAdd];
    const row = new Discord.ActionRowBuilder()
        .addComponents(buttonArray);

    const delMessageChan = client.channels.cache.get(deletedMessagesChanID);
    if (delMessageChan)
    {
        delMessageChan.send({
            embeds: [embed],
            components: [row]
        });
    }
}
