import SQLite from "better-sqlite3";

const sql = new SQLite(`./src/db.sqlite`);

export default {
    modifyPunishmentsProfile,
    getPunishmentData,
    hasRecord
};

/**
 * @name newPunishmentsProfile
 * @description Creates a new punishments profile from a user ID.
 * @param {string} guildAndUserId The guild and user ID to create a punishments profile for.
 * @param {string} guildId The guild ID we're in.
 * @param {string} id The user ID to create a punishments profile for.
 */
async function newPunishmentsProfile(guildAndUserId, guildId, id)
{
    let data = global.getInfoById.get(guildId, id);
    if (data)
        return;

    // doesn't exist, lets create!
    data = {
        guildAndUserId,
        guildId,
        id,
        warns: 0,
        mutes: 0,
        kicks: 0,
        bans: 0,
        punishmentreason: `[]`,
        punishmentinfo: `[]`
    };
    global.setInfo.run(data);
}

/**
 * @name getPunishmentData
 * @description Returns the punishment count for a specific type for a user.
 * @param {string} guildId The guild ID we're in.
 * @param {string} id The user ID to get punishment data for.
 * @param {string} type The punishment type to get data for.
 * @returns {number} The punishment count for the user & specific type.
 */
function getPunishmentData(guildId, id, type)
{
    const data = global.getInfoById.get(guildId, id);
    if (!data)
        return 0;

    switch (type)
    {
        case `warns`: {
            return data.warns;
        }
        case `mutes`: {
            return data.mutes;
        }
        case `kicks`: {
            return data.kicks;
        }
        case `bans`: {
            return data.bans;
        }
        default:
    }
}

/**
 * @name modifyPunishmentsProfile
 * @description Modifies the punishment data for a user.
 * @param {*} client The client object.
 * @param {string} id The user ID to modify punishment data for.
 * @param {string} moderatorId The moderator ID who modified the punishment data.
 * @param {string} guildId The guild ID we're in.
 * @param {string} type The punishment type to modify data for. (warns/mutes/kicks/bans)
 * @param {number} amount The amount to modify the punishment type by. (Usually 1 or -1)
 * @param {string} reason The reason for the punishment.
 * @param {string} timeForMuteMeta Time in friendly format for the punishment meta. (e.g. 1d/1h/etc)
 */
async function modifyPunishmentsProfile(client, id, moderatorId, guildId, type, amount, reason, timeForMuteMeta)
{
    let data = global.getInfoById.get(guildId, id);
    if (!data)
    {
        const guildAndUserId = `${ guildId }_${ id }`;
        await newPunishmentsProfile(guildAndUserId, guildId, id);
    }

    data = global.getInfoById.get(guildId, id); // need to get again

    if ((!timeForMuteMeta) && (type === `mutes` || type === `bans`))
        var timeForMuteMeta = `${ global.locale[global.findLocale(guildId)].permanently }`;
    if (timeForMuteMeta)
        var info = `<@!${ moderatorId }> ${ global.locale[global.findLocale(guildId)].at } <t:${ Math.round(Date.now() / 1000) }> ${ timeForMuteMeta }`;
    else
        var info = `<@!${ moderatorId }> ${ global.locale[global.findLocale(guildId)].at } <t:${ Math.round(Date.now() / 1000) }>`;
    if (!moderatorId)
        var info = `${ global.locale[global.findLocale(guildId)].unknownMetaSQL }`;

    // make JSON array nice, then parse it to recognize the object
    let punishReason = JSON.parse(data.punishmentreason);
    if (reason.length > 250)
        reason = reason.slice(0, 250);
    punishReason.push(reason);
    punishReason = JSON.stringify(punishReason);
    let punishInfo = JSON.parse(data.punishmentinfo);
    punishInfo.push(info);
    punishInfo = JSON.stringify(punishInfo);

    data = {
        guildAndUserId: `${ data.guildId }_${ id }`,
        guildId: data.guildId,
        id,
        warns: data.warns,
        mutes: data.mutes,
        kicks: data.kicks,
        bans: data.bans,
        punishmentreason: punishReason,
        punishmentinfo: punishInfo
    };

    switch (type)
    {
        case `warns`: {
            data.warns += amount;
            break;
        }
        case `mutes`: {
            data.mutes += amount;
            break;
        }
        case `kicks`: {
            data.kicks += amount;
            break;
        }
        case `bans`: {
            data.bans += amount;
            break;
        }
        default: {
            return;
        }
    }

    global.setInfo.run(data);
}

/**
 * @name hasRecord
 * @description Checks if a user has any data in their punishments profile.
 * @param {string} guildId The guild ID we're in.
 * @param {string} id The user ID to check for punishment data.
 * @returns {boolean} Whether or not the user has punishment data.
 */
function hasRecord(guildId, id)
{
    const record = sql.prepare(`SELECT punishmentreason FROM punishments WHERE guildId = ? AND id = ?`).get(guildId, id);
    if (record.punishmentreason === `[]`)
        return false;
    return true;
}
