import { ButtonStyle, ChannelType } from "discord.js";
import common from "./common.js";

export default {
    makeTicket, closeTicket, reopenTicket, saveTranscriptAdmin, saveTranscriptUser
};

/**
 * @name checkTicketPerms
 * @description Checks if the user has permission to either open a ticket at all, or administrate a ticket.
 * @param {Discord.Client} client The client
 * @param {Discord.Interaction} interaction The interaction object.
 * @param {boolean} checkIfBanned Whether or not to check if the user is banned.
 * @param {boolean} checkIfAdmin Whether or not to check if the user is a ticket admin.
 */
async function checkTicketPerms(client, interaction, checkIfBanned, checkIfAdmin)
{
    let requester = interaction.user;
    const guildId = interaction.guild.id;
    const { bannedfromTicketsRoleId, ticketAdminRoleId } = common.getGuildConfig(guildId);

    requester = await common.getGuildMember(client, guildId, requester.id, true);
    if (!requester)
        return false;

    if (bannedfromTicketsRoleId.length > 10 && requester.roles.cache.has(bannedfromTicketsRoleId) && checkIfBanned)
        return false;
    if (!requester.roles.cache.has(ticketAdminRoleId) && checkIfAdmin)
        return false;
    return true;
}

/**
 * @name makeTicket
 * @description Creates a ticket for the user.
 * @param {Discord.Interaction} interaction The interaction object.
 * @param {*} Discord The Discord object.
 * @param {*} client The client object.
 */
export async function makeTicket(interaction, Discord, client)
{
    await interaction.deferReply({ ephemeral: true });

    const isBanned = await checkTicketPerms(client, interaction, true, false);
    if (!isBanned)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`);

    const requesterId = interaction.user.id;
    let { ticketAdminRoleId } = common.getGuildConfig(interaction.guild.id);
    const { openTixCatId, closedTixCatId } = common.getGuildConfig(interaction.guild.id);

    if (!ticketAdminRoleId)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].cantCreateTicket }`);

    // Check if there is already an existing channel with the same name
    const channelName = `ticket-${ requesterId }`;
    const existingChannel = interaction.guild.channels.cache.find((channel) => channel.name === channelName);
    if (existingChannel)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].alreadyHaveAnOpenTickets }`);

    let openTixCat = interaction.guild.channels.cache.find((channel) => channel.id === openTixCatId);
    let closedTixCat = interaction.guild.channels.cache.find((channel) => channel.id === closedTixCatId);
    if (!openTixCat || !closedTixCat)
    {
        if (!openTixCat)
        {
            openTixCat = await interaction.guild.channels.create(
                {
                    name: `${ global.locale[global.findLocale(interaction.guild.id)].openTickets }`,
                    type: ChannelType.GuildCategory,
                    permissionOverwrites: [
                        {
                            id: interaction.guild.id,
                            deny: [`ViewChannel`]
                        },
                        {
                            id: client.user.id,
                            allow: [`ViewChannel`, `SendMessages`, `AttachFiles`, `AddReactions`, `ManageMessages`, `ManageChannels`],
                        },
                        {
                            id: ticketAdminRoleId,
                            allow: [`ViewChannel`, `SendMessages`, `AttachFiles`, `AddReactions`, `ManageMessages`],
                        }
                    ]
                }
            );
        }
        if (!closedTixCat)
        {
            closedTixCat = await interaction.guild.channels.create(
                {
                    name: `${ global.locale[global.findLocale(interaction.guild.id)].closedTickets }`,
                    type: ChannelType.GuildCategory,
                    permissionOverwrites: [
                        {
                            id: interaction.guild.id,
                            deny: [`ViewChannel`]
                        },
                        {
                            id: client.user.id,
                            allow: [`ViewChannel`, `SendMessages`, `AttachFiles`, `AddReactions`, `ManageMessages`, `ManageChannels`],
                        },
                        {
                            id: ticketAdminRoleId,
                            allow: [`ViewChannel`, `SendMessages`, `AttachFiles`, `AddReactions`, `ManageMessages`],
                        }
                    ]
                }
            );
        }
        // Save cat ID's
        const guildId = interaction.guild.id;
        const guildConfig = global.getConfiguration.get(guildId);
        const GUILD_CONFIG = JSON.parse(guildConfig.config);
        GUILD_CONFIG.openTixCatId = openTixCat.id;
        GUILD_CONFIG.closedTixCatId = closedTixCat.id;
        const data = {
            guildId,
            config: JSON.stringify(GUILD_CONFIG)
        };
        global.setConfiguration.run(data);
    }

    // Create the channel
    const channel = await interaction.guild.channels.create(
        {
            name: channelName,
            type: ChannelType.GuildText,
            parent: `${ openTixCat.id }`,
            permissionOverwrites: [
                {
                    id: interaction.guild.id,
                    deny: [`ViewChannel`],
                },
                {
                    id: requesterId,
                    allow: [`ViewChannel`, `SendMessages`, `AttachFiles`, `AddReactions`],
                },
                {
                    id: client.user.id,
                    allow: [`ViewChannel`],
                },
                {
                    id: ticketAdminRoleId,
                    allow: [`ViewChannel`, `SendMessages`, `AttachFiles`, `AddReactions`, `ManageMessages`],
                },
            ],
        }
    ).catch(() => interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].cantCreateTicket }`));

    // Send the admin buttons in the channel & pin the message.
    // Create a button to open the ticket
    const close = new Discord.ButtonBuilder()
        .setCustomId(`closeticket-${ requesterId }`)
        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].closeTicket }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(false);

    const buttonArray = [close];
    const row = new Discord.ActionRowBuilder()
        .addComponents(buttonArray);

    if (interaction.guild.id === `872330479565680640`)
        ticketAdminRoleId = `<snip>`;
    const welcomeMessage = `${ global.locale[global.findLocale(interaction.guild.id)].ticketWelcomeMessage.replaceAll(`%requesterId%`, `${ requesterId }`).replaceAll(`%ticketAdminRoleId%`, `${ ticketAdminRoleId }`) }`;
    channel.send({ content: welcomeMessage, components: [row] })
        .then((message) =>
        {
            message.pin();
            interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].ticketCreated } <#${ channel.id }>`);
        });
}

/**
 * @name closeTicket
 * @description Closes a ticket for the user.
 * @param {Discord.Interaction} interaction The interaction object.
 * @param {*} Discord The Discord object.
 */
export async function closeTicket(interaction, Discord)
{
    await interaction.deferReply({ ephemeral: true });

    // Get channel
    const channel = interaction.channel;
    // Ticket Owner ID (from customId)
    const ticketOwnerId = interaction.customId.split(`-`)[1];
    // Hide channel from ticket owner
    await channel.permissionOverwrites.edit(ticketOwnerId, {
        ViewChannel: false,
        SendMessages: false,
    });

    // Rename channel
    const channelName = `closed-${ ticketOwnerId }`;
    await channel.setName(channelName);
    await channel.setParent(common.getGuildConfig(interaction.guild.id).closedTixCatId);

    // Send new message in channel
    const reopen = new Discord.ButtonBuilder()
        .setCustomId(`reopenticket-${ ticketOwnerId }`)
        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].reopenTicket }`)
        .setStyle(ButtonStyle.Success)
        .setDisabled(false);

    const saveTranscriptAdmin = new Discord.ButtonBuilder()
        .setCustomId(`savetranscriptadmin-${ ticketOwnerId }`)
        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].saveAdminTranscript }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(false);

    const saveTranscriptUser = new Discord.ButtonBuilder()
        .setCustomId(`savetranscriptuser-${ ticketOwnerId }`)
        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].saveUserTranscript }`)
        .setStyle(ButtonStyle.Secondary)
        .setDisabled(false);

    const buttonArray = [reopen, saveTranscriptAdmin, saveTranscriptUser];
    const row = new Discord.ActionRowBuilder()
        .addComponents(buttonArray);

    const message = `${ global.locale[global.findLocale(interaction.guild.id)].ticketCloseMessage.replaceAll(`%ticketOwnerId%`, `${ ticketOwnerId }`).replaceAll(`%interaction.user.id%`, `${ interaction.user.id }`) }`;
    channel.send({ content: message, components: [row] })
        .then((message) =>
        {
            message.pin();
        });
    interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].ticketClosed }`);
    const disableClose = new Discord.ButtonBuilder()
        .setCustomId(`closeticket-${ ticketOwnerId }`)
        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].closeTicket }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(true);
    const disableButtonArray = [disableClose];
    const disableRow = new Discord.ActionRowBuilder()
        .addComponents(disableButtonArray);
    interaction.message.edit({ components: [disableRow] }); // Disable Close Ticket button when ticket is closed.
}

/**
 * @name reopenTicket
 * @description Reopens a ticket for the user.
 * @param {Discord.Interaction} interaction The interaction object.
 * @param {*} Discord The Discord object.
 * @param {Discord.Client} client The client
 */
export async function reopenTicket(interaction, Discord, client)
{
    await interaction.deferReply({ ephemeral: true });

    const isAdmin = await checkTicketPerms(client, interaction, false, true);
    if (!isAdmin)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`);

    // Get channel
    const channel = interaction.channel;
    // Ticket Owner ID (from customId)
    const ticketOwnerId = interaction.customId.split(`-`)[1];

    await channel.permissionOverwrites.edit(ticketOwnerId, {
        ViewChannel: true,
        SendMessages: true,
    });

    // Rename channel
    const channelName = `ticket-${ ticketOwnerId }`;
    await channel.setName(channelName);
    await channel.setParent(common.getGuildConfig(interaction.guild.id).openTixCatId);

    // Delete previous ticket close message
    await interaction.message.delete();

    // Unlock close ticket button
    const enableClose = new Discord.ButtonBuilder()
        .setCustomId(`closeticket-${ ticketOwnerId }`)
        .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].closeTicket }`)
        .setStyle(ButtonStyle.Danger)
        .setDisabled(false);
    const disableButtonArray = [enableClose];
    const disableRow = new Discord.ActionRowBuilder()
        .addComponents(disableButtonArray);
    // Find first message in channel
    await channel.messages.fetch().then(async (messages) =>
    {
        let firstMessage = messages.reverse();
        [firstMessage] = messages.values();
        firstMessage.edit({ components: [disableRow] });
    });

    // Send new message in channel
    channel.send(`<@!${ ticketOwnerId }>, ${ global.locale[global.findLocale(interaction.guild.id)].ticketHasBeenReopened }`);
    return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id)].ticketReopened }\n\n${ global.locale[global.findLocale(interaction.guild.id)].weMightBeRateLimited } <t:${ Math.round((Date.now() + 600_000) / 1000) }:T> (<t:${ Math.round((Date.now() + 600_000) / 1000) }:R>) ${ global.locale[global.findLocale(interaction.guild.id)].weMightBeRateLimited2 }`);
}

/**
 * @name handleTimeForTicketTranscription
 * @description Converts Epoch into Month/Day/Year Hour:Minute:Second AM/PM.
 * @param {number} timestamp The Epoch timestamp.
 */
async function handleTimeForTicketTranscription(timestamp)
{
    const date = new Date(timestamp);
    return `[${ date.toLocaleDateString(`en-US`, { year: `numeric`, month: `2-digit`, day: `2-digit` }) } ${ date.toLocaleTimeString(`en-US`, { hour: `2-digit`, minute: `2-digit`, second: `2-digit` }) }]`;
}

/**
 * @name saveTranscriptAdmin
 * @description Saves the transcript of a ticket (and everything after) for the admin.
 * @param {Discord.Interaction} interaction The interaction object.
 * @param {*} Discord The Discord object.
 * @param {Discord.Client} client The client
 */
export async function saveTranscriptAdmin(interaction, Discord, client)
{
    await interaction.deferReply({ ephemeral: true });

    const isAdmin = await checkTicketPerms(client, interaction, false, true);
    if (!isAdmin)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`);

    const { auditChannelID } = common.getGuildConfig(interaction.guild.id);
    const auditChannel = interaction.guild.channels.cache.get(auditChannelID);
    if (!auditChannel)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noAuditChannel }`);

    // Get channel
    const channel = interaction.channel;
    // Ticket Owner ID (from customId)
    const ticketOwnerId = interaction.customId.split(`-`)[1];

    const finalArray = [];
    // Save every message in channel
    await channel.messages.fetch().then(async (messages) =>
    {
        for (const message of messages.reverse())
        {
            const realMessage = message[1];
            const messageMeta = `${ await handleTimeForTicketTranscription(realMessage.createdTimestamp) } ${ realMessage.author.username }: ${ realMessage.content }`;
            finalArray.push(messageMeta);
        }
    });
    const finalString = finalArray.join(`\n`);
    const attachment = new Discord.AttachmentBuilder(Buffer.from(finalString), { name: `transcript-${ channel.name }.txt` });
    auditChannel.send({
        content: `<@!${ ticketOwnerId }>'s ${ global.locale[global.findLocale(interaction.guild.id)].ticketTranscript }`,
        files: [attachment]
    });
    await interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].savedTranscript }`);
    return channel.delete();
}

/**
 * @name saveTranscriptUser
 * @description Saves the transcript of a ticket (but not after ticket was closed) for the user.
 * @param {Discord.Interaction} interaction The interaction object.
 * @param {*} Discord The Discord object.
 * @param {Discord.Client} client The client
 */
export async function saveTranscriptUser(interaction, Discord, client)
{
    await interaction.deferReply({ ephemeral: true });

    const isAdmin = await checkTicketPerms(client, interaction, false, true);
    if (!isAdmin)
        return interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].noPermission }`);

    // Get channel
    const channel = interaction.channel;
    // Ticket Owner ID (from customId)
    const ticketOwnerId = interaction.customId.split(`-`)[1];

    const finalArray = [];
    // Save every message in channel
    await channel.messages.fetch({
        before: interaction.message.id,
    }).then(async (messages) =>
    {
        for (const message of messages.reverse())
        {
            const realMessage = message[1];
            const messageMeta = `${ await handleTimeForTicketTranscription(realMessage.createdTimestamp) } ${ realMessage.author.username }: ${ realMessage.content }`;
            finalArray.push(messageMeta);
        }
    });
    const finalString = finalArray.join(`\n`);
    const attachment = new Discord.AttachmentBuilder(Buffer.from(finalString), { name: `transcript-${ channel.name }.txt` });
    const ticketOwner = await common.getGuildMember(client, interaction.guild.id, ticketOwnerId, true);
    if (ticketOwner)
    {
        await ticketOwner.send({
            content: `${ global.locale[global.findLocale(interaction.guild.id)].yourTicketTranscript }`,
            files: [attachment]
        }).then(
            () =>
            {
                interaction.followUp(`${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].sentCopyOfTranscript } <@!${ ticketOwnerId }>!`);
                const reopen = new Discord.ButtonBuilder()
                    .setCustomId(`reopenticket-${ ticketOwnerId }`)
                    .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].reopenTicket }`)
                    .setStyle(ButtonStyle.Success)
                    .setDisabled(false);
                const saveTranscriptAdmin = new Discord.ButtonBuilder()
                    .setCustomId(`savetranscriptadmin-${ ticketOwnerId }`)
                    .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].saveAdminTranscript }`)
                    .setStyle(ButtonStyle.Danger)
                    .setDisabled(false);
                const disableUserButton = new Discord.ButtonBuilder()
                    .setCustomId(`savetranscriptuser-${ ticketOwnerId }`)
                    .setLabel(`${ global.locale[global.findLocale(interaction.guild.id)].saveUserTranscript }`)
                    .setStyle(ButtonStyle.Secondary)
                    .setDisabled(true);
                const disableButtonArray = [reopen, saveTranscriptAdmin, disableUserButton];
                const disableRow = new Discord.ActionRowBuilder()
                    .addComponents(disableButtonArray);
                interaction.message.edit({ components: [disableRow] }); // Disable Send transcript to user if we're already sent it.
            }
        )
            .catch(
                () =>
                {
                    interaction.followUp(
                        {
                            content: `${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].iCouldntDM } <@!${ ticketOwnerId }>. ${ global.locale[global.findLocale(interaction.guild.id, interaction.locale)].pleaseSendThemTranscript }`,
                            files: [attachment]
                        }
                    );
                }
            );
    }
}
