import { createRequire } from "node:module";
import DarkDashboard from "dbd-dark-dashboard";
import DBD from "discord-dashboard";
import fs from "node:fs";
import utils from "./utils.js";
import common from "../utils/common.js";
import indexCards from "./index-cards.js";
import commands from "./commands.js";
import epochRender from "./epoch/render.js";

const require = createRequire(import.meta.url);

/**
 * @name bootWeb
 * @description Starts the bot dashboard.
 * @param {*} client The client object.
 * @param {Discord} Discord The Discord object.
 */
export default async function bootWeb(client, Discord)
{
    const {
        webfrontAPIKey, clientSecret, webfrontPort, webfrontURL, botOwnerID
    } = require(`../config.json`);

    await DBD.useLicense(webfrontAPIKey);
    DBD.Dashboard = DBD.UpdatedClass();

    // read in ./webfront/privacy-policy.html as text
    const privacyPolicyHTML = fs.readFileSync(`./src/webfront/privacy-policy.html`, `utf8`);

    // take webfrontURL and remove the trailing port number
    const domain = webfrontURL.replace(/:\d+$/, ``);

    // generate array for guild language option
    const guildLanguages = [];
    const guildLanguageNames = [];
    const locales = fs.readdirSync(`./src/locale`).filter((file) => file.endsWith(`.json`));
    for (const locale of locales)
    {
        const localeName = locale.split(`.`)[0];
        guildLanguages.push(localeName);
        const data = JSON.parse(fs.readFileSync(`./src/locale/${ locale }`));
        // eslint-disable-next-line no-underscore-dangle
        guildLanguageNames.push(data._name);
    }
    const guildLanguageMap = {};
    for (let index = 0; index < guildLanguages.length; index++)
        guildLanguageMap[guildLanguageNames[index]] = guildLanguages[index];

    const clientId = client.user.id;
    const Dashboard = new DBD.Dashboard({
        bot: client,
        Discord,
        requiredPermissions: [DBD.DISCORD_FLAGS.Permissions.KICK_MEMBERS],
        acceptPrivacyPolicy: true,
        port: webfrontPort,
        client: {
            id: clientId,
            secret: clientSecret
        },
        redirectUri: `${ webfrontURL }/discord/callback`,
        domain: `${ domain }`,
        minimizedConsoleLogs: true,
        disableExternalStatistics: true,
        invite: {
            clientId,
            scopes: [`bot`, `applications.commands`],
            permissions: `8`
        },
        customPages: [
            DBD.customPagesTypes.renderHtml(`/epoch`, () => epochRender()),
            DBD.customPagesTypes.renderHtml(`/embed`, () => `<h1>Use /embed in Discord.</h1>`),
        ],
        theme: DarkDashboard({
            information: {
                createdBy: ``,
                websiteTitle: `Bot Dashboard`,
                websiteName: `Bot Dashboard`,
                websiteUrl: ``,
                dashboardUrl: `${ webfrontURL }`,
                supporteMail: ``,
                supportServer: ``,
                imageFavicon: ``,
                iconURL: ``,
                pageBackGround: `linear-gradient(#2CA8FF, #155b8d)`,
                loggedIn: `Successfully signed in.`,
                mainColor: `#2CA8FF`,
                subColor: `#ebdbdb`,
            },
            index: indexCards,
            commands: [commands],
            privacyPolicy: {
                pp: `${ privacyPolicyHTML }`
            },
            sidebar: {
                keepDefault: true,
                list: [
                    {
                        // Icon does not have to be an svg, any html can be used
                        icon: `<svg style="position: absolute; margin-left: 8px; margin-top: 2px;" xmlns="http://www.w3.org/2000/svg" height="24px" viewBox="0 0 24 24" width="24px" fill="#2CA8FF"><path d="M0 0h24v24H0V0z" fill="none"/><path d="M11.99 2C6.47 2 2 6.48 2 12s4.47 10 9.99 10C17.52 22 22 17.52 22 12S17.52 2 11.99 2zM12 20c-4.42 0-8-3.58-8-8s3.58-8 8-8 8 3.58 8 8-3.58 8-8 8zm.5-13H11v6l5.25 3.15.75-1.23-4.5-2.67z"/></svg>`,
                        title: `Epoch Generator`,
                        link: `/epoch`,
                        id: `epoch`,
                    },
                ],
            },
        }),
        settings: [
            {
                categoryId: `rolesettings`,
                categoryName: `Role Settings`,
                categoryDescription: ``,
                categoryOptionsList: [
                    {
                        optionId: `viproleid`,
                        optionName: `VIP Role (Required)`,
                        optionDescription: `Anyone who has this role's ID or higher will be able to: bypass bad word detection, send stickers, mention people with No Mentions role, send whatever links they want, send zalgo, and will not trigger the auto responses. If you do not have a "VIP" role, set this to your Moderator role ID.`,
                        optionType: DBD.formTypes.rolesSelect(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `vipRoleID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `vipRoleID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `vipRoleID`, newData);
                            await sendAlert(guild.id, user.id, `VIP Role ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `mutedroleid`,
                        optionName: `Muted Role`,
                        optionDescription: `If not used, then none of the mute related commands will run.`,
                        optionType: DBD.formTypes.rolesSelect(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `mutedRoleID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `mutedRoleID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `mutedRoleID`, newData);
                            await sendAlert(guild.id, user.id, `Muted Role ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `nopingroleid`,
                        optionName: `No Mentions Role`,
                        optionDescription: `Used for admins who do not want to be mentioned. Target will be auto warned if they mention anyone with the role.`,
                        optionType: DBD.formTypes.rolesSelect(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `noPingRoleId`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `noPingRoleId`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `noPingRoleId`, newData);
                            await sendAlert(guild.id, user.id, `No Mentions Role ID`, newData, client, Discord);
                            return {};
                        },
                    },
                    {
                        optionId: `autojoinroleid`,
                        optionName: `Auto Join Role`,
                        optionDescription: `New members will have this role automatically added when they join.`,
                        optionType: DBD.formTypes.rolesSelect(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `autoJoinRoleID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `autoJoinRoleID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `autoJoinRoleID`, newData);
                            await sendAlert(guild.id, user.id, `Auto Join Role ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `bannedfromticketsid`,
                        optionName: `Banned from Tickets Role`,
                        optionDescription: `Anyone who has this role will be unable to create tickets.`,
                        optionType: DBD.formTypes.rolesSelect(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `bannedfromTicketsRoleId`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `bannedfromTicketsRoleId`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `bannedfromTicketsRoleId`, newData);
                            await sendAlert(guild.id, user.id, `Banned from Tickets Role ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `ticketadminrole`,
                        optionName: `Ticket Access Role`,
                        optionDescription: `Anyone who has this role will be able to manage tickets.`,
                        optionType: DBD.formTypes.rolesSelect(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `ticketAdminRoleId`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `ticketAdminRoleId`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `ticketAdminRoleId`, newData);
                            await sendAlert(guild.id, user.id, `Ticket Access Role ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                ],
            },
            {
                categoryId: `channelsettings`,
                categoryName: `Channel Settings`,
                categoryDescription: ``,
                categoryOptionsList: [
                    {
                        optionId: `auditChannelid`,
                        optionName: `Audit Channel (Required)`,
                        optionDescription: `Where the bot will send audit logs.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `auditChannelID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `auditChannelID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `auditChannelID`, newData);
                            await sendAlert(guild.id, user.id, `Audit Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `deletedmessageschanid`,
                        optionName: `Deleted Messages / Edits Channel`,
                        optionDescription: `Where the bot will send automatically deleted messages and edited messages as a log.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `deletedMessagesChanID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `deletedMessagesChanID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `deletedMessagesChanID`, newData);
                            await sendAlert(guild.id, user.id, `Deleted Messages Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `generalchannelid`,
                        optionName: `General Channel`,
                        optionDescription: `If not used, then the Steam Sale for MW2 will not fire, and the warn command will not attempt to contact the user if their DMs are closed.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `generalChannelID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `generalChannelID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `generalChannelID`, newData);
                            await sendAlert(guild.id, user.id, `General Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `suggestionchannelid`,
                        optionName: `Suggestions Message Channel`,
                        optionDescription: `If not used, then the suggest command will not work.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `suggestionChannelID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `suggestionChannelID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `suggestionChannelID`, newData);
                            await sendAlert(guild.id, user.id, `Suggestions Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `regionalchannelid`,
                        optionName: `Regional Message Channel`,
                        optionDescription: `If not used, then regional keywords are ignored.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `regionalChannelID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `regionalChannelID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `regionalChannelID`, newData);
                            await sendAlert(guild.id, user.id, `Regional Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `lockdownchannelids`,
                        optionName: `Lockdown Channels`,
                        optionDescription: `If not used, then the lockdown command will not work.`,
                        optionType: DBD.formTypes.channelsMultiSelect(false, false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `lockdownChannelIDs`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `lockdownChannelIDs`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `lockdownChannelIDs`, newData);
                            await sendAlert(guild.id, user.id, `Lockdown Channel IDs`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `banappealschannelid`,
                        optionName: `Ban Appeals Channel`,
                        optionDescription: `Used for automated unbans. If you do not use our ban appeals platform, you can safely ignore this.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `banAppealsChannelID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `banAppealsChannelID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `banAppealsChannelID`, newData);
                            await sendAlert(guild.id, user.id, `Ban Appeals Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },

                    {
                        optionType: `spacer`,
                        title: `Server Statistics Settings`,
                        description: `Settings that are used for server statistics.`
                    },
                    {
                        optionId: `serverstatscatname`,
                        optionName: `Statistics Category Name`,
                        optionDescription: `Name of the category for server stats`,
                        optionType: DBD.formTypes.input(`📊 Server Stats 📊`, 0, 100, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `serverStatsCatName`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `serverStatsCatName`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `serverStatsCatName`, newData);
                            await sendAlert(guild.id, user.id, `Statistics Category Name`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `serverstatsmemberchanname`,
                        optionName: `Statistics Members Channel Name`,
                        optionDescription: `Name of the channel for Members stats`,
                        optionType: DBD.formTypes.input(`Members: `, 0, 100, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `serverStatsMemberChanName`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `serverStatsMemberChanName`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `serverStatsMemberChanName`, newData);
                            await sendAlert(guild.id, user.id, `Statistics Members Channel Name`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `serverstatsboosterchanname`,
                        optionName: `Statistics Boosters Channel Name`,
                        optionDescription: `Name of the channel for Booster stats`,
                        optionType: DBD.formTypes.input(`Boosters: `, 0, 100, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `serverStatsBoosterChanName`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `serverStatsBoosterChanName`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `serverStatsBoosterChanName`, newData);
                            await sendAlert(guild.id, user.id, `Statistics Boosters Channel Name`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },

                    {
                        optionType: `spacer`,
                        title: `Welcome/Goodbye Settings`,
                        description: `Settings that are used for welcoming/saying goodbye to members.`,
                    },
                    {
                        optionId: `welcomechannelid`,
                        optionName: `Welcome/Leave Channel`,
                        optionDescription: `Channel to welcome/say goodbye in.`,
                        optionType: DBD.formTypes.channelsSelect(false, [`GUILD_TEXT`]),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `welcomeChannelID`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `welcomeChannelID`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `welcomeChannelID`, newData);
                            await sendAlert(guild.id, user.id, `Welcome Channel ID`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `welcomemsg`,
                        optionName: `Welcome Message`,
                        optionDescription: `Message to send to the welcome channel when a user joins. Use {user} to mention the user.`,
                        optionType: DBD.formTypes.textarea(``, 0, 1000, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `welcomeMessage`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `welcomeMessage`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `welcomeMessage`, newData);
                            await sendAlert(guild.id, user.id, `Welcome Message`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `goodbyemsg`,
                        optionName: `Goodbye Message`,
                        optionDescription: `Message to send to the welcome channel when a user leaves. Use {user} to mention the user.`,
                        optionType: DBD.formTypes.textarea(``, 0, 1000, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `leavingMessage`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `leavingMessage`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `leavingMessage`, newData);
                            await sendAlert(guild.id, user.id, `Leaving Message`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                ],
            },
            {
                categoryId: `manipulationsettings`,
                categoryName: `Auto Mod Settings`,
                categoryDescription: ``,
                categoryOptionsList: [
                    {
                        optionId: `automodmode`,
                        optionName: `Auto Mod Mode`,
                        optionDescription: `Note: We do not warn for using bad words/sending links at this time.`,
                        optionType: DBD.formTypes.select(({ "Delete Message": `delete`, "Delete Message & Warn User": `warn` }), false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `autoModMode`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `autoModMode`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `autoModMode`, newData);
                            await sendAlert(guild.id, user.id, `Auto Mod Mode`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `automodcaps`,
                        optionName: `Auto Moderate against excessive caps?`,
                        optionDescription: `If 70% of the message is caps, and the message is longer than 20 characters, we will moderate against it.`,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `autoModCaps`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `autoModCaps`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `autoModCaps`, newData);
                            await sendAlert(guild.id, user.id, `Auto Mod Caps`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `autowarnforamountofmentions`,
                        optionName: `Warn for X amount of mentions?`,
                        optionDescription: `If this is set to any number higher than 0, the bot will warn the user if they mention that amount of people in a single message.`,
                        optionType: DBD.formTypes.input(`4`, 1, 2, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `autoWarnForAmountOfMentions`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `autoWarnForAmountOfMentions`);
                            if (Number(oldData) === Number(newData))
                                return {};
                            if (Number.isNaN(newData))
                                return { error: `This should be a number.` };
                            utils.saveGuildConfigValue(guild, `autoWarnForAmountOfMentions`, Number(newData));
                            await sendAlert(guild.id, user.id, `Warn for Amount of Mentions`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `amountofspoilers`,
                        optionName: `Amount of Spoilers to warn for?`,
                        optionDescription: `If set to any number higher than 0, the bot will warn the user if they have that amount of spoilers in a single message.`,
                        optionType: DBD.formTypes.input(`-1`, 1, 2, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `amountOfSpoilers`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `amountOfSpoilers`);
                            if (Number(oldData) === Number(newData))
                                return {};
                            if (Number.isNaN(newData))
                                return { error: `This should be a number.` };
                            utils.saveGuildConfigValue(guild, `amountOfSpoilers`, Number(newData));
                            await sendAlert(guild.id, user.id, `Amount of Spoilers`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: ``,
                        description: ``,
                    },
                    /*
                    TODO: Change mentions of `textarea` below this to array type when dbd-dashboard updates.
                    */
                    {
                        optionId: `badwords`,
                        optionName: `Bad Words`,
                        optionDescription: `A list of words that will be automatically deleted.`,
                        optionType: DBD.formTypes.textarea(``, 0, 999_999_999, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = await utils.getArrayData(guild, `badWords`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `badWords`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            await utils.saveArrayData(guild, `badWords`, newData);
                            await sendAlert(guild.id, user.id, `Bad Words`, newData, client, Discord);
                            return {};
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: `Auto Ban Settings`,
                        description: `Settings that are used for the automated ban system.`,
                    },
                    {
                        optionId: `autobanmode`,
                        optionName: `Auto Ban Mode`,
                        optionDescription: `What should we detect in order to auto ban someone who uses a phrase in the auto ban list?`,
                        optionType: DBD.formTypes.select(({ "Just the phrase (NOT RECOMMENDED)": `phrase`, "Phrase + Link": `link`, "Phrase + Link + @everyone": `all` }), false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `autoBanMode`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `autoBanMode`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `autoBanMode`, newData);
                            await sendAlert(guild.id, user.id, `Auto Ban Mode`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `autobannedphrases`,
                        optionName: `Auto Banned Phrases`,
                        optionDescription: `A list of phrases that will result in the user being automatically banned.`,
                        optionType: DBD.formTypes.textarea(``, 0, 999_999_999, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = await utils.getArrayData(guild, `autoBannedPhrases`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `autoBannedPhrases`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            await utils.saveArrayData(guild, `autoBannedPhrases`, newData);
                            await sendAlert(guild.id, user.id, `Auto Banned Phrases`, newData, client, Discord);
                            return {};
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: `Links / Discord Settings`,
                        description: `Settings that affect links and discord invites.`,
                    },
                    {
                        optionId: `allowdenymode`,
                        optionName: `Links / Discord Mode`,
                        optionDescription: `Do we treat the links list & discord list as a allowlist or a denylist?`,
                        optionType: DBD.formTypes.select(({ Allowlist: `allow`, Denylist: `deny` }), false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `allowDenyMode`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowDenyMode`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `allowDenyMode`, newData);
                            await sendAlert(guild.id, user.id, `Links / Discord Mode`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `allowedlinks`,
                        optionName: `Links List`,
                        optionDescription: `A list of links that are either allowed or blocked based on your Auto Moderation settings.`,
                        optionType: DBD.formTypes.textarea(``, 0, 999_999_999, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = await utils.getArrayData(guild, `allowedLinks`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowedLinks`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            await utils.saveArrayData(guild, `allowedLinks`, newData);
                            await sendAlert(guild.id, user.id, `Links List`, newData, client, Discord);
                            return {};
                        },
                    },
                    {
                        optionId: `alloweddiscords`,
                        optionName: `Discord Server IDs`,
                        optionDescription: `A list of Discord Server IDs that are either allowed or blocked based on your Auto Moderation settings.`,
                        optionType: DBD.formTypes.textarea(``, 0, 999_999_999, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = await utils.getArrayData(guild, `allowedDiscords`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowedDiscords`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            await utils.saveArrayData(guild, `allowedDiscords`, newData);
                            await sendAlert(guild.id, user.id, `Discord List`, newData, client, Discord);
                            return {};
                        },
                    },
                ],
            },
            {
                categoryId: `othersettings`,
                categoryName: `Other Settings`,
                categoryDescription: ``,
                categoryOptionsList: [
                    {
                        optionId: `locale`,
                        optionName: `Guild Language?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.select(guildLanguageMap, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `locale`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `locale`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `locale`, newData);
                            await sendAlert(guild.id, user.id, `Guild Language`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: `True / False Settings`,
                        description: `Settings that are toggleable.`,
                    },
                    {
                        optionId: `allowanyonetosendstickers`,
                        optionName: `Allow Anyone to send stickers?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `allowAnyoneToSendStickers`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowAnyoneToSendStickers`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `allowAnyoneToSendStickers`, newData);
                            await sendAlert(guild.id, user.id, `Allow Anyone to send stickers`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `allowanyonewithroletobypassbadwordbanning`,
                        optionName: `Allow Anyone with a role to bypass bad word auto deletion?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `allowAnyoneWithRoleToBypassBadWordBanning`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowAnyoneWithRoleToBypassBadWordBanning`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `allowAnyoneWithRoleToBypassBadWordBanning`, newData);
                            await sendAlert(guild.id, user.id, `Allow anyone with role to bypass bad words`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `allowanyonewithroletobypasstriggerresponse`,
                        optionName: `Allow Anyone with a role to bypass the bot's auto replies?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `allowAnyoneWithRoleToBypassTriggerResponse`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowAnyoneWithRoleToBypassTriggerResponse`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `allowAnyoneWithRoleToBypassTriggerResponse`, newData);
                            await sendAlert(guild.id, user.id, `Allow anyone with role to bypass auto replies`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `allowanyonewithroletosendlinks`,
                        optionName: `Allow Anyone with a role to send any link?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `allowAnyoneWithRoleToSendLinks`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `allowAnyoneWithRoleToSendLinks`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `allowAnyoneWithRoleToSendLinks`, newData);
                            await sendAlert(guild.id, user.id, `Allow anyone with a role to send links`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `renameuntaggablenames`,
                        optionName: `Do we rename untaggable usernames?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `renameUntaggableNames`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `renameUntaggableNames`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `renameUntaggableNames`, newData);
                            await sendAlert(guild.id, user.id, `Rename Untaggable Names`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `renameto`,
                        optionName: `What should we rename untaggable names to (if enabled)?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.input(``, 3, 30, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `renameTo`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `renameTo`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `renameTo`, newData);
                            await sendAlert(guild.id, user.id, `Rename Untaggable Names To`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `logselfdeletes`,
                        optionName: `Log Self Deletes?`,
                        optionDescription: `Note: This will log when commands are automatically deleted, there is nothing we can do about this.`,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `logSelfDeletes`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `logSelfDeletes`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `logSelfDeletes`, newData);
                            await sendAlert(guild.id, user.id, `Log Self Deletes`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `logselfedits`,
                        optionName: `Log Self Edits?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `logSelfEdits`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `logSelfEdits`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `logSelfEdits`, newData);
                            await sendAlert(guild.id, user.id, `Log Self Edits`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `trackinvites`,
                        optionName: `Track Invites?`,
                        optionDescription: `If enabled we will send a message in audit channel when someone joins attempting to find out who invited them.`,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `trackInvites`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `trackInvites`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `trackInvites`, newData);
                            await sendAlert(guild.id, user.id, `Track Invites`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `reporting`,
                        optionName: `Allow Users to report messages in the App Menu?`,
                        optionDescription: `If you change this setting it will take affect on the next bot restart.`,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `reportApplet`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `reportApplet`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `reportApplet`, newData);
                            await sendAlert(guild.id, user.id, `Report Applet`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: `Ban Settings`,
                        description: `Settings that affect bans.`,
                    },
                    {
                        optionId: `senddmonban`,
                        optionName: `Do we DM the user upon banning?`,
                        optionDescription: ``,
                        optionType: DBD.formTypes.switch(false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `sendDmOnBan`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `sendDmOnBan`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `sendDmOnBan`, newData);
                            await sendAlert(guild.id, user.id, `Send DM on ban`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `banappealslink`,
                        optionName: `Ban Appeals Link`,
                        optionDescription: `Should be a string of your ban appeals site/form.`,
                        optionType: DBD.formTypes.input(`https://`, 5, 100, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `banAppealsLink`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `banAppealsLink`);

                            const dontSend = await dontSendChanges(oldData, newData);
                            if (dontSend)
                                return {};

                            if (newData.startsWith(`http`))
                            {
                                if (newData.endsWith(`/`))
                                    newData = newData.slice(0, -1);
                                utils.saveGuildConfigValue(guild, `banAppealsLink`, newData);
                                await sendAlert(guild.id, user.id, `Ban Appeals Link`, newData, client, Discord);
                                return {};
                            }
                            if (newData === ``)
                            {
                                utils.saveGuildConfigValue(guild, `banAppealsLink`, ``);
                                await sendAlert(guild.id, user.id, `Ban Appeals Link`, newData, client, Discord);
                                return {};
                            }
                            return { error: `This should start with http.` };
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: ``,
                        description: ``,
                    },
                    {
                        optionId: `altdetectiondays`,
                        optionName: `Minimum Account Detection Days`,
                        optionDescription: `DO NOT SET THIS UNLESS YOU KNOW WHAT YOU ARE DOING. If this is set to any number higher than 0, we will automatically ban any user below this threshold and DM them how to restore their account. Do not use this feature unless you use our appeals project.`,
                        optionType: DBD.formTypes.input(`2`, 1, 10, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `altDetectionDays`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `altDetectionDays`);
                            if (Number(oldData) === Number(newData))
                                return {};
                            if (Number.isNaN(newData))
                                return { error: `This should be a number.` };
                            utils.saveGuildConfigValue(guild, `altDetectionDays`, Number(newData));
                            await sendAlert(guild.id, user.id, `Minimum Account Detection Days`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: `Keyword Spam Settings`,
                        description: `Settings that are used for modifying how we detect / punish people who trigger keywords on purpose.`,
                    },
                    {
                        optionId: `amountofkeywordspam`,
                        optionName: `Amount of Keywords to detect as spamming`,
                        optionDescription: `How many times can they trigger the keywords before we treat them as a spammer.`,
                        optionType: DBD.formTypes.input(`-1`, 1, 2, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `amountOfKeywordSpam`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `amountOfKeywordSpam`);
                            if (Number(oldData) === Number(newData))
                                return {};
                            if (Number.isNaN(newData))
                                return { error: `This should be a number.` };
                            utils.saveGuildConfigValue(guild, `amountOfKeywordSpam`, Number(newData));
                            await sendAlert(guild.id, user.id, `Amount of Keywords to detect as spamming`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `timebeforerefreshingkeywordspaminminutes`,
                        optionName: `Time (in minutes) before refreshing keywords spam`,
                        optionDescription: `Amount of time that needs to pass before we reset a potential keyword spammer's spam counter.`,
                        optionType: DBD.formTypes.input(`-1`, 1, 3, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `timeBeforeRefreshingKeywordSpamInMinutes`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `timeBeforeRefreshingKeywordSpamInMinutes`);
                            if (Number(oldData) === Number(newData))
                                return {};
                            if (Number.isNaN(newData))
                                return { error: `This should be a number.` };
                            utils.saveGuildConfigValue(guild, `timeBeforeRefreshingKeywordSpamInMinutes`, Number(newData));
                            await sendAlert(guild.id, user.id, `Time before refreshing keyword spam`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `timetomuteforkeywordspaminminutes`,
                        optionName: `Time (in minutes) we mute for`,
                        optionDescription: `Amount of time to mute a user spamming keywords for.`,
                        optionType: DBD.formTypes.input(`-1`, 1, 5, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const value = utils.getGuildConfigValue(guild, `timeToMuteForKeywordSpamInMinutes`);
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            const oldData = utils.getGuildConfigValue(guild, `timeToMuteForKeywordSpamInMinutes`);
                            if (Number(oldData) === Number(newData))
                                return {};
                            if (Number.isNaN(newData))
                                return { error: `This should be a number.` };
                            utils.saveGuildConfigValue(guild, `timeToMuteForKeywordSpamInMinutes`, Number(newData));
                            await sendAlert(guild.id, user.id, `Time to mute keyword spammers for`, newData, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionType: `spacer`,
                        title: `Chat Command Settings`,
                        description: `Settings that are used for chat commands.`,
                    },
                    {
                        optionId: `prefix`,
                        optionName: `Prefix`,
                        optionDescription: `Starting prefix for chat commands. (To use "," as a [part of a] prefix, placeholder it with "%c")`,
                        optionType: DBD.formTypes.textarea(``, 0, 999_999_999, false, false),
                        getActualSet: async ({ guild }) =>
                        {
                            const prefix = utils.getGuildConfigValue(guild, `prefix`);

                            let string = ``;
                            for (let element of prefix)
                            {
                                if (element !== `cdb!`)
                                {
                                    if (element.includes(`,`))
                                        element = element.replace(`,`, `%c`);
                                    string += `${ element },`;
                                }
                            }

                            if (string.charAt(string.length - 1) === `,`)
                                string = string.slice(0, -1);

                            return string;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            // get new data
                            const newDataArray = [];

                            newDataArray.push(`cdb!`); // add internal prefix
                            if (newData.includes(`,`))
                            {
                                const splittedNewData = newData.split(`,`);
                                for (let prefix of splittedNewData)
                                {
                                    prefix = prefix.replace(`%c`, `,`);
                                    newDataArray.push(prefix);
                                }
                            } // don't break if there's only one value.
                            else
                                newDataArray.push(newData);

                            // remove duplicates from newDataArray
                            const newDataArrayNoDuplicates = [];
                            for (const element of newDataArray)
                            {
                                if (!newDataArrayNoDuplicates.includes(element) && element !== ``)
                                    newDataArrayNoDuplicates.push(element);
                            }

                            const oldData = utils.getGuildConfigValue(guild, `prefix`);
                            const dontSend = await dontSendChanges(oldData, newDataArrayNoDuplicates);
                            if (dontSend)
                                return {};

                            utils.saveGuildConfigValue(guild, `prefix`, newDataArrayNoDuplicates);
                            await sendAlert(guild.id, user.id, `Prefix`, newDataArrayNoDuplicates, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                ],
            },
            {
                categoryId: `apikeys`,
                categoryName: `API Keys`,
                categoryDescription: ``,
                categoryOptionsList: [
                    {
                        optionId: `itadapikey`,
                        optionName: `ITAD API Key`,
                        optionDescription: `Your API Key from IsThereAnyDeal.com. If not used, then the Steam Sale for MW2 will not fire.`,
                        optionType: DBD.formTypes.input(``, 5, 100, false, false),
                        getActualSet: async () =>
                        {
                            // const value = utils.getGuildConfigValue(guild, `ITADAPIKey`);
                            const value = `SECRET`;
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            utils.saveGuildConfigValue(guild, `ITADAPIKey`, newData);
                            await sendAlert(guild.id, user.id, `ITAD API Key`, `SECRET`, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                    {
                        optionId: `deeplapikey`,
                        optionName: `DeepL API Key`,
                        optionDescription: `Your API Key from DeepL. If not used, then the translate functionality will not work.`,
                        optionType: DBD.formTypes.input(``, 5, 100, false, false),
                        getActualSet: async () =>
                        {
                            // const value = utils.getGuildConfigValue(guild, `DeepLAPIKey`);
                            const value = `SECRET`;
                            return value;
                        },
                        setNew: async ({ guild, user, newData }) =>
                        {
                            utils.saveGuildConfigValue(guild, `DeepLAPIKey`, newData);
                            await sendAlert(guild.id, user.id, `DeepL API Key`, `SECRET`, client, Discord);
                            return {};
                        },
                        allowedCheck: async ({ guild, user }) =>
                        {
                            const properUser = await common.getGuildMember(client, guild.id, user.id);
                            if (properUser && !properUser.permissions.has(`ADMINISTRATOR`)
                                && !botOwnerID.includes(user.id))
                                return { allowed: false, errorMessage: `You cannot use this option.` };
                            return { allowed: true, errorMessage: undefined };
                        },
                    },
                ],
            },
        ],
    });

    Dashboard.init();
}

/**
 * @param {string} guildId The guild to get the data for.
 * @param {string} userId The user who is changing the data.
 * @param {string} option The data option to get.
 * @param {string} value The data to save.
 * @param {Discord.Client} client The client object.
 * @param {Discord} Discord The Discord object.
 */
async function sendAlert(guildId, userId, option, value, client, Discord)
{
    const guild = client.guilds.cache.get(guildId);
    const auditChannelID = utils.getGuildConfigValue(guild, `auditChannelID`);
    const auditChannel = await guild.channels.cache.get(auditChannelID);
    if (!auditChannel) return;
    if (value.length === 0)
        value = `<empty>`;
    value = common.truncateString(value, 850);
    const message = common.truncateString(`\`${ option }\` has been set to \`${ value }\` by <@!${ userId }>.`, 1024);
    const embed = new Discord.EmbedBuilder()
        .setColor(`#0099ff`)
        .setTitle(`🔧 Setting Changed`)
        .setDescription(message);
    auditChannel.send({ embeds: [embed] });
}

/**
 * @param {string} oldData Old data of action.
 * @param {string} newData New data of action.
 */
async function dontSendChanges(oldData, newData)
{
    if (!oldData && !newData)
        return true;
    if (oldData && oldData.length === 0 && newData && newData.length === 0)
        return true;
    // alternate to comparing objects? javascript is completely stupid
    if (typeof oldData === `object` && typeof newData === `object` && oldData.toString() === newData.toString())
        return true;
    if (oldData === newData)
        return true;

    return false;
}
