import SQLite from "better-sqlite3";

export default {
    getGuildConfigValue,
    saveGuildConfigValue,
    getArrayData,
    saveArrayData
};

/**
 * @name getGuildConfigValue
 * @description Returns the value of a guild config.
 * @param {object} guild The guild to get the config for.
 * @param {string} option The config option to get.
 * @returns {string} The value of the config option.
 */
function getGuildConfigValue(guild, option)
{
    const guildId = guild.id;
    const guildConfig = global.getConfiguration.get(guildId);
    const GUILD_CONFIG = JSON.parse(guildConfig.config);
    return GUILD_CONFIG[option];
}

/**
 * @name saveGuildConfigValue
 * @description Saves a value to a guild config.
 * @param {object} guild The guild to save the config for.
 * @param {string} option The config option to save.
 * @param {string} value The value to save.
 */
function saveGuildConfigValue(guild, option, value)
{
    const guildId = guild.id;
    const guildConfig = global.getConfiguration.get(guildId);
    const GUILD_CONFIG = JSON.parse(guildConfig.config);
    GUILD_CONFIG[option] = value;
    if (value === `SECRET`)
        return; // Don't allow overwriting SECRET values.
    const guildConfigString = JSON.stringify(GUILD_CONFIG);
    const data = {
        guildId,
        config: guildConfigString
    };
    global.setConfiguration.run(data);
}

/**
 * @name getArrayData
 * @description Returns data as an array.
 * @param {object} guild The guild to get the data for.
 * @param {string} option The data option to get.
 * @returns {string[]} The data as an array.
 */
async function getArrayData(guild, option)
{
    const guildId = guild.id;
    const sql = new SQLite(`./src/db.sqlite`);
    switch (option)
    {
        case `badWords`: {
            const badWordsH = sql.prepare(`SELECT word FROM badWords WHERE guildId = ${ guildId }`).raw().all();
            var value = badWordsH.flat().toString();
            break;
        }
        case `allowedLinks`: {
            const allowedLinksH = sql.prepare(`SELECT link FROM allowedLinks WHERE guildId = ${ guildId }`).raw().all();
            var value = allowedLinksH.flat().toString();
            break;
        }
        case `autoBannedPhrases`: {
            const autoBannedPhrasesH = sql.prepare(`SELECT phrase FROM autoBannedPhrases WHERE guildId = ${ guildId }`).raw().all();
            var value = autoBannedPhrasesH.flat().toString();
            break;
        }
        case `allowedDiscords`: {
            const allowedDiscordsH = sql.prepare(`SELECT id FROM allowedDiscords WHERE guildId = ${ guildId }`).raw().all();
            var value = allowedDiscordsH.flat().toString();
            break;
        }
        // No default
    }
    return value;
}

/**
 * @name saveArrayData
 * @description Saves data as an array.
 * @param {object} guild The guild to save the data for.
 * @param {string} option The data option to save.
 * @param {string} value The data to save.
 * @returns {string[]} The data as an array.
 */
async function saveArrayData(guild, option, value)
{
    const guildId = guild.id;
    const sql = new SQLite(`./src/db.sqlite`);
    switch (option)
    {
        case `badWords`: {
            // Empty out the bad words table for guildId
            sql.prepare(`DELETE FROM badWords WHERE guildId = ${ guildId }`).run();
            // Take value, split it into an array, and then save each word to the database.
            if (value.includes(`,`))
                var badWords = value.split(`,`);
            else // Don't break if there's only one value.
                var badWords = [value];
            for (let word of badWords)
            {
                word = word.toLowerCase().trim(); // removes whitespace
                const data = {
                    guildId,
                    word
                };
                sql.prepare(`INSERT INTO badWords (guildId, word) VALUES (@guildId, @word)`).run(data);
            }
            break;
        }
        case `allowedLinks`: {
            // Empty out the allowed links table for guildId
            sql.prepare(`DELETE FROM allowedLinks WHERE guildId = ${ guildId }`).run();
            // Take value, split it into an array, and then save each link to the database.
            if (value.includes(`,`))
                var allowedLinks = value.split(`,`);
            else // Don't break if there's only one value.
                var allowedLinks = [value];
            for (let link of allowedLinks)
            {
                link = link.toLowerCase().trim(); // removes whitespace
                link = link.replace(/http:\/\//gi, ``).replace(/https:\/\//gi, ``).replace(/www\./gi, ``); // Get raw domain only
                link = link.split(`?`)[0]; // Strip anything after ?
                link = link.split(`#`)[0]; // Strip anything after #
                link = link.replace(/\/$/, ``); // Strip last / if it exists
                const data = {
                    guildId,
                    link
                };
                sql.prepare(`INSERT INTO allowedLinks (guildId, link) VALUES (@guildId, @link)`).run(data);
            }
            break;
        }
        case `autoBannedPhrases`: {
            // Empty out the auto banned phrases table for guildId
            sql.prepare(`DELETE FROM autoBannedPhrases WHERE guildId = ${ guildId }`).run();
            // Take value, split it into an array, and then save each phrase to the database.
            if (value.includes(`,`))
                var autoBannedPhrases = value.split(`,`);
            else // Don't break if there's only one value.
                var autoBannedPhrases = [value];
            for (let phrase of autoBannedPhrases)
            {
                phrase = phrase.toLowerCase().trim(); // removes whitespace
                const data = {
                    guildId,
                    phrase
                };
                sql.prepare(`INSERT INTO autoBannedPhrases (guildId, phrase) VALUES (@guildId, @phrase)`).run(data);
            }
            break;
        }
        case `allowedDiscords`: {
            // Empty out the allowedDiscords table for guildId
            sql.prepare(`DELETE FROM allowedDiscords WHERE guildId = ${ guildId }`).run();
            // Take value, split it into an array, and then save each id to the database.
            if (value.includes(`,`))
                var allowedDiscords = value.split(`,`);
            else // Don't break if there's only one value.
                var allowedDiscords = [value];
            for (let id of allowedDiscords)
            {
                id = id.trim(); // removes whitespace
                const data = {
                    guildId,
                    id
                };
                sql.prepare(`INSERT INTO allowedDiscords (guildId, id) VALUES (@guildId, @id)`).run(data);
            }
            break;
        }
        // No default
    }
}
